;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; secondlife setup.nsi
;; Copyright 2004-2008, Linden Research, Inc.
;;
;; NSIS Unicode 2.38.1 or higher required
;; http://www.scratchpaper.com/
;;
;; Author: James Cook, Don Kjer, Callum Prentice
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Compiler flags
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
SetOverwrite on				; overwrite files
SetCompress auto			; compress iff saves space
SetCompressor /solid lzma	; compress whole installer as one block
SetDatablockOptimize off	; only saves us 0.1%, not worth it
XPStyle on                  ; add an XP manifest to the installer
RequestExecutionLevel admin	; on Vista we must be admin because we write to Program Files

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Project flags
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


        !define INSTEXE  "Ascent.exe"
        !define VERSION "1.4.0"
        !define VERSION_LONG "1.4.0.0"
        !define VERSION_DASHES "1-4-0-0"
        

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Tweak for different servers/builds (this placeholder is replaced by viewer_manifest.py)
;; For example:
;; !define INSTFLAGS "%(flags)s"
;; !define INSTNAME   "SecondLife%(grid_caps)s"
;; !define SHORTCUT   "Second Life (%(grid_caps)s)"
;; !define URLNAME   "secondlife%(grid)s"
;; !define UNINSTALL_SETTINGS 1


                OutFile "Ascent_1-4-0-0_Setup.exe"
                !define VIEWERNAME "Ascent"
                !define INSTFLAGS "--channel $\"Ascent$\"  --settings settings_ascent.xml"
                !define INSTNAME   "Ascent"
                !define SHORTCUT   "Ascent"
                !define URLNAME   "ascent"
                !define INSTALL_ICON "install_icon_snowglobe.ico"
                !define UNINSTALL_ICON "uninstall_icon_snowglobe.ico"
                Caption "Ascent ${VERSION}"
                

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; - language files - one for each language (or flavor thereof)
;; (these files are in the same place as the nsi template but the python script generates a new nsi file in the 
;; application directory so we have to add a path to these include files)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
!include "E:\ascentviewer\indra\newview\installers\windows\lang_de.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_en-us.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_es.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_fr.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_ja.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_it.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_ko.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_nl.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_pt-br.nsi"
!include "E:\ascentviewer\indra\newview\installers\windows\lang_zh.nsi"

# *TODO: Move these into the language files themselves
LangString LanguageCode ${LANG_GERMAN}   "de"
LangString LanguageCode ${LANG_ENGLISH}  "en"
LangString LanguageCode ${LANG_SPANISH}  "es"
LangString LanguageCode ${LANG_FRENCH}   "fr"
LangString LanguageCode ${LANG_JAPANESE} "ja"
LangString LanguageCode ${LANG_ITALIAN}  "it"
LangString LanguageCode ${LANG_KOREAN}   "ko"
LangString LanguageCode ${LANG_DUTCH}    "nl"
LangString LanguageCode ${LANG_PORTUGUESEBR} "pt"
LangString LanguageCode ${LANG_SIMPCHINESE}  "zh"

Name ${INSTNAME}

SubCaption 0 $(LicenseSubTitleSetup)	; override "license agreement" text

BrandingText " "						; bottom of window text
Icon          E:\ascentviewer\indra\newview\installers\windows\install_icon_snowglobe.ico
UninstallIcon E:\ascentviewer\indra\newview\installers\windows\uninstall_icon_snowglobe.ico
WindowIcon on							; show our icon in left corner
BGGradient off							; no big background window
CRCCheck on								; make sure CRC is OK
InstProgressFlags smooth colored		; new colored smooth look
ShowInstDetails nevershow				; no details, no "show" button
SetOverwrite on							; stomp files by default
AutoCloseWindow true					; after all files install, close window

InstallDir "$PROGRAMFILES\${INSTNAME}"
InstallDirRegKey HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\${INSTNAME}" ""
DirText $(DirectoryChooseTitle) $(DirectoryChooseSetup)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Var INSTPROG
Var INSTEXE
Var INSTFLAGS
Var INSTSHORTCUT
Var COMMANDLINE         ; command line passed to this installer, set in .onInit
Var SHORTCUT_LANG_PARAM ; "--set InstallLanguage de", passes language to viewer

;;; Function definitions should go before file includes, because calls to
;;; DLLs like LangDLL trigger an implicit file include, so if that call is at
;;; the end of this script NSIS has to decompress the whole installer before 
;;; it can call the DLL function. JC

!include "FileFunc.nsh"     ; For GetParameters, GetOptions
!insertmacro GetParameters
!insertmacro GetOptions

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; After install completes, launch app
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function .onInstSuccess
    Push $R0	# Option value, unused
    ${GetOptions} $COMMANDLINE "/AUTOSTART" $R0
    # If parameter was there (no error) just launch
    # Otherwise ask
    IfErrors label_ask_launch label_launch
    
label_ask_launch:
    # Don't launch by default when silent
    IfSilent label_no_launch
	MessageBox MB_YESNO $(InstSuccesssQuestion) \
        IDYES label_launch IDNO label_no_launch
        
label_launch:
	# Assumes SetOutPath $INSTDIR
	Exec '"$INSTDIR\$INSTEXE" $INSTFLAGS $SHORTCUT_LANG_PARAM'
label_no_launch:
	Pop $R0
FunctionEnd


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Make sure we're not on Windows 98 / ME
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function CheckWindowsVersion
	DetailPrint "Checking Windows version..."
	Call GetWindowsVersion
	Pop $R0
	; Just get first two characters, ignore 4.0 part of "NT 4.0"
	StrCpy $R0 $R0 2
	; Blacklist certain OS versions
	StrCmp $R0 "95" win_ver_bad
	StrCmp $R0 "98" win_ver_bad
	StrCmp $R0 "ME" win_ver_bad
	StrCmp $R0 "NT" win_ver_bad
	Return
win_ver_bad:
	MessageBox MB_YESNO $(CheckWindowsVersionMB) IDNO win_ver_abort
	Return
win_ver_abort:
	Quit
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Make sure the user can install/uninstall
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function CheckIfAdministrator
    DetailPrint $(CheckAdministratorInstDP)
    UserInfo::GetAccountType
    Pop $R0
    StrCmp $R0 "Admin" lbl_is_admin
        MessageBox MB_OK $(CheckAdministratorInstMB)
        Quit
lbl_is_admin:
    Return
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function un.CheckIfAdministrator
    DetailPrint $(CheckAdministratorUnInstDP)
    UserInfo::GetAccountType
    Pop $R0
    StrCmp $R0 "Admin" lbl_is_admin
        MessageBox MB_OK $(CheckAdministratorUnInstMB)
        Quit
lbl_is_admin:
    Return
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Checks to see if the current version has already been installed (according to the registry).
; If it has, allow user to bail out of install process.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function CheckIfAlreadyCurrent
  Push $0
	ReadRegStr $0 HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG" "Version"
    StrCmp $0 ${VERSION_LONG} 0 DONE
	MessageBox MB_OKCANCEL $(CheckIfCurrentMB) /SD IDOK IDOK DONE
    Quit

  DONE:
    Pop $0
    Return
FunctionEnd
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Close the program, if running. Modifies no variables.
; Allows user to bail out of install process.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function CloseSecondLife
  Push $0
  FindWindow $0 "Second Life" ""
  IntCmp $0 0 DONE
  MessageBox MB_OKCANCEL $(CloseSecondLifeInstMB) IDOK CLOSE IDCANCEL CANCEL_INSTALL

  CANCEL_INSTALL:
    Quit

  CLOSE:
    DetailPrint $(CloseSecondLifeInstDP)
    SendMessage $0 16 0 0

  LOOP:
	  FindWindow $0 "Second Life" ""
	  IntCmp $0 0 DONE
	  Sleep 500
	  Goto LOOP

  DONE:
    Pop $0
    Return
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Test our connection to secondlife.com
; Also allows us to count attempted installs by examining web logs.
; *TODO: Return current SL version info and have installer check
; if it is up to date.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function CheckNetworkConnection
    Push $0
    Push $1
    Push $2	# Option value for GetOptions
    DetailPrint $(CheckNetworkConnectionDP)
    ; Look for a tag value from the stub installer, used for statistics
    ; to correlate installs.  Default to "" if not found on command line.
    StrCpy $2 ""
    ${GetOptions} $COMMANDLINE "/STUBTAG=" $2
    GetTempFileName $0
    !define HTTP_TIMEOUT 5000 ; milliseconds
    ; Don't show secondary progress bar, this will be quick.
    NSISdl::download_quiet \
        /TIMEOUT=${HTTP_TIMEOUT} \
        "http://install.secondlife.com/check/?stubtag=$2&version=${VERSION_LONG}" \
        $0
    Pop $1 ; Return value, either "success", "cancel" or an error message
    ; MessageBox MB_OK "Download result: $1"
    ; Result ignored for now
	; StrCmp $1 "success" +2
	;	DetailPrint "Connection failed: $1"
    Delete $0 ; temporary file
    Pop $2
    Pop $1
    Pop $0
    Return
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Delete files in Documents and Settings\<user>\SecondLife\cache
; Delete files in Documents and Settings\All Users\SecondLife\cache
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Function RemoveCacheFiles
;
;; Delete files in Documents and Settings\<user>\SecondLife
;Push $0
;Push $1
;Push $2
;  DetailPrint $(RemoveCacheFilesDP)
;
;  StrCpy $0 0 ; Index number used to iterate via EnumRegKey
;
;  LOOP:
;    EnumRegKey $1 HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProfileList" $0
;    StrCmp $1 "" DONE               ; no more users
;
;    ReadRegStr $2 HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProfileList\$1" "ProfileImagePath" 
;    StrCmp $2 "" CONTINUE 0         ; "ProfileImagePath" value is missing
;
;    ; Required since ProfileImagePath is of type REG_EXPAND_SZ
;    ExpandEnvStrings $2 $2
;
;	; When explicitly uninstalling, everything goes away
;    RMDir /r "$2\Application Data\SecondLife\cache"
;
;  CONTINUE:
;    IntOp $0 $0 + 1
;    Goto LOOP
;  DONE:
;Pop $2
;Pop $1
;Pop $0
;
;; Delete files in Documents and Settings\All Users\SecondLife
;Push $0
;  ReadRegStr $0 HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders" "Common AppData"
;  StrCmp $0 "" +2
;  RMDir /r "$0\SecondLife\cache"
;Pop $0
;
;; Delete filse in C:\Windows\Application Data\SecondLife
;; If the user is running on a pre-NT system, Application Data lives here instead of
;; in Documents and Settings.
;RMDir /r "$WINDIR\Application Data\SecondLife\cache"
;
;FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Delete the installed shader files
;;; Since shaders are in active development, we'll likely need to shuffle them
;;; around a bit from build to build.  This ensures that shaders that we move
;;; or rename in the dev tree don't get left behind in the install.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function RemoveOldShaders

;; Remove old shader files first so fallbacks will work. see DEV-5663
RMDir /r "$INSTDIR\app_settings\shaders\*"

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Delete the installed XUI files
;;; We've changed the directory hierarchy for skins, putting all XUI and texture
;;; files under a specific skin directory, i.e. skins/default/xui/en-us as opposed
;;; to skins/xui/en-us.  Need to clean up the old path when upgrading
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function RemoveOldXUI

;; remove old XUI and texture files
RmDir /r "$INSTDIR\skins\html"
RmDir /r "$INSTDIR\skins\xui"
RmDir /r "$INSTDIR\skins\textures"
Delete "$INSTDIR\skins\*.txt"

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Remove any releasenotes files.
;;; We are no longer including release notes with the viewer. This will delete
;;; any that were left behind by an older installer. Delete will not fail if
;;; the files do not exist
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function RemoveOldReleaseNotes

;; remove releasenotes.txt file from application directory, and the shortcut
;; from the start menu.
Delete "$SMPROGRAMS\$INSTSHORTCUT\SL Release Notes.lnk"
Delete "$INSTDIR\releasenotes.txt"

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Delete a xui file that causes crash in Silver skin in cases where it was
;;; left behind by an older installer.
;;; See SNOW-348
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function RemoveOldAboutLandSilver

Delete "$INSTDIR\skins\silver\xui\en-us\floater_about_land.xml"

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Delete files in Documents and Settings\<user>\SecondLife
; Delete files in Documents and Settings\All Users\SecondLife
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function un.DocumentsAndSettingsFolder

; Delete files in Documents and Settings\<user>\SecondLife
Push $0
Push $1
Push $2

  DetailPrint "Deleting files in Documents and Settings folder"

  StrCpy $0 0 ; Index number used to iterate via EnumRegKey

  LOOP:
    EnumRegKey $1 HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProfileList" $0
    StrCmp $1 "" DONE               ; no more users

    ReadRegStr $2 HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProfileList\$1" "ProfileImagePath" 
    StrCmp $2 "" CONTINUE 0         ; "ProfileImagePath" value is missing

    ; Required since ProfileImagePath is of type REG_EXPAND_SZ
    ExpandEnvStrings $2 $2

	; If uninstalling a normal install remove everything
	; Otherwise (preview/dmz etc) just remove cache
    StrCmp $INSTFLAGS "" RM_ALL RM_CACHE
      RM_ALL:
        RMDir /r "$2\Application Data\SecondLife"
      RM_CACHE:
        # Local Settings directory is the cache, there is no "cache" subdir
        RMDir /r "$2\Local Settings\Application Data\SecondLife"
        # Vista version of the same
        RMDir /r "$2\AppData\Local\SecondLife"
        Delete "$2\Application Data\SecondLife\user_settings\settings_windlight.xml"

  CONTINUE:
    IntOp $0 $0 + 1
    Goto LOOP
  DONE:

Pop $2
Pop $1
Pop $0

; Delete files in Documents and Settings\All Users\SecondLife
Push $0
  ReadRegStr $0 HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders" "Common AppData"
  StrCmp $0 "" +2
  RMDir /r "$0\SecondLife"
Pop $0

; Delete filse in C:\Windows\Application Data\SecondLife
; If the user is running on a pre-NT system, Application Data lives here instead of
; in Documents and Settings.
RMDir /r "$WINDIR\Application Data\SecondLife"

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Close the program, if running. Modifies no variables.
; Allows user to bail out of uninstall process.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function un.CloseSecondLife
  Push $0
  FindWindow $0 "Second Life" ""
  IntCmp $0 0 DONE
  MessageBox MB_OKCANCEL $(CloseSecondLifeUnInstMB) IDOK CLOSE IDCANCEL CANCEL_UNINSTALL

  CANCEL_UNINSTALL:
    Quit

  CLOSE:
    DetailPrint $(CloseSecondLifeUnInstDP)
    SendMessage $0 16 0 0

  LOOP:
	  FindWindow $0 "Second Life" ""
	  IntCmp $0 0 DONE
	  Sleep 500
	  Goto LOOP

  DONE:
    Pop $0
    Return
FunctionEnd


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;   Delete the stored password for the current Windows user
;   DEV-10821 -- Unauthorised user can gain access to an SL account after a real user has uninstalled
;
Function un.RemovePassword

DetailPrint "Removing Second Life password"

SetShellVarContext current
Delete "$APPDATA\SecondLife\user_settings\password.dat"
SetShellVarContext all

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Delete the installed files
;;; This deletes the uninstall executable, but it works 
;;; because it is copied to temp directory before running
;;;
;;; Note:  You must list all files here, because we only
;;; want to delete our files, not things users left in the
;;; application directories.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function un.ProgramFiles

;; Remove mozilla file first so recursive directory deletion doesn't get hung up
Delete "$INSTDIR\app_settings\mozilla\components"

;; This placeholder is replaced by the complete list of files to uninstall by viewer_manifest.py
Delete $INSTDIR\skins\default\html\zh\loading-error\index.html
Delete $INSTDIR\skins\default\html\zh\loading-error\index.html
Delete $INSTDIR\skins\default\html\pt\loading-error\index.html
Delete $INSTDIR\skins\default\html\pt\loading-error\index.html
Delete $INSTDIR\skins\default\html\ko\loading-error\index.html
Delete $INSTDIR\skins\default\html\ko\loading-error\index.html
Delete $INSTDIR\skins\default\html\ja\loading-error\index.html
Delete $INSTDIR\skins\default\html\ja\loading-error\index.html
Delete $INSTDIR\skins\default\html\fr\loading-error\index.html
Delete $INSTDIR\skins\default\html\fr\loading-error\index.html
Delete $INSTDIR\skins\default\html\es\loading-error\index.html
Delete $INSTDIR\skins\default\html\es\loading-error\index.html
Delete $INSTDIR\skins\default\html\en-us\loading\sl_logo_rotate_black.gif
Delete $INSTDIR\skins\default\html\en-us\loading\sl_logo_rotate_black.gif
Delete $INSTDIR\skins\default\html\en-us\loading\loading.html
Delete $INSTDIR\skins\default\html\en-us\loading\loading.html
Delete $INSTDIR\skins\default\html\en-us\loading-error\index.html
Delete $INSTDIR\skins\default\html\en-us\loading-error\index.html
Delete $INSTDIR\skins\default\html\de\loading-error\index.html
Delete $INSTDIR\skins\default\html\de\loading-error\index.html
Delete $INSTDIR\skins\silver\xui\en-us\panel_voice_controls.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_voice_controls.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_toolbar.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_toolbar.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_status_bar.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_status_bar.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_media_controls.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_media_controls.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_avatar.xml
Delete $INSTDIR\skins\silver\xui\en-us\panel_avatar.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_windlight_options.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_windlight_options.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_water.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_water.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_tools.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_tools.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_script_queue.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_script_queue.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_script_ed_panel.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_script_ed_panel.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_live_lsleditor.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_live_lsleditor.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_env_settings.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_env_settings.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_directory.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_directory.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_chatterbox.xml
Delete $INSTDIR\skins\silver\xui\en-us\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\zh\xui_version.xml
Delete $INSTDIR\skins\default\xui\zh\xui_version.xml
Delete $INSTDIR\skins\default\xui\zh\ui_strings.xml
Delete $INSTDIR\skins\default\xui\zh\ui_strings.xml
Delete $INSTDIR\skins\default\xui\zh\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\zh\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\zh\strings.xml
Delete $INSTDIR\skins\default\xui\zh\strings.xml
Delete $INSTDIR\skins\default\xui\zh\role_actions.xml
Delete $INSTDIR\skins\default\xui\zh\role_actions.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\zh\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\zh\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\zh\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\zh\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\zh\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\zh\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\zh\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\zh\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\zh\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\zh\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\zh\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\zh\panel_place.xml
Delete $INSTDIR\skins\default\xui\zh\panel_place.xml
Delete $INSTDIR\skins\default\xui\zh\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\zh\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\zh\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\zh\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\zh\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\zh\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\zh\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\zh\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\zh\panel_login.xml
Delete $INSTDIR\skins\default\xui\zh\panel_login.xml
Delete $INSTDIR\skins\default\xui\zh\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\zh\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\zh\panel_groups.xml
Delete $INSTDIR\skins\default\xui\zh\panel_groups.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group.xml
Delete $INSTDIR\skins\default\xui\zh\panel_group.xml
Delete $INSTDIR\skins\default\xui\zh\panel_friends.xml
Delete $INSTDIR\skins\default\xui\zh\panel_friends.xml
Delete $INSTDIR\skins\default\xui\zh\panel_event.xml
Delete $INSTDIR\skins\default\xui\zh\panel_event.xml
Delete $INSTDIR\skins\default\xui\zh\panel_classified.xml
Delete $INSTDIR\skins\default\xui\zh\panel_classified.xml
Delete $INSTDIR\skins\default\xui\zh\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\zh\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\zh\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\zh\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\zh\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\zh\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\zh\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\zh\panel_audio.xml
Delete $INSTDIR\skins\default\xui\zh\panel_audio.xml
Delete $INSTDIR\skins\default\xui\zh\notify.xml
Delete $INSTDIR\skins\default\xui\zh\notify.xml
Delete $INSTDIR\skins\default\xui\zh\notifications.xml
Delete $INSTDIR\skins\default\xui\zh\notifications.xml
Delete $INSTDIR\skins\default\xui\zh\mime_types.xml
Delete $INSTDIR\skins\default\xui\zh\mime_types.xml
Delete $INSTDIR\skins\default\xui\zh\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\zh\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\zh\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\zh\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\zh\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\zh\menu_login.xml
Delete $INSTDIR\skins\default\xui\zh\menu_login.xml
Delete $INSTDIR\skins\default\xui\zh\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\zh\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\zh\fonts.xml
Delete $INSTDIR\skins\default\xui\zh\fonts.xml
Delete $INSTDIR\skins\default\xui\zh\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\zh\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\zh\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\zh\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\zh\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\zh\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\zh\floater_water.xml
Delete $INSTDIR\skins\default\xui\zh\floater_water.xml
Delete $INSTDIR\skins\default\xui\zh\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\zh\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\zh\floater_tos.xml
Delete $INSTDIR\skins\default\xui\zh\floater_tos.xml
Delete $INSTDIR\skins\default\xui\zh\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\zh\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\zh\floater_tools.xml
Delete $INSTDIR\skins\default\xui\zh\floater_tools.xml
Delete $INSTDIR\skins\default\xui\zh\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\zh\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\zh\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\zh\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\zh\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\zh\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\zh\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\zh\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\zh\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\zh\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\zh\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\zh\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\zh\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\zh\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\zh\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\zh\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\zh\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\zh\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\zh\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\zh\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\zh\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\zh\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\zh\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\zh\floater_profile.xml
Delete $INSTDIR\skins\default\xui\zh\floater_profile.xml
Delete $INSTDIR\skins\default\xui\zh\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\zh\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_new_landmark.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_new_landmark.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\zh\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\zh\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\zh\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\zh\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\zh\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\zh\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\zh\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\zh\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\zh\floater_pay.xml
Delete $INSTDIR\skins\default\xui\zh\floater_pay.xml
Delete $INSTDIR\skins\default\xui\zh\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\zh\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\zh\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\zh\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\zh\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\zh\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\zh\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\zh\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\zh\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\zh\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\zh\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\zh\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mute.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mute.xml
Delete $INSTDIR\skins\default\xui\zh\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\zh\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\zh\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\zh\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\zh\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\zh\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\zh\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\zh\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\zh\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\zh\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\zh\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\zh\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\zh\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\zh\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\zh\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\zh\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\zh\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\zh\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\zh\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\zh\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\zh\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\zh\floater_import.xml
Delete $INSTDIR\skins\default\xui\zh\floater_import.xml
Delete $INSTDIR\skins\default\xui\zh\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_im.xml
Delete $INSTDIR\skins\default\xui\zh\floater_im.xml
Delete $INSTDIR\skins\default\xui\zh\floater_hud.xml
Delete $INSTDIR\skins\default\xui\zh\floater_hud.xml
Delete $INSTDIR\skins\default\xui\zh\floater_html.xml
Delete $INSTDIR\skins\default\xui\zh\floater_html.xml
Delete $INSTDIR\skins\default\xui\zh\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\zh\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\zh\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\zh\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\zh\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\zh\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\zh\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\zh\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\zh\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_directory.xml
Delete $INSTDIR\skins\default\xui\zh\floater_directory.xml
Delete $INSTDIR\skins\default\xui\zh\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\zh\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\zh\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\zh\floater_customize.xml
Delete $INSTDIR\skins\default\xui\zh\floater_customize.xml
Delete $INSTDIR\skins\default\xui\zh\floater_critical.xml
Delete $INSTDIR\skins\default\xui\zh\floater_critical.xml
Delete $INSTDIR\skins\default\xui\zh\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\zh\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\zh\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\zh\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\zh\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\zh\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\zh\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\zh\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\zh\floater_camera.xml
Delete $INSTDIR\skins\default\xui\zh\floater_camera.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\zh\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\zh\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\zh\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\zh\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\zh\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\zh\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\zh\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\zh\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\zh\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\zh\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\zh\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\zh\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\zh\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\zh\floater_auction.xml
Delete $INSTDIR\skins\default\xui\zh\floater_auction.xml
Delete $INSTDIR\skins\default\xui\zh\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\zh\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\zh\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\zh\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\zh\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\zh\floater_about.xml
Delete $INSTDIR\skins\default\xui\zh\floater_about.xml
Delete $INSTDIR\skins\default\xui\zh\alerts.xml
Delete $INSTDIR\skins\default\xui\zh\alerts.xml
Delete $INSTDIR\skins\default\xui\uk\xui_version.xml
Delete $INSTDIR\skins\default\xui\uk\xui_version.xml
Delete $INSTDIR\skins\default\xui\uk\ui_strings.xml
Delete $INSTDIR\skins\default\xui\uk\ui_strings.xml
Delete $INSTDIR\skins\default\xui\uk\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\uk\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\uk\strings.xml
Delete $INSTDIR\skins\default\xui\uk\strings.xml
Delete $INSTDIR\skins\default\xui\uk\role_actions.xml
Delete $INSTDIR\skins\default\xui\uk\role_actions.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\uk\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\uk\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\uk\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\uk\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\uk\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\uk\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\uk\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\uk\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\uk\panel_progress.xml
Delete $INSTDIR\skins\default\xui\uk\panel_progress.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\uk\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\uk\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\uk\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\uk\panel_place.xml
Delete $INSTDIR\skins\default\xui\uk\panel_place.xml
Delete $INSTDIR\skins\default\xui\uk\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\uk\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\uk\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\uk\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\uk\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\uk\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\uk\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\uk\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\uk\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\uk\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\uk\panel_login.xml
Delete $INSTDIR\skins\default\xui\uk\panel_login.xml
Delete $INSTDIR\skins\default\xui\uk\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\uk\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\uk\panel_groups.xml
Delete $INSTDIR\skins\default\xui\uk\panel_groups.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group.xml
Delete $INSTDIR\skins\default\xui\uk\panel_group.xml
Delete $INSTDIR\skins\default\xui\uk\panel_friends.xml
Delete $INSTDIR\skins\default\xui\uk\panel_friends.xml
Delete $INSTDIR\skins\default\xui\uk\panel_event.xml
Delete $INSTDIR\skins\default\xui\uk\panel_event.xml
Delete $INSTDIR\skins\default\xui\uk\panel_classified.xml
Delete $INSTDIR\skins\default\xui\uk\panel_classified.xml
Delete $INSTDIR\skins\default\xui\uk\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\uk\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\uk\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\uk\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\uk\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\uk\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\uk\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\uk\panel_audio.xml
Delete $INSTDIR\skins\default\xui\uk\panel_audio.xml
Delete $INSTDIR\skins\default\xui\uk\notify.xml
Delete $INSTDIR\skins\default\xui\uk\notify.xml
Delete $INSTDIR\skins\default\xui\uk\notifications.xml
Delete $INSTDIR\skins\default\xui\uk\notifications.xml
Delete $INSTDIR\skins\default\xui\uk\mime_types.xml
Delete $INSTDIR\skins\default\xui\uk\mime_types.xml
Delete $INSTDIR\skins\default\xui\uk\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\uk\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\uk\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\uk\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\uk\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\uk\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\uk\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\uk\menu_login.xml
Delete $INSTDIR\skins\default\xui\uk\menu_login.xml
Delete $INSTDIR\skins\default\xui\uk\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\uk\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\uk\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\uk\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\uk\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\uk\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\uk\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\uk\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\uk\floater_water.xml
Delete $INSTDIR\skins\default\xui\uk\floater_water.xml
Delete $INSTDIR\skins\default\xui\uk\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\uk\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\uk\floater_tos.xml
Delete $INSTDIR\skins\default\xui\uk\floater_tos.xml
Delete $INSTDIR\skins\default\xui\uk\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\uk\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\uk\floater_tools.xml
Delete $INSTDIR\skins\default\xui\uk\floater_tools.xml
Delete $INSTDIR\skins\default\xui\uk\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\uk\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\uk\floater_test.xml
Delete $INSTDIR\skins\default\xui\uk\floater_test.xml
Delete $INSTDIR\skins\default\xui\uk\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\uk\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\uk\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\uk\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\uk\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\uk\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\uk\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\uk\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\uk\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\uk\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\uk\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\uk\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\uk\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\uk\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\uk\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\uk\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\uk\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\uk\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\uk\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\uk\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\uk\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\uk\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\uk\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\uk\floater_profile.xml
Delete $INSTDIR\skins\default\xui\uk\floater_profile.xml
Delete $INSTDIR\skins\default\xui\uk\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\uk\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\uk\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\uk\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\uk\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\uk\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\uk\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\uk\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\uk\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\uk\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\uk\floater_pay.xml
Delete $INSTDIR\skins\default\xui\uk\floater_pay.xml
Delete $INSTDIR\skins\default\xui\uk\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\uk\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\uk\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\uk\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\uk\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\uk\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\uk\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\uk\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\uk\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\uk\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\uk\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\uk\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\uk\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\uk\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\uk\floater_mute.xml
Delete $INSTDIR\skins\default\xui\uk\floater_mute.xml
Delete $INSTDIR\skins\default\xui\uk\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\uk\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\uk\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\uk\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\uk\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\uk\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\uk\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\uk\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\uk\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\uk\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\uk\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\uk\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\uk\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\uk\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\uk\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\uk\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\uk\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\uk\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\uk\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\uk\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\uk\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\uk\floater_import.xml
Delete $INSTDIR\skins\default\xui\uk\floater_import.xml
Delete $INSTDIR\skins\default\xui\uk\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_im.xml
Delete $INSTDIR\skins\default\xui\uk\floater_im.xml
Delete $INSTDIR\skins\default\xui\uk\floater_hud.xml
Delete $INSTDIR\skins\default\xui\uk\floater_hud.xml
Delete $INSTDIR\skins\default\xui\uk\floater_html.xml
Delete $INSTDIR\skins\default\xui\uk\floater_html.xml
Delete $INSTDIR\skins\default\xui\uk\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\uk\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\uk\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\uk\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\uk\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\uk\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\uk\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\uk\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\uk\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_directory.xml
Delete $INSTDIR\skins\default\xui\uk\floater_directory.xml
Delete $INSTDIR\skins\default\xui\uk\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\uk\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\uk\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\uk\floater_customize.xml
Delete $INSTDIR\skins\default\xui\uk\floater_customize.xml
Delete $INSTDIR\skins\default\xui\uk\floater_critical.xml
Delete $INSTDIR\skins\default\xui\uk\floater_critical.xml
Delete $INSTDIR\skins\default\xui\uk\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\uk\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\uk\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\uk\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\uk\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\uk\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\uk\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\uk\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\uk\floater_camera.xml
Delete $INSTDIR\skins\default\xui\uk\floater_camera.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\uk\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\uk\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\uk\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\uk\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\uk\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\uk\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\uk\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\uk\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\uk\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\uk\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\uk\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\uk\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\uk\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\uk\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\uk\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\uk\floater_auction.xml
Delete $INSTDIR\skins\default\xui\uk\floater_auction.xml
Delete $INSTDIR\skins\default\xui\uk\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\uk\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\uk\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\uk\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\uk\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\uk\floater_about.xml
Delete $INSTDIR\skins\default\xui\uk\floater_about.xml
Delete $INSTDIR\skins\default\xui\uk\alerts.xml
Delete $INSTDIR\skins\default\xui\uk\alerts.xml
Delete $INSTDIR\skins\default\xui\tr\xui_version.xml
Delete $INSTDIR\skins\default\xui\tr\xui_version.xml
Delete $INSTDIR\skins\default\xui\tr\ui_strings.xml
Delete $INSTDIR\skins\default\xui\tr\ui_strings.xml
Delete $INSTDIR\skins\default\xui\tr\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\tr\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\tr\strings.xml
Delete $INSTDIR\skins\default\xui\tr\strings.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\tr\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\tr\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\tr\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\tr\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\tr\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\tr\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\tr\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\tr\panel_place.xml
Delete $INSTDIR\skins\default\xui\tr\panel_place.xml
Delete $INSTDIR\skins\default\xui\tr\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\tr\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\tr\panel_login.xml
Delete $INSTDIR\skins\default\xui\tr\panel_login.xml
Delete $INSTDIR\skins\default\xui\tr\panel_groups.xml
Delete $INSTDIR\skins\default\xui\tr\panel_groups.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group.xml
Delete $INSTDIR\skins\default\xui\tr\panel_group.xml
Delete $INSTDIR\skins\default\xui\tr\panel_friends.xml
Delete $INSTDIR\skins\default\xui\tr\panel_friends.xml
Delete $INSTDIR\skins\default\xui\tr\panel_event.xml
Delete $INSTDIR\skins\default\xui\tr\panel_event.xml
Delete $INSTDIR\skins\default\xui\tr\panel_classified.xml
Delete $INSTDIR\skins\default\xui\tr\panel_classified.xml
Delete $INSTDIR\skins\default\xui\tr\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\tr\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\tr\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\tr\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\tr\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\tr\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\tr\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\tr\panel_audio.xml
Delete $INSTDIR\skins\default\xui\tr\panel_audio.xml
Delete $INSTDIR\skins\default\xui\tr\notify.xml
Delete $INSTDIR\skins\default\xui\tr\notify.xml
Delete $INSTDIR\skins\default\xui\tr\notifications.xml
Delete $INSTDIR\skins\default\xui\tr\notifications.xml
Delete $INSTDIR\skins\default\xui\tr\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\tr\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\tr\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\tr\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\tr\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\tr\menu_login.xml
Delete $INSTDIR\skins\default\xui\tr\menu_login.xml
Delete $INSTDIR\skins\default\xui\tr\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\tr\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\tr\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\tr\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\tr\floater_tos.xml
Delete $INSTDIR\skins\default\xui\tr\floater_tos.xml
Delete $INSTDIR\skins\default\xui\tr\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\tr\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\tr\floater_tools.xml
Delete $INSTDIR\skins\default\xui\tr\floater_tools.xml
Delete $INSTDIR\skins\default\xui\tr\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\tr\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\tr\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\tr\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\tr\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\tr\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\tr\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\tr\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\tr\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\tr\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\tr\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\tr\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\tr\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\tr\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\tr\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\tr\floater_profile.xml
Delete $INSTDIR\skins\default\xui\tr\floater_profile.xml
Delete $INSTDIR\skins\default\xui\tr\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\tr\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\tr\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\tr\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\tr\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\tr\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\tr\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\tr\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\tr\floater_pay.xml
Delete $INSTDIR\skins\default\xui\tr\floater_pay.xml
Delete $INSTDIR\skins\default\xui\tr\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\tr\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\tr\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\tr\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\tr\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\tr\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\tr\floater_mute.xml
Delete $INSTDIR\skins\default\xui\tr\floater_mute.xml
Delete $INSTDIR\skins\default\xui\tr\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\tr\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\tr\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\tr\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\tr\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\tr\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\tr\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\tr\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\tr\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\tr\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\tr\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\tr\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\tr\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\tr\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\tr\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\tr\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\tr\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\tr\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_hud.xml
Delete $INSTDIR\skins\default\xui\tr\floater_hud.xml
Delete $INSTDIR\skins\default\xui\tr\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\tr\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\tr\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\tr\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\tr\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\tr\floater_directory.xml
Delete $INSTDIR\skins\default\xui\tr\floater_directory.xml
Delete $INSTDIR\skins\default\xui\tr\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\tr\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\tr\floater_customize.xml
Delete $INSTDIR\skins\default\xui\tr\floater_customize.xml
Delete $INSTDIR\skins\default\xui\tr\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\tr\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\tr\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\tr\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\tr\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\tr\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\tr\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\tr\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\tr\floater_camera.xml
Delete $INSTDIR\skins\default\xui\tr\floater_camera.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\tr\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\tr\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\tr\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\tr\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\tr\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\tr\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\tr\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\tr\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\tr\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\tr\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\tr\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\tr\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\tr\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\tr\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\tr\floater_about.xml
Delete $INSTDIR\skins\default\xui\tr\floater_about.xml
Delete $INSTDIR\skins\default\xui\tr\alerts.xml
Delete $INSTDIR\skins\default\xui\tr\alerts.xml
Delete $INSTDIR\skins\default\xui\ru\xui_version.xml
Delete $INSTDIR\skins\default\xui\ru\xui_version.xml
Delete $INSTDIR\skins\default\xui\ru\ui_strings.xml
Delete $INSTDIR\skins\default\xui\ru\ui_strings.xml
Delete $INSTDIR\skins\default\xui\ru\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\ru\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\ru\strings.xml
Delete $INSTDIR\skins\default\xui\ru\strings.xml
Delete $INSTDIR\skins\default\xui\ru\role_actions.xml
Delete $INSTDIR\skins\default\xui\ru\role_actions.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\ru\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\ru\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\ru\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\ru\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\ru\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\ru\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\ru\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\ru\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\ru\panel_progress.xml
Delete $INSTDIR\skins\default\xui\ru\panel_progress.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\ru\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\ru\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\ru\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\ru\panel_place.xml
Delete $INSTDIR\skins\default\xui\ru\panel_place.xml
Delete $INSTDIR\skins\default\xui\ru\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\ru\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\ru\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ru\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ru\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\ru\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\ru\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\ru\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\ru\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\ru\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\ru\panel_login.xml
Delete $INSTDIR\skins\default\xui\ru\panel_login.xml
Delete $INSTDIR\skins\default\xui\ru\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\ru\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\ru\panel_groups.xml
Delete $INSTDIR\skins\default\xui\ru\panel_groups.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group.xml
Delete $INSTDIR\skins\default\xui\ru\panel_group.xml
Delete $INSTDIR\skins\default\xui\ru\panel_friends.xml
Delete $INSTDIR\skins\default\xui\ru\panel_friends.xml
Delete $INSTDIR\skins\default\xui\ru\panel_event.xml
Delete $INSTDIR\skins\default\xui\ru\panel_event.xml
Delete $INSTDIR\skins\default\xui\ru\panel_classified.xml
Delete $INSTDIR\skins\default\xui\ru\panel_classified.xml
Delete $INSTDIR\skins\default\xui\ru\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\ru\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\ru\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\ru\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\ru\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\ru\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\ru\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\ru\panel_audio.xml
Delete $INSTDIR\skins\default\xui\ru\panel_audio.xml
Delete $INSTDIR\skins\default\xui\ru\notify.xml
Delete $INSTDIR\skins\default\xui\ru\notify.xml
Delete $INSTDIR\skins\default\xui\ru\notifications.xml
Delete $INSTDIR\skins\default\xui\ru\notifications.xml
Delete $INSTDIR\skins\default\xui\ru\mime_types.xml
Delete $INSTDIR\skins\default\xui\ru\mime_types.xml
Delete $INSTDIR\skins\default\xui\ru\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\ru\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\ru\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\ru\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\ru\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\ru\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\ru\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\ru\menu_login.xml
Delete $INSTDIR\skins\default\xui\ru\menu_login.xml
Delete $INSTDIR\skins\default\xui\ru\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\ru\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\ru\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\ru\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\ru\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\ru\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\ru\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\ru\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\ru\floater_water.xml
Delete $INSTDIR\skins\default\xui\ru\floater_water.xml
Delete $INSTDIR\skins\default\xui\ru\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\ru\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\ru\floater_tos.xml
Delete $INSTDIR\skins\default\xui\ru\floater_tos.xml
Delete $INSTDIR\skins\default\xui\ru\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\ru\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\ru\floater_tools.xml
Delete $INSTDIR\skins\default\xui\ru\floater_tools.xml
Delete $INSTDIR\skins\default\xui\ru\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\ru\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\ru\floater_test.xml
Delete $INSTDIR\skins\default\xui\ru\floater_test.xml
Delete $INSTDIR\skins\default\xui\ru\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\ru\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\ru\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\ru\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\ru\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\ru\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\ru\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\ru\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\ru\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\ru\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\ru\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\ru\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\ru\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\ru\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\ru\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\ru\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\ru\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\ru\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\ru\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\ru\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\ru\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\ru\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\ru\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\ru\floater_profile.xml
Delete $INSTDIR\skins\default\xui\ru\floater_profile.xml
Delete $INSTDIR\skins\default\xui\ru\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\ru\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\ru\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\ru\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\ru\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\ru\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\ru\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\ru\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\ru\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\ru\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\ru\floater_pay.xml
Delete $INSTDIR\skins\default\xui\ru\floater_pay.xml
Delete $INSTDIR\skins\default\xui\ru\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\ru\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\ru\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\ru\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\ru\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\ru\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\ru\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\ru\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\ru\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\ru\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\ru\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\ru\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\ru\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\ru\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\ru\floater_mute.xml
Delete $INSTDIR\skins\default\xui\ru\floater_mute.xml
Delete $INSTDIR\skins\default\xui\ru\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\ru\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\ru\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\ru\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\ru\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\ru\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\ru\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\ru\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\ru\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\ru\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\ru\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\ru\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\ru\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\ru\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\ru\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\ru\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\ru\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\ru\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\ru\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\ru\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\ru\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\ru\floater_import.xml
Delete $INSTDIR\skins\default\xui\ru\floater_import.xml
Delete $INSTDIR\skins\default\xui\ru\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_im.xml
Delete $INSTDIR\skins\default\xui\ru\floater_im.xml
Delete $INSTDIR\skins\default\xui\ru\floater_hud.xml
Delete $INSTDIR\skins\default\xui\ru\floater_hud.xml
Delete $INSTDIR\skins\default\xui\ru\floater_html.xml
Delete $INSTDIR\skins\default\xui\ru\floater_html.xml
Delete $INSTDIR\skins\default\xui\ru\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\ru\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\ru\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\ru\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\ru\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\ru\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\ru\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\ru\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\ru\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_directory.xml
Delete $INSTDIR\skins\default\xui\ru\floater_directory.xml
Delete $INSTDIR\skins\default\xui\ru\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\ru\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\ru\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\ru\floater_customize.xml
Delete $INSTDIR\skins\default\xui\ru\floater_customize.xml
Delete $INSTDIR\skins\default\xui\ru\floater_critical.xml
Delete $INSTDIR\skins\default\xui\ru\floater_critical.xml
Delete $INSTDIR\skins\default\xui\ru\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\ru\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\ru\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\ru\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\ru\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\ru\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\ru\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\ru\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\ru\floater_camera.xml
Delete $INSTDIR\skins\default\xui\ru\floater_camera.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\ru\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\ru\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\ru\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\ru\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\ru\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\ru\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\ru\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\ru\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\ru\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\ru\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\ru\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\ru\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\ru\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\ru\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\ru\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\ru\floater_auction.xml
Delete $INSTDIR\skins\default\xui\ru\floater_auction.xml
Delete $INSTDIR\skins\default\xui\ru\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\ru\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\ru\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\ru\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\ru\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\ru\floater_about.xml
Delete $INSTDIR\skins\default\xui\ru\floater_about.xml
Delete $INSTDIR\skins\default\xui\ru\alerts.xml
Delete $INSTDIR\skins\default\xui\ru\alerts.xml
Delete $INSTDIR\skins\default\xui\pt\xui_version.xml
Delete $INSTDIR\skins\default\xui\pt\xui_version.xml
Delete $INSTDIR\skins\default\xui\pt\ui_strings.xml
Delete $INSTDIR\skins\default\xui\pt\ui_strings.xml
Delete $INSTDIR\skins\default\xui\pt\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\pt\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\pt\strings.xml
Delete $INSTDIR\skins\default\xui\pt\strings.xml
Delete $INSTDIR\skins\default\xui\pt\role_actions.xml
Delete $INSTDIR\skins\default\xui\pt\role_actions.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\pt\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\pt\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\pt\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\pt\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\pt\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\pt\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\pt\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\pt\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\pt\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\pt\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\pt\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\pt\panel_place.xml
Delete $INSTDIR\skins\default\xui\pt\panel_place.xml
Delete $INSTDIR\skins\default\xui\pt\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\pt\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\pt\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pt\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pt\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\pt\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\pt\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\pt\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\pt\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\pt\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\pt\panel_login.xml
Delete $INSTDIR\skins\default\xui\pt\panel_login.xml
Delete $INSTDIR\skins\default\xui\pt\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\pt\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\pt\panel_groups.xml
Delete $INSTDIR\skins\default\xui\pt\panel_groups.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group.xml
Delete $INSTDIR\skins\default\xui\pt\panel_group.xml
Delete $INSTDIR\skins\default\xui\pt\panel_friends.xml
Delete $INSTDIR\skins\default\xui\pt\panel_friends.xml
Delete $INSTDIR\skins\default\xui\pt\panel_event.xml
Delete $INSTDIR\skins\default\xui\pt\panel_event.xml
Delete $INSTDIR\skins\default\xui\pt\panel_classified.xml
Delete $INSTDIR\skins\default\xui\pt\panel_classified.xml
Delete $INSTDIR\skins\default\xui\pt\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\pt\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\pt\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\pt\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\pt\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\pt\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\pt\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\pt\panel_audio.xml
Delete $INSTDIR\skins\default\xui\pt\panel_audio.xml
Delete $INSTDIR\skins\default\xui\pt\notify.xml
Delete $INSTDIR\skins\default\xui\pt\notify.xml
Delete $INSTDIR\skins\default\xui\pt\notifications.xml
Delete $INSTDIR\skins\default\xui\pt\notifications.xml
Delete $INSTDIR\skins\default\xui\pt\mime_types.xml
Delete $INSTDIR\skins\default\xui\pt\mime_types.xml
Delete $INSTDIR\skins\default\xui\pt\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\pt\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\pt\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\pt\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\pt\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\pt\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\pt\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\pt\menu_login.xml
Delete $INSTDIR\skins\default\xui\pt\menu_login.xml
Delete $INSTDIR\skins\default\xui\pt\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\pt\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\pt\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\pt\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\pt\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\pt\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\pt\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\pt\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\pt\floater_water.xml
Delete $INSTDIR\skins\default\xui\pt\floater_water.xml
Delete $INSTDIR\skins\default\xui\pt\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\pt\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\pt\floater_tos.xml
Delete $INSTDIR\skins\default\xui\pt\floater_tos.xml
Delete $INSTDIR\skins\default\xui\pt\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\pt\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\pt\floater_tools.xml
Delete $INSTDIR\skins\default\xui\pt\floater_tools.xml
Delete $INSTDIR\skins\default\xui\pt\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\pt\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\pt\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\pt\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\pt\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\pt\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\pt\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\pt\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\pt\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\pt\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\pt\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\pt\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\pt\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\pt\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\pt\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\pt\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\pt\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\pt\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\pt\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\pt\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\pt\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\pt\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\pt\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\pt\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\pt\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\pt\floater_profile.xml
Delete $INSTDIR\skins\default\xui\pt\floater_profile.xml
Delete $INSTDIR\skins\default\xui\pt\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\pt\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\pt\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\pt\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\pt\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\pt\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\pt\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\pt\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\pt\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\pt\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\pt\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\pt\floater_pay.xml
Delete $INSTDIR\skins\default\xui\pt\floater_pay.xml
Delete $INSTDIR\skins\default\xui\pt\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\pt\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\pt\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\pt\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\pt\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\pt\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\pt\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\pt\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\pt\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\pt\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\pt\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\pt\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mute.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mute.xml
Delete $INSTDIR\skins\default\xui\pt\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\pt\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\pt\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\pt\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\pt\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\pt\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\pt\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\pt\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\pt\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\pt\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\pt\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\pt\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\pt\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\pt\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\pt\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\pt\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\pt\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\pt\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\pt\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\pt\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\pt\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\pt\floater_import.xml
Delete $INSTDIR\skins\default\xui\pt\floater_import.xml
Delete $INSTDIR\skins\default\xui\pt\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_im.xml
Delete $INSTDIR\skins\default\xui\pt\floater_im.xml
Delete $INSTDIR\skins\default\xui\pt\floater_hud.xml
Delete $INSTDIR\skins\default\xui\pt\floater_hud.xml
Delete $INSTDIR\skins\default\xui\pt\floater_html.xml
Delete $INSTDIR\skins\default\xui\pt\floater_html.xml
Delete $INSTDIR\skins\default\xui\pt\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\pt\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\pt\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\pt\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\pt\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\pt\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\pt\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\pt\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\pt\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_directory.xml
Delete $INSTDIR\skins\default\xui\pt\floater_directory.xml
Delete $INSTDIR\skins\default\xui\pt\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\pt\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\pt\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\pt\floater_customize.xml
Delete $INSTDIR\skins\default\xui\pt\floater_customize.xml
Delete $INSTDIR\skins\default\xui\pt\floater_critical.xml
Delete $INSTDIR\skins\default\xui\pt\floater_critical.xml
Delete $INSTDIR\skins\default\xui\pt\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\pt\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\pt\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\pt\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\pt\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\pt\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\pt\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\pt\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\pt\floater_camera.xml
Delete $INSTDIR\skins\default\xui\pt\floater_camera.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\pt\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\pt\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\pt\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\pt\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\pt\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\pt\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\pt\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\pt\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\pt\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\pt\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\pt\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\pt\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\pt\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\pt\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\pt\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\pt\floater_auction.xml
Delete $INSTDIR\skins\default\xui\pt\floater_auction.xml
Delete $INSTDIR\skins\default\xui\pt\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\pt\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\pt\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\pt\floater_account_history.xml
Delete $INSTDIR\skins\default\xui\pt\floater_account_history.xml
Delete $INSTDIR\skins\default\xui\pt\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\pt\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\pt\floater_about.xml
Delete $INSTDIR\skins\default\xui\pt\floater_about.xml
Delete $INSTDIR\skins\default\xui\pt\alerts.xml
Delete $INSTDIR\skins\default\xui\pt\alerts.xml
Delete $INSTDIR\skins\default\xui\pl\xui_version.xml
Delete $INSTDIR\skins\default\xui\pl\xui_version.xml
Delete $INSTDIR\skins\default\xui\pl\ui_strings.xml
Delete $INSTDIR\skins\default\xui\pl\ui_strings.xml
Delete $INSTDIR\skins\default\xui\pl\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\pl\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\pl\strings.xml
Delete $INSTDIR\skins\default\xui\pl\strings.xml
Delete $INSTDIR\skins\default\xui\pl\role_actions.xml
Delete $INSTDIR\skins\default\xui\pl\role_actions.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\pl\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\pl\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\pl\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\pl\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\pl\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\pl\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\pl\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\pl\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\pl\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\pl\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\pl\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\pl\panel_place.xml
Delete $INSTDIR\skins\default\xui\pl\panel_place.xml
Delete $INSTDIR\skins\default\xui\pl\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\pl\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\pl\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pl\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\pl\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\pl\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\pl\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\pl\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\pl\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\pl\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\pl\panel_login.xml
Delete $INSTDIR\skins\default\xui\pl\panel_login.xml
Delete $INSTDIR\skins\default\xui\pl\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\pl\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\pl\panel_groups.xml
Delete $INSTDIR\skins\default\xui\pl\panel_groups.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group.xml
Delete $INSTDIR\skins\default\xui\pl\panel_group.xml
Delete $INSTDIR\skins\default\xui\pl\panel_friends.xml
Delete $INSTDIR\skins\default\xui\pl\panel_friends.xml
Delete $INSTDIR\skins\default\xui\pl\panel_event.xml
Delete $INSTDIR\skins\default\xui\pl\panel_event.xml
Delete $INSTDIR\skins\default\xui\pl\panel_classified.xml
Delete $INSTDIR\skins\default\xui\pl\panel_classified.xml
Delete $INSTDIR\skins\default\xui\pl\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\pl\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\pl\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\pl\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\pl\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\pl\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\pl\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\pl\panel_audio.xml
Delete $INSTDIR\skins\default\xui\pl\panel_audio.xml
Delete $INSTDIR\skins\default\xui\pl\notify.xml
Delete $INSTDIR\skins\default\xui\pl\notify.xml
Delete $INSTDIR\skins\default\xui\pl\notifications.xml
Delete $INSTDIR\skins\default\xui\pl\notifications.xml
Delete $INSTDIR\skins\default\xui\pl\mime_types.xml
Delete $INSTDIR\skins\default\xui\pl\mime_types.xml
Delete $INSTDIR\skins\default\xui\pl\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\pl\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\pl\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\pl\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\pl\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\pl\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\pl\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\pl\menu_login.xml
Delete $INSTDIR\skins\default\xui\pl\menu_login.xml
Delete $INSTDIR\skins\default\xui\pl\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\pl\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\pl\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\pl\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\pl\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\pl\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\pl\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\pl\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\pl\floater_water.xml
Delete $INSTDIR\skins\default\xui\pl\floater_water.xml
Delete $INSTDIR\skins\default\xui\pl\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\pl\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\pl\floater_tos.xml
Delete $INSTDIR\skins\default\xui\pl\floater_tos.xml
Delete $INSTDIR\skins\default\xui\pl\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\pl\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\pl\floater_tools.xml
Delete $INSTDIR\skins\default\xui\pl\floater_tools.xml
Delete $INSTDIR\skins\default\xui\pl\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\pl\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\pl\floater_test.xml
Delete $INSTDIR\skins\default\xui\pl\floater_test.xml
Delete $INSTDIR\skins\default\xui\pl\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\pl\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\pl\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\pl\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\pl\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\pl\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\pl\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\pl\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\pl\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\pl\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\pl\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\pl\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\pl\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\pl\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\pl\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\pl\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\pl\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\pl\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\pl\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\pl\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\pl\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\pl\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\pl\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\pl\floater_profile.xml
Delete $INSTDIR\skins\default\xui\pl\floater_profile.xml
Delete $INSTDIR\skins\default\xui\pl\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\pl\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\pl\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\pl\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\pl\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\pl\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\pl\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\pl\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\pl\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\pl\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\pl\floater_pay.xml
Delete $INSTDIR\skins\default\xui\pl\floater_pay.xml
Delete $INSTDIR\skins\default\xui\pl\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\pl\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\pl\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\pl\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\pl\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\pl\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\pl\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\pl\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\pl\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\pl\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\pl\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\pl\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mute.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mute.xml
Delete $INSTDIR\skins\default\xui\pl\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\pl\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\pl\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\pl\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\pl\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\pl\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\pl\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\pl\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\pl\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\pl\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\pl\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\pl\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\pl\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\pl\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\pl\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\pl\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\pl\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\pl\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\pl\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\pl\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\pl\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\pl\floater_import.xml
Delete $INSTDIR\skins\default\xui\pl\floater_import.xml
Delete $INSTDIR\skins\default\xui\pl\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_im.xml
Delete $INSTDIR\skins\default\xui\pl\floater_im.xml
Delete $INSTDIR\skins\default\xui\pl\floater_hud.xml
Delete $INSTDIR\skins\default\xui\pl\floater_hud.xml
Delete $INSTDIR\skins\default\xui\pl\floater_html.xml
Delete $INSTDIR\skins\default\xui\pl\floater_html.xml
Delete $INSTDIR\skins\default\xui\pl\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\pl\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\pl\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\pl\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\pl\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\pl\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\pl\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\pl\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\pl\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_directory.xml
Delete $INSTDIR\skins\default\xui\pl\floater_directory.xml
Delete $INSTDIR\skins\default\xui\pl\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\pl\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\pl\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\pl\floater_customize.xml
Delete $INSTDIR\skins\default\xui\pl\floater_customize.xml
Delete $INSTDIR\skins\default\xui\pl\floater_critical.xml
Delete $INSTDIR\skins\default\xui\pl\floater_critical.xml
Delete $INSTDIR\skins\default\xui\pl\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\pl\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\pl\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\pl\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\pl\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\pl\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\pl\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\pl\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\pl\floater_camera.xml
Delete $INSTDIR\skins\default\xui\pl\floater_camera.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\pl\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\pl\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\pl\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\pl\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\pl\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\pl\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\pl\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\pl\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\pl\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\pl\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\pl\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\pl\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\pl\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\pl\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\pl\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\pl\floater_auction.xml
Delete $INSTDIR\skins\default\xui\pl\floater_auction.xml
Delete $INSTDIR\skins\default\xui\pl\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\pl\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\pl\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\pl\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\pl\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\pl\floater_about.xml
Delete $INSTDIR\skins\default\xui\pl\floater_about.xml
Delete $INSTDIR\skins\default\xui\pl\alerts.xml
Delete $INSTDIR\skins\default\xui\pl\alerts.xml
Delete $INSTDIR\skins\default\xui\ko\xui_version.xml
Delete $INSTDIR\skins\default\xui\ko\xui_version.xml
Delete $INSTDIR\skins\default\xui\ko\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\ko\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\ko\strings.xml
Delete $INSTDIR\skins\default\xui\ko\strings.xml
Delete $INSTDIR\skins\default\xui\ko\role_actions.xml
Delete $INSTDIR\skins\default\xui\ko\role_actions.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\ko\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\ko\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\ko\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\ko\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\ko\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\ko\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\ko\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\ko\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_LCD.xml
Delete $INSTDIR\skins\default\xui\ko\panel_preferences_LCD.xml
Delete $INSTDIR\skins\default\xui\ko\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\ko\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\ko\panel_place.xml
Delete $INSTDIR\skins\default\xui\ko\panel_place.xml
Delete $INSTDIR\skins\default\xui\ko\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ko\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ko\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\ko\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\ko\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\ko\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\ko\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\ko\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\ko\panel_login.xml
Delete $INSTDIR\skins\default\xui\ko\panel_login.xml
Delete $INSTDIR\skins\default\xui\ko\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\ko\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\ko\panel_groups.xml
Delete $INSTDIR\skins\default\xui\ko\panel_groups.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group.xml
Delete $INSTDIR\skins\default\xui\ko\panel_group.xml
Delete $INSTDIR\skins\default\xui\ko\panel_friends.xml
Delete $INSTDIR\skins\default\xui\ko\panel_friends.xml
Delete $INSTDIR\skins\default\xui\ko\panel_event.xml
Delete $INSTDIR\skins\default\xui\ko\panel_event.xml
Delete $INSTDIR\skins\default\xui\ko\panel_classified.xml
Delete $INSTDIR\skins\default\xui\ko\panel_classified.xml
Delete $INSTDIR\skins\default\xui\ko\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\ko\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\ko\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\ko\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\ko\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\ko\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\ko\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\ko\panel_audio.xml
Delete $INSTDIR\skins\default\xui\ko\panel_audio.xml
Delete $INSTDIR\skins\default\xui\ko\notify.xml
Delete $INSTDIR\skins\default\xui\ko\notify.xml
Delete $INSTDIR\skins\default\xui\ko\notifications.xml
Delete $INSTDIR\skins\default\xui\ko\notifications.xml
Delete $INSTDIR\skins\default\xui\ko\mime_types.xml
Delete $INSTDIR\skins\default\xui\ko\mime_types.xml
Delete $INSTDIR\skins\default\xui\ko\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\ko\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\ko\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\ko\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\ko\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\ko\menu_login.xml
Delete $INSTDIR\skins\default\xui\ko\menu_login.xml
Delete $INSTDIR\skins\default\xui\ko\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\ko\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\ko\fonts.xml
Delete $INSTDIR\skins\default\xui\ko\fonts.xml
Delete $INSTDIR\skins\default\xui\ko\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\ko\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\ko\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\ko\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\ko\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\ko\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\ko\floater_water.xml
Delete $INSTDIR\skins\default\xui\ko\floater_water.xml
Delete $INSTDIR\skins\default\xui\ko\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\ko\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\ko\floater_tos.xml
Delete $INSTDIR\skins\default\xui\ko\floater_tos.xml
Delete $INSTDIR\skins\default\xui\ko\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\ko\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\ko\floater_tools.xml
Delete $INSTDIR\skins\default\xui\ko\floater_tools.xml
Delete $INSTDIR\skins\default\xui\ko\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\ko\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\ko\floater_test.xml
Delete $INSTDIR\skins\default\xui\ko\floater_test.xml
Delete $INSTDIR\skins\default\xui\ko\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\ko\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\ko\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\ko\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\ko\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\ko\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\ko\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\ko\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\ko\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\ko\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\ko\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\ko\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\ko\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\ko\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\ko\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\ko\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\ko\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\ko\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\ko\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\ko\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\ko\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\ko\floater_profile.xml
Delete $INSTDIR\skins\default\xui\ko\floater_profile.xml
Delete $INSTDIR\skins\default\xui\ko\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\ko\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\ko\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\ko\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\ko\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\ko\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\ko\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\ko\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\ko\floater_pay.xml
Delete $INSTDIR\skins\default\xui\ko\floater_pay.xml
Delete $INSTDIR\skins\default\xui\ko\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\ko\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\ko\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\ko\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\ko\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\ko\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\ko\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\ko\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\ko\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\ko\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\ko\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\ko\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\ko\floater_mute.xml
Delete $INSTDIR\skins\default\xui\ko\floater_mute.xml
Delete $INSTDIR\skins\default\xui\ko\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\ko\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\ko\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\ko\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\ko\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\ko\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\ko\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\ko\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\ko\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\ko\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\ko\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\ko\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\ko\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\ko\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\ko\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\ko\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\ko\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\ko\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\ko\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\ko\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\ko\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\ko\floater_import.xml
Delete $INSTDIR\skins\default\xui\ko\floater_import.xml
Delete $INSTDIR\skins\default\xui\ko\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_im.xml
Delete $INSTDIR\skins\default\xui\ko\floater_im.xml
Delete $INSTDIR\skins\default\xui\ko\floater_hud.xml
Delete $INSTDIR\skins\default\xui\ko\floater_hud.xml
Delete $INSTDIR\skins\default\xui\ko\floater_html.xml
Delete $INSTDIR\skins\default\xui\ko\floater_html.xml
Delete $INSTDIR\skins\default\xui\ko\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\ko\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\ko\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\ko\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\ko\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\ko\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\ko\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_directory.xml
Delete $INSTDIR\skins\default\xui\ko\floater_directory.xml
Delete $INSTDIR\skins\default\xui\ko\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\ko\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\ko\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\ko\floater_customize.xml
Delete $INSTDIR\skins\default\xui\ko\floater_customize.xml
Delete $INSTDIR\skins\default\xui\ko\floater_critical.xml
Delete $INSTDIR\skins\default\xui\ko\floater_critical.xml
Delete $INSTDIR\skins\default\xui\ko\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\ko\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\ko\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\ko\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\ko\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\ko\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\ko\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\ko\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\ko\floater_camera.xml
Delete $INSTDIR\skins\default\xui\ko\floater_camera.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\ko\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\ko\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\ko\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\ko\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\ko\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\ko\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\ko\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\ko\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\ko\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\ko\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\ko\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\ko\floater_auction.xml
Delete $INSTDIR\skins\default\xui\ko\floater_auction.xml
Delete $INSTDIR\skins\default\xui\ko\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\ko\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\ko\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\ko\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\ko\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\ko\floater_about.xml
Delete $INSTDIR\skins\default\xui\ko\floater_about.xml
Delete $INSTDIR\skins\default\xui\ko\alerts.xml
Delete $INSTDIR\skins\default\xui\ko\alerts.xml
Delete $INSTDIR\skins\default\xui\ko\LCD_text.xml
Delete $INSTDIR\skins\default\xui\ko\LCD_text.xml
Delete $INSTDIR\skins\default\xui\ja\xui_version.xml
Delete $INSTDIR\skins\default\xui\ja\xui_version.xml
Delete $INSTDIR\skins\default\xui\ja\ui_strings.xml
Delete $INSTDIR\skins\default\xui\ja\ui_strings.xml
Delete $INSTDIR\skins\default\xui\ja\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\ja\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\ja\strings.xml
Delete $INSTDIR\skins\default\xui\ja\strings.xml
Delete $INSTDIR\skins\default\xui\ja\role_actions.xml
Delete $INSTDIR\skins\default\xui\ja\role_actions.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\ja\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\ja\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\ja\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\ja\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\ja\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\ja\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\ja\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\ja\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_LCD.xml
Delete $INSTDIR\skins\default\xui\ja\panel_preferences_LCD.xml
Delete $INSTDIR\skins\default\xui\ja\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\ja\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\ja\panel_place.xml
Delete $INSTDIR\skins\default\xui\ja\panel_place.xml
Delete $INSTDIR\skins\default\xui\ja\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\ja\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\ja\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\ja\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\ja\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ja\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\ja\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\ja\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\ja\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\ja\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\ja\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\ja\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\ja\panel_login.xml
Delete $INSTDIR\skins\default\xui\ja\panel_login.xml
Delete $INSTDIR\skins\default\xui\ja\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\ja\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\ja\panel_groups.xml
Delete $INSTDIR\skins\default\xui\ja\panel_groups.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group.xml
Delete $INSTDIR\skins\default\xui\ja\panel_group.xml
Delete $INSTDIR\skins\default\xui\ja\panel_friends.xml
Delete $INSTDIR\skins\default\xui\ja\panel_friends.xml
Delete $INSTDIR\skins\default\xui\ja\panel_event.xml
Delete $INSTDIR\skins\default\xui\ja\panel_event.xml
Delete $INSTDIR\skins\default\xui\ja\panel_classified.xml
Delete $INSTDIR\skins\default\xui\ja\panel_classified.xml
Delete $INSTDIR\skins\default\xui\ja\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\ja\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\ja\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\ja\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\ja\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\ja\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\ja\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\ja\panel_audio.xml
Delete $INSTDIR\skins\default\xui\ja\panel_audio.xml
Delete $INSTDIR\skins\default\xui\ja\notify.xml
Delete $INSTDIR\skins\default\xui\ja\notify.xml
Delete $INSTDIR\skins\default\xui\ja\notifications.xml
Delete $INSTDIR\skins\default\xui\ja\notifications.xml
Delete $INSTDIR\skins\default\xui\ja\mime_types.xml
Delete $INSTDIR\skins\default\xui\ja\mime_types.xml
Delete $INSTDIR\skins\default\xui\ja\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\ja\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\ja\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\ja\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\ja\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\ja\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\ja\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\ja\menu_login.xml
Delete $INSTDIR\skins\default\xui\ja\menu_login.xml
Delete $INSTDIR\skins\default\xui\ja\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\ja\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\ja\fonts.xml
Delete $INSTDIR\skins\default\xui\ja\fonts.xml
Delete $INSTDIR\skins\default\xui\ja\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\ja\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\ja\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\ja\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\ja\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\ja\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\ja\floater_water.xml
Delete $INSTDIR\skins\default\xui\ja\floater_water.xml
Delete $INSTDIR\skins\default\xui\ja\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\ja\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\ja\floater_tos.xml
Delete $INSTDIR\skins\default\xui\ja\floater_tos.xml
Delete $INSTDIR\skins\default\xui\ja\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\ja\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\ja\floater_tools.xml
Delete $INSTDIR\skins\default\xui\ja\floater_tools.xml
Delete $INSTDIR\skins\default\xui\ja\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\ja\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\ja\floater_test.xml
Delete $INSTDIR\skins\default\xui\ja\floater_test.xml
Delete $INSTDIR\skins\default\xui\ja\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\ja\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\ja\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\ja\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\ja\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\ja\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\ja\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\ja\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\ja\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\ja\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\ja\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\ja\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\ja\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\ja\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\ja\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\ja\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\ja\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\ja\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\ja\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\ja\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\ja\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\ja\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\ja\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\ja\floater_profile.xml
Delete $INSTDIR\skins\default\xui\ja\floater_profile.xml
Delete $INSTDIR\skins\default\xui\ja\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\ja\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\ja\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\ja\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\ja\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\ja\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\ja\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\ja\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\ja\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\ja\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\ja\floater_pay.xml
Delete $INSTDIR\skins\default\xui\ja\floater_pay.xml
Delete $INSTDIR\skins\default\xui\ja\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\ja\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\ja\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\ja\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\ja\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\ja\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\ja\floater_notification.xml
Delete $INSTDIR\skins\default\xui\ja\floater_notification.xml
Delete $INSTDIR\skins\default\xui\ja\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\ja\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\ja\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\ja\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\ja\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\ja\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\ja\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\ja\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mute.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mute.xml
Delete $INSTDIR\skins\default\xui\ja\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\ja\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\ja\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\ja\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\ja\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\ja\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\ja\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\ja\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\ja\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\ja\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\ja\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\ja\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\ja\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\ja\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\ja\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\ja\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\ja\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\ja\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\ja\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\ja\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\ja\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\ja\floater_import.xml
Delete $INSTDIR\skins\default\xui\ja\floater_import.xml
Delete $INSTDIR\skins\default\xui\ja\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_im.xml
Delete $INSTDIR\skins\default\xui\ja\floater_im.xml
Delete $INSTDIR\skins\default\xui\ja\floater_hud.xml
Delete $INSTDIR\skins\default\xui\ja\floater_hud.xml
Delete $INSTDIR\skins\default\xui\ja\floater_html.xml
Delete $INSTDIR\skins\default\xui\ja\floater_html.xml
Delete $INSTDIR\skins\default\xui\ja\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\ja\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\ja\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\ja\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\ja\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\ja\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\ja\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\ja\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\ja\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_directory.xml
Delete $INSTDIR\skins\default\xui\ja\floater_directory.xml
Delete $INSTDIR\skins\default\xui\ja\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\ja\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\ja\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\ja\floater_customize.xml
Delete $INSTDIR\skins\default\xui\ja\floater_customize.xml
Delete $INSTDIR\skins\default\xui\ja\floater_critical.xml
Delete $INSTDIR\skins\default\xui\ja\floater_critical.xml
Delete $INSTDIR\skins\default\xui\ja\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\ja\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\ja\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\ja\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\ja\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\ja\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\ja\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\ja\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\ja\floater_camera.xml
Delete $INSTDIR\skins\default\xui\ja\floater_camera.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\ja\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\ja\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\ja\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\ja\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\ja\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\ja\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\ja\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\ja\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\ja\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\ja\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\ja\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\ja\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\ja\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\ja\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\ja\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\ja\floater_auction.xml
Delete $INSTDIR\skins\default\xui\ja\floater_auction.xml
Delete $INSTDIR\skins\default\xui\ja\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\ja\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\ja\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\ja\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\ja\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\ja\floater_about.xml
Delete $INSTDIR\skins\default\xui\ja\floater_about.xml
Delete $INSTDIR\skins\default\xui\ja\alerts.xml
Delete $INSTDIR\skins\default\xui\ja\alerts.xml
Delete $INSTDIR\skins\default\xui\ja\LCD_text.xml
Delete $INSTDIR\skins\default\xui\ja\LCD_text.xml
Delete $INSTDIR\skins\default\xui\it\ui_strings.xml
Delete $INSTDIR\skins\default\xui\it\ui_strings.xml
Delete $INSTDIR\skins\default\xui\it\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\it\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\it\strings.xml
Delete $INSTDIR\skins\default\xui\it\strings.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\it\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\it\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\it\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\it\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\it\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\it\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\it\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\it\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\it\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\it\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\it\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\it\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\it\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\it\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\it\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\it\panel_place.xml
Delete $INSTDIR\skins\default\xui\it\panel_place.xml
Delete $INSTDIR\skins\default\xui\it\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\it\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\it\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\it\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\it\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\it\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\it\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\it\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\it\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\it\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\it\panel_login.xml
Delete $INSTDIR\skins\default\xui\it\panel_login.xml
Delete $INSTDIR\skins\default\xui\it\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\it\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\it\panel_groups.xml
Delete $INSTDIR\skins\default\xui\it\panel_groups.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\it\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\it\panel_group.xml
Delete $INSTDIR\skins\default\xui\it\panel_group.xml
Delete $INSTDIR\skins\default\xui\it\panel_friends.xml
Delete $INSTDIR\skins\default\xui\it\panel_friends.xml
Delete $INSTDIR\skins\default\xui\it\panel_event.xml
Delete $INSTDIR\skins\default\xui\it\panel_event.xml
Delete $INSTDIR\skins\default\xui\it\panel_classified.xml
Delete $INSTDIR\skins\default\xui\it\panel_classified.xml
Delete $INSTDIR\skins\default\xui\it\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\it\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\it\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\it\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\it\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\it\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\it\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\it\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\it\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\it\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\it\panel_audio.xml
Delete $INSTDIR\skins\default\xui\it\panel_audio.xml
Delete $INSTDIR\skins\default\xui\it\notify.xml
Delete $INSTDIR\skins\default\xui\it\notify.xml
Delete $INSTDIR\skins\default\xui\it\notifications.xml
Delete $INSTDIR\skins\default\xui\it\notifications.xml
Delete $INSTDIR\skins\default\xui\it\mime_types.xml
Delete $INSTDIR\skins\default\xui\it\mime_types.xml
Delete $INSTDIR\skins\default\xui\it\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\it\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\it\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\it\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\it\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\it\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\it\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\it\menu_login.xml
Delete $INSTDIR\skins\default\xui\it\menu_login.xml
Delete $INSTDIR\skins\default\xui\it\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\it\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\it\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\it\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\it\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\it\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\it\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\it\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\it\floater_water.xml
Delete $INSTDIR\skins\default\xui\it\floater_water.xml
Delete $INSTDIR\skins\default\xui\it\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\it\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\it\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\it\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\it\floater_tos.xml
Delete $INSTDIR\skins\default\xui\it\floater_tos.xml
Delete $INSTDIR\skins\default\xui\it\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\it\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\it\floater_tools.xml
Delete $INSTDIR\skins\default\xui\it\floater_tools.xml
Delete $INSTDIR\skins\default\xui\it\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\it\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\it\floater_test.xml
Delete $INSTDIR\skins\default\xui\it\floater_test.xml
Delete $INSTDIR\skins\default\xui\it\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\it\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\it\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\it\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\it\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\it\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\it\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\it\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\it\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\it\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\it\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\it\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\it\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\it\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\it\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\it\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\it\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\it\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\it\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\it\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\it\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\it\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\it\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\it\floater_profile.xml
Delete $INSTDIR\skins\default\xui\it\floater_profile.xml
Delete $INSTDIR\skins\default\xui\it\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\it\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\it\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\it\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\it\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\it\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\it\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\it\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\it\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\it\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\it\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\it\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\it\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\it\floater_pay.xml
Delete $INSTDIR\skins\default\xui\it\floater_pay.xml
Delete $INSTDIR\skins\default\xui\it\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\it\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\it\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\it\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\it\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\it\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\it\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\it\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\it\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\it\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\it\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\it\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\it\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\it\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\it\floater_mute.xml
Delete $INSTDIR\skins\default\xui\it\floater_mute.xml
Delete $INSTDIR\skins\default\xui\it\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\it\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\it\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\it\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\it\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\it\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\it\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\it\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\it\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\it\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\it\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\it\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\it\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\it\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\it\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\it\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\it\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\it\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\it\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\it\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\it\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\it\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\it\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\it\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\it\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\it\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\it\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\it\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\it\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\it\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\it\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\it\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\it\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\it\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\it\floater_import.xml
Delete $INSTDIR\skins\default\xui\it\floater_import.xml
Delete $INSTDIR\skins\default\xui\it\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_im.xml
Delete $INSTDIR\skins\default\xui\it\floater_im.xml
Delete $INSTDIR\skins\default\xui\it\floater_hud.xml
Delete $INSTDIR\skins\default\xui\it\floater_hud.xml
Delete $INSTDIR\skins\default\xui\it\floater_html.xml
Delete $INSTDIR\skins\default\xui\it\floater_html.xml
Delete $INSTDIR\skins\default\xui\it\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\it\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\it\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\it\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\it\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\it\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\it\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\it\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\it\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\it\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\it\floater_directory.xml
Delete $INSTDIR\skins\default\xui\it\floater_directory.xml
Delete $INSTDIR\skins\default\xui\it\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\it\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\it\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\it\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\it\floater_customize.xml
Delete $INSTDIR\skins\default\xui\it\floater_customize.xml
Delete $INSTDIR\skins\default\xui\it\floater_critical.xml
Delete $INSTDIR\skins\default\xui\it\floater_critical.xml
Delete $INSTDIR\skins\default\xui\it\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\it\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\it\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\it\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\it\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\it\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\it\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\it\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\it\floater_camera.xml
Delete $INSTDIR\skins\default\xui\it\floater_camera.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\it\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\it\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\it\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\it\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\it\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\it\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\it\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\it\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\it\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\it\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\it\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\it\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\it\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\it\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\it\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\it\floater_auction.xml
Delete $INSTDIR\skins\default\xui\it\floater_auction.xml
Delete $INSTDIR\skins\default\xui\it\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\it\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\it\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\it\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\it\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\it\floater_about.xml
Delete $INSTDIR\skins\default\xui\it\floater_about.xml
Delete $INSTDIR\skins\default\xui\it\alerts.xml
Delete $INSTDIR\skins\default\xui\it\alerts.xml
Delete $INSTDIR\skins\default\xui\hu\xui_version.xml
Delete $INSTDIR\skins\default\xui\hu\xui_version.xml
Delete $INSTDIR\skins\default\xui\hu\ui_strings.xml
Delete $INSTDIR\skins\default\xui\hu\ui_strings.xml
Delete $INSTDIR\skins\default\xui\hu\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\hu\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\hu\strings.xml
Delete $INSTDIR\skins\default\xui\hu\strings.xml
Delete $INSTDIR\skins\default\xui\hu\role_actions.xml
Delete $INSTDIR\skins\default\xui\hu\role_actions.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\hu\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\hu\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\hu\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\hu\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\hu\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\hu\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\hu\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\hu\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\hu\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\hu\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\hu\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\hu\panel_place.xml
Delete $INSTDIR\skins\default\xui\hu\panel_place.xml
Delete $INSTDIR\skins\default\xui\hu\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\hu\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\hu\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\hu\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\hu\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\hu\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\hu\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\hu\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\hu\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\hu\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\hu\panel_login.xml
Delete $INSTDIR\skins\default\xui\hu\panel_login.xml
Delete $INSTDIR\skins\default\xui\hu\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\hu\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\hu\panel_groups.xml
Delete $INSTDIR\skins\default\xui\hu\panel_groups.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group.xml
Delete $INSTDIR\skins\default\xui\hu\panel_group.xml
Delete $INSTDIR\skins\default\xui\hu\panel_friends.xml
Delete $INSTDIR\skins\default\xui\hu\panel_friends.xml
Delete $INSTDIR\skins\default\xui\hu\panel_event.xml
Delete $INSTDIR\skins\default\xui\hu\panel_event.xml
Delete $INSTDIR\skins\default\xui\hu\panel_classified.xml
Delete $INSTDIR\skins\default\xui\hu\panel_classified.xml
Delete $INSTDIR\skins\default\xui\hu\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\hu\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\hu\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\hu\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\hu\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\hu\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\hu\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\hu\panel_audio.xml
Delete $INSTDIR\skins\default\xui\hu\panel_audio.xml
Delete $INSTDIR\skins\default\xui\hu\notify.xml
Delete $INSTDIR\skins\default\xui\hu\notify.xml
Delete $INSTDIR\skins\default\xui\hu\notifications.xml
Delete $INSTDIR\skins\default\xui\hu\notifications.xml
Delete $INSTDIR\skins\default\xui\hu\mime_types.xml
Delete $INSTDIR\skins\default\xui\hu\mime_types.xml
Delete $INSTDIR\skins\default\xui\hu\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\hu\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\hu\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\hu\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\hu\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\hu\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\hu\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\hu\menu_login.xml
Delete $INSTDIR\skins\default\xui\hu\menu_login.xml
Delete $INSTDIR\skins\default\xui\hu\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\hu\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\hu\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\hu\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\hu\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\hu\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\hu\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\hu\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\hu\floater_water.xml
Delete $INSTDIR\skins\default\xui\hu\floater_water.xml
Delete $INSTDIR\skins\default\xui\hu\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\hu\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\hu\floater_tos.xml
Delete $INSTDIR\skins\default\xui\hu\floater_tos.xml
Delete $INSTDIR\skins\default\xui\hu\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\hu\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\hu\floater_tools.xml
Delete $INSTDIR\skins\default\xui\hu\floater_tools.xml
Delete $INSTDIR\skins\default\xui\hu\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\hu\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\hu\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\hu\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\hu\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\hu\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\hu\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\hu\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\hu\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\hu\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\hu\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\hu\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\hu\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\hu\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\hu\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\hu\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\hu\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\hu\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\hu\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\hu\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\hu\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\hu\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\hu\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\hu\floater_profile.xml
Delete $INSTDIR\skins\default\xui\hu\floater_profile.xml
Delete $INSTDIR\skins\default\xui\hu\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\hu\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\hu\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\hu\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\hu\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\hu\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\hu\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\hu\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\hu\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\hu\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\hu\floater_pay.xml
Delete $INSTDIR\skins\default\xui\hu\floater_pay.xml
Delete $INSTDIR\skins\default\xui\hu\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\hu\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\hu\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\hu\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\hu\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\hu\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\hu\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\hu\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\hu\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\hu\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\hu\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\hu\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\hu\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\hu\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\hu\floater_mute.xml
Delete $INSTDIR\skins\default\xui\hu\floater_mute.xml
Delete $INSTDIR\skins\default\xui\hu\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\hu\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\hu\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\hu\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\hu\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\hu\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\hu\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\hu\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\hu\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\hu\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\hu\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\hu\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\hu\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\hu\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\hu\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\hu\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\hu\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\hu\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\hu\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\hu\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\hu\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\hu\floater_import.xml
Delete $INSTDIR\skins\default\xui\hu\floater_import.xml
Delete $INSTDIR\skins\default\xui\hu\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_im.xml
Delete $INSTDIR\skins\default\xui\hu\floater_im.xml
Delete $INSTDIR\skins\default\xui\hu\floater_hud.xml
Delete $INSTDIR\skins\default\xui\hu\floater_hud.xml
Delete $INSTDIR\skins\default\xui\hu\floater_html.xml
Delete $INSTDIR\skins\default\xui\hu\floater_html.xml
Delete $INSTDIR\skins\default\xui\hu\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\hu\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\hu\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\hu\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\hu\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\hu\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\hu\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\hu\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_directory.xml
Delete $INSTDIR\skins\default\xui\hu\floater_directory.xml
Delete $INSTDIR\skins\default\xui\hu\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\hu\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\hu\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\hu\floater_customize.xml
Delete $INSTDIR\skins\default\xui\hu\floater_customize.xml
Delete $INSTDIR\skins\default\xui\hu\floater_critical.xml
Delete $INSTDIR\skins\default\xui\hu\floater_critical.xml
Delete $INSTDIR\skins\default\xui\hu\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\hu\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\hu\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\hu\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\hu\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\hu\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\hu\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\hu\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\hu\floater_camera.xml
Delete $INSTDIR\skins\default\xui\hu\floater_camera.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\hu\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\hu\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\hu\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\hu\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\hu\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\hu\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\hu\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\hu\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\hu\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\hu\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\hu\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\hu\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\hu\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\hu\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\hu\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\hu\floater_auction.xml
Delete $INSTDIR\skins\default\xui\hu\floater_auction.xml
Delete $INSTDIR\skins\default\xui\hu\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\hu\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\hu\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\hu\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\hu\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\hu\floater_about.xml
Delete $INSTDIR\skins\default\xui\hu\floater_about.xml
Delete $INSTDIR\skins\default\xui\hu\alerts.xml
Delete $INSTDIR\skins\default\xui\hu\alerts.xml
Delete $INSTDIR\skins\default\xui\fr\xui_version.xml
Delete $INSTDIR\skins\default\xui\fr\xui_version.xml
Delete $INSTDIR\skins\default\xui\fr\ui_strings.xml
Delete $INSTDIR\skins\default\xui\fr\ui_strings.xml
Delete $INSTDIR\skins\default\xui\fr\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\fr\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\fr\strings.xml
Delete $INSTDIR\skins\default\xui\fr\strings.xml
Delete $INSTDIR\skins\default\xui\fr\role_actions.xml
Delete $INSTDIR\skins\default\xui\fr\role_actions.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\fr\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\fr\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\fr\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\fr\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\fr\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\fr\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\fr\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\fr\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\fr\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\fr\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\fr\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\fr\panel_place.xml
Delete $INSTDIR\skins\default\xui\fr\panel_place.xml
Delete $INSTDIR\skins\default\xui\fr\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\fr\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\fr\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\fr\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\fr\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\fr\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\fr\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\fr\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\fr\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\fr\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\fr\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\fr\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\fr\panel_login.xml
Delete $INSTDIR\skins\default\xui\fr\panel_login.xml
Delete $INSTDIR\skins\default\xui\fr\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\fr\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\fr\panel_groups.xml
Delete $INSTDIR\skins\default\xui\fr\panel_groups.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group.xml
Delete $INSTDIR\skins\default\xui\fr\panel_group.xml
Delete $INSTDIR\skins\default\xui\fr\panel_friends.xml
Delete $INSTDIR\skins\default\xui\fr\panel_friends.xml
Delete $INSTDIR\skins\default\xui\fr\panel_event.xml
Delete $INSTDIR\skins\default\xui\fr\panel_event.xml
Delete $INSTDIR\skins\default\xui\fr\panel_classified.xml
Delete $INSTDIR\skins\default\xui\fr\panel_classified.xml
Delete $INSTDIR\skins\default\xui\fr\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\fr\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\fr\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\fr\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\fr\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\fr\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\fr\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\fr\panel_audio.xml
Delete $INSTDIR\skins\default\xui\fr\panel_audio.xml
Delete $INSTDIR\skins\default\xui\fr\notify.xml
Delete $INSTDIR\skins\default\xui\fr\notify.xml
Delete $INSTDIR\skins\default\xui\fr\notifications.xml
Delete $INSTDIR\skins\default\xui\fr\notifications.xml
Delete $INSTDIR\skins\default\xui\fr\mime_types.xml
Delete $INSTDIR\skins\default\xui\fr\mime_types.xml
Delete $INSTDIR\skins\default\xui\fr\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\fr\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\fr\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\fr\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\fr\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\fr\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\fr\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\fr\menu_login.xml
Delete $INSTDIR\skins\default\xui\fr\menu_login.xml
Delete $INSTDIR\skins\default\xui\fr\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\fr\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\fr\fonts.xml
Delete $INSTDIR\skins\default\xui\fr\fonts.xml
Delete $INSTDIR\skins\default\xui\fr\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\fr\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\fr\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\fr\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\fr\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\fr\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\fr\floater_water.xml
Delete $INSTDIR\skins\default\xui\fr\floater_water.xml
Delete $INSTDIR\skins\default\xui\fr\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\fr\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\fr\floater_tos.xml
Delete $INSTDIR\skins\default\xui\fr\floater_tos.xml
Delete $INSTDIR\skins\default\xui\fr\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\fr\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\fr\floater_tools.xml
Delete $INSTDIR\skins\default\xui\fr\floater_tools.xml
Delete $INSTDIR\skins\default\xui\fr\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\fr\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\fr\floater_test.xml
Delete $INSTDIR\skins\default\xui\fr\floater_test.xml
Delete $INSTDIR\skins\default\xui\fr\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\fr\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\fr\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\fr\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\fr\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\fr\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\fr\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\fr\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\fr\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\fr\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\fr\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\fr\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\fr\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\fr\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\fr\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\fr\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\fr\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\fr\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\fr\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\fr\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\fr\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\fr\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\fr\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\fr\floater_profile.xml
Delete $INSTDIR\skins\default\xui\fr\floater_profile.xml
Delete $INSTDIR\skins\default\xui\fr\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\fr\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\fr\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\fr\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\fr\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\fr\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\fr\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\fr\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\fr\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\fr\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\fr\floater_pay.xml
Delete $INSTDIR\skins\default\xui\fr\floater_pay.xml
Delete $INSTDIR\skins\default\xui\fr\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\fr\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\fr\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\fr\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\fr\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\fr\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\fr\floater_notification.xml
Delete $INSTDIR\skins\default\xui\fr\floater_notification.xml
Delete $INSTDIR\skins\default\xui\fr\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\fr\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\fr\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\fr\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\fr\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\fr\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\fr\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\fr\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mute.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mute.xml
Delete $INSTDIR\skins\default\xui\fr\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\fr\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\fr\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\fr\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\fr\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\fr\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\fr\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\fr\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\fr\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\fr\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\fr\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\fr\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\fr\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\fr\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\fr\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\fr\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\fr\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\fr\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\fr\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\fr\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\fr\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\fr\floater_import.xml
Delete $INSTDIR\skins\default\xui\fr\floater_import.xml
Delete $INSTDIR\skins\default\xui\fr\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_im.xml
Delete $INSTDIR\skins\default\xui\fr\floater_im.xml
Delete $INSTDIR\skins\default\xui\fr\floater_hud.xml
Delete $INSTDIR\skins\default\xui\fr\floater_hud.xml
Delete $INSTDIR\skins\default\xui\fr\floater_html.xml
Delete $INSTDIR\skins\default\xui\fr\floater_html.xml
Delete $INSTDIR\skins\default\xui\fr\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\fr\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\fr\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\fr\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\fr\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\fr\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\fr\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\fr\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\fr\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_directory.xml
Delete $INSTDIR\skins\default\xui\fr\floater_directory.xml
Delete $INSTDIR\skins\default\xui\fr\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\fr\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\fr\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\fr\floater_customize.xml
Delete $INSTDIR\skins\default\xui\fr\floater_customize.xml
Delete $INSTDIR\skins\default\xui\fr\floater_critical.xml
Delete $INSTDIR\skins\default\xui\fr\floater_critical.xml
Delete $INSTDIR\skins\default\xui\fr\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\fr\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\fr\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\fr\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\fr\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\fr\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\fr\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\fr\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\fr\floater_camera.xml
Delete $INSTDIR\skins\default\xui\fr\floater_camera.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\fr\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\fr\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\fr\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\fr\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\fr\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\fr\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\fr\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\fr\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\fr\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\fr\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\fr\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\fr\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\fr\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\fr\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\fr\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\fr\floater_auction.xml
Delete $INSTDIR\skins\default\xui\fr\floater_auction.xml
Delete $INSTDIR\skins\default\xui\fr\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\fr\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\fr\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\fr\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\fr\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\fr\floater_about.xml
Delete $INSTDIR\skins\default\xui\fr\floater_about.xml
Delete $INSTDIR\skins\default\xui\fr\alerts.xml
Delete $INSTDIR\skins\default\xui\fr\alerts.xml
Delete $INSTDIR\skins\default\xui\es\xui_version.xml
Delete $INSTDIR\skins\default\xui\es\xui_version.xml
Delete $INSTDIR\skins\default\xui\es\ui_strings.xml
Delete $INSTDIR\skins\default\xui\es\ui_strings.xml
Delete $INSTDIR\skins\default\xui\es\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\es\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\es\strings.xml
Delete $INSTDIR\skins\default\xui\es\strings.xml
Delete $INSTDIR\skins\default\xui\es\role_actions.xml
Delete $INSTDIR\skins\default\xui\es\role_actions.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\es\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\es\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\es\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\es\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\es\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\es\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\es\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\es\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\es\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\es\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\es\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\es\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\es\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\es\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\es\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\es\panel_place.xml
Delete $INSTDIR\skins\default\xui\es\panel_place.xml
Delete $INSTDIR\skins\default\xui\es\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\es\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\es\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\es\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\es\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\es\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\es\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\es\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\es\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\es\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\es\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\es\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\es\panel_login.xml
Delete $INSTDIR\skins\default\xui\es\panel_login.xml
Delete $INSTDIR\skins\default\xui\es\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\es\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\es\panel_groups.xml
Delete $INSTDIR\skins\default\xui\es\panel_groups.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\es\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\es\panel_group.xml
Delete $INSTDIR\skins\default\xui\es\panel_group.xml
Delete $INSTDIR\skins\default\xui\es\panel_friends.xml
Delete $INSTDIR\skins\default\xui\es\panel_friends.xml
Delete $INSTDIR\skins\default\xui\es\panel_event.xml
Delete $INSTDIR\skins\default\xui\es\panel_event.xml
Delete $INSTDIR\skins\default\xui\es\panel_classified.xml
Delete $INSTDIR\skins\default\xui\es\panel_classified.xml
Delete $INSTDIR\skins\default\xui\es\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\es\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\es\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\es\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\es\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\es\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\es\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\es\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\es\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\es\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\es\panel_audio.xml
Delete $INSTDIR\skins\default\xui\es\panel_audio.xml
Delete $INSTDIR\skins\default\xui\es\notify.xml
Delete $INSTDIR\skins\default\xui\es\notify.xml
Delete $INSTDIR\skins\default\xui\es\notifications.xml
Delete $INSTDIR\skins\default\xui\es\notifications.xml
Delete $INSTDIR\skins\default\xui\es\mime_types.xml
Delete $INSTDIR\skins\default\xui\es\mime_types.xml
Delete $INSTDIR\skins\default\xui\es\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\es\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\es\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\es\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\es\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\es\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\es\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\es\menu_login.xml
Delete $INSTDIR\skins\default\xui\es\menu_login.xml
Delete $INSTDIR\skins\default\xui\es\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\es\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\es\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\es\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\es\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\es\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\es\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\es\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\es\floater_water.xml
Delete $INSTDIR\skins\default\xui\es\floater_water.xml
Delete $INSTDIR\skins\default\xui\es\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\es\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\es\floater_tos.xml
Delete $INSTDIR\skins\default\xui\es\floater_tos.xml
Delete $INSTDIR\skins\default\xui\es\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\es\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\es\floater_tools.xml
Delete $INSTDIR\skins\default\xui\es\floater_tools.xml
Delete $INSTDIR\skins\default\xui\es\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\es\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\es\floater_test.xml
Delete $INSTDIR\skins\default\xui\es\floater_test.xml
Delete $INSTDIR\skins\default\xui\es\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\es\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\es\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\es\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\es\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\es\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\es\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\es\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\es\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\es\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\es\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\es\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\es\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\es\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\es\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\es\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\es\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\es\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\es\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\es\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\es\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\es\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\es\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\es\floater_profile.xml
Delete $INSTDIR\skins\default\xui\es\floater_profile.xml
Delete $INSTDIR\skins\default\xui\es\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\es\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\es\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\es\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\es\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\es\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\es\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\es\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\es\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\es\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\es\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\es\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\es\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\es\floater_pay.xml
Delete $INSTDIR\skins\default\xui\es\floater_pay.xml
Delete $INSTDIR\skins\default\xui\es\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\es\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\es\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\es\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\es\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\es\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\es\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\es\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\es\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\es\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\es\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\es\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\es\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\es\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\es\floater_mute.xml
Delete $INSTDIR\skins\default\xui\es\floater_mute.xml
Delete $INSTDIR\skins\default\xui\es\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\es\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\es\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\es\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\es\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\es\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\es\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\es\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\es\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\es\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\es\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\es\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\es\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\es\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\es\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\es\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\es\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\es\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\es\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\es\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\es\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\es\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\es\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\es\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\es\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\es\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\es\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\es\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\es\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\es\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\es\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\es\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\es\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\es\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\es\floater_import.xml
Delete $INSTDIR\skins\default\xui\es\floater_import.xml
Delete $INSTDIR\skins\default\xui\es\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_im.xml
Delete $INSTDIR\skins\default\xui\es\floater_im.xml
Delete $INSTDIR\skins\default\xui\es\floater_hud.xml
Delete $INSTDIR\skins\default\xui\es\floater_hud.xml
Delete $INSTDIR\skins\default\xui\es\floater_html.xml
Delete $INSTDIR\skins\default\xui\es\floater_html.xml
Delete $INSTDIR\skins\default\xui\es\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\es\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\es\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\es\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\es\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\es\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\es\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\es\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\es\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\es\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\es\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\es\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\es\floater_directory.xml
Delete $INSTDIR\skins\default\xui\es\floater_directory.xml
Delete $INSTDIR\skins\default\xui\es\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\es\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\es\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\es\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\es\floater_customize.xml
Delete $INSTDIR\skins\default\xui\es\floater_customize.xml
Delete $INSTDIR\skins\default\xui\es\floater_critical.xml
Delete $INSTDIR\skins\default\xui\es\floater_critical.xml
Delete $INSTDIR\skins\default\xui\es\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\es\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\es\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\es\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\es\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\es\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\es\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\es\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\es\floater_camera.xml
Delete $INSTDIR\skins\default\xui\es\floater_camera.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\es\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\es\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\es\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\es\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\es\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\es\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\es\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\es\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\es\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\es\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\es\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\es\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\es\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\es\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\es\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\es\floater_auction.xml
Delete $INSTDIR\skins\default\xui\es\floater_auction.xml
Delete $INSTDIR\skins\default\xui\es\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\es\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\es\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\es\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\es\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\es\floater_about.xml
Delete $INSTDIR\skins\default\xui\es\floater_about.xml
Delete $INSTDIR\skins\default\xui\es\alerts.xml
Delete $INSTDIR\skins\default\xui\es\alerts.xml
Delete $INSTDIR\skins\default\xui\en-us\xui_version.xml
Delete $INSTDIR\skins\default\xui\en-us\xui_version.xml
Delete $INSTDIR\skins\default\xui\en-us\ui_strings.xml
Delete $INSTDIR\skins\default\xui\en-us\ui_strings.xml
Delete $INSTDIR\skins\default\xui\en-us\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\en-us\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\en-us\strings.xml
Delete $INSTDIR\skins\default\xui\en-us\strings.xml
Delete $INSTDIR\skins\default\xui\en-us\role_actions.xml
Delete $INSTDIR\skins\default\xui\en-us\role_actions.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_progress.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_progress.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_ascent_vanity.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_ascent_vanity.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_ascent_system.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_ascent_system.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_place.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_place.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_hud.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_hud.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_login.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_login.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_hud.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_hud.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_groups.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_groups.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_group.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_friends.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_friends.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_event.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_event.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_classified.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_classified.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_bg_toolbar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_bg_toolbar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_bg_tab.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_bg_tab.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_bars.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_bars.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_audio.xml
Delete $INSTDIR\skins\default\xui\en-us\panel_audio.xml
Delete $INSTDIR\skins\default\xui\en-us\notify.xml
Delete $INSTDIR\skins\default\xui\en-us\notify.xml
Delete $INSTDIR\skins\default\xui\en-us\notifications.xml
Delete $INSTDIR\skins\default\xui\en-us\notifications.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types_windows.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types_windows.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types_mac.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types_mac.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types_linux.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types_linux.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_login.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_login.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\en-us\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\en-us\fonts.xml
Delete $INSTDIR\skins\default\xui\en-us\fonts.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_water.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_water.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_vfs_explorer.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_vfs_explorer.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_vfs.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_vfs.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_tos.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_tos.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_tools.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_tools.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_test.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_test.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_teleport_history.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_teleport_history.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_teleport.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_teleport.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_radar.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_radar.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_profile.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_profile.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_new_landmark.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_new_landmark.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_pay.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_pay.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_notification.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_notification.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_new_local_inventory.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_new_local_inventory.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mute.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mute.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_message_log.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_message_log.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_message_builder.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_message_builder.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_backup_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_backup_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_backup.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory_backup.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_import_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_import_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_import.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_import.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_im.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_im.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_hud.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_hud.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_html_simple.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_html_simple.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_html.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_html.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_hex.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_hex.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_export.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_export.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_explore_sounds.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_explore_sounds.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_explore_animations.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_explore_animations.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_directory.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_directory.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_dickdongs.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_dickdongs.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_customize.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_customize.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_critical.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_critical.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_clothing.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_clothing.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_camera.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_camera.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_blacklist.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_blacklist.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_auction.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_auction.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_asset_text_editor.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_asset_text_editor.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_area_search.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_area_search.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_ao.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_ao.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_about.xml
Delete $INSTDIR\skins\default\xui\en-us\floater_about.xml
Delete $INSTDIR\skins\default\xui\en-us\alerts.xml
Delete $INSTDIR\skins\default\xui\en-us\alerts.xml
Delete $INSTDIR\skins\default\xui\de\xui_version.xml
Delete $INSTDIR\skins\default\xui\de\xui_version.xml
Delete $INSTDIR\skins\default\xui\de\ui_strings.xml
Delete $INSTDIR\skins\default\xui\de\ui_strings.xml
Delete $INSTDIR\skins\default\xui\de\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\de\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\de\strings.xml
Delete $INSTDIR\skins\default\xui\de\strings.xml
Delete $INSTDIR\skins\default\xui\de\role_actions.xml
Delete $INSTDIR\skins\default\xui\de\role_actions.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_remote.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\de\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\de\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\de\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\de\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\de\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\de\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\de\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\de\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\de\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\de\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\de\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\de\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_LCD.xml
Delete $INSTDIR\skins\default\xui\de\panel_preferences_LCD.xml
Delete $INSTDIR\skins\default\xui\de\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\de\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\de\panel_place.xml
Delete $INSTDIR\skins\default\xui\de\panel_place.xml
Delete $INSTDIR\skins\default\xui\de\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\de\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\de\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\de\panel_notifications_channel.xml
Delete $INSTDIR\skins\default\xui\de\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\de\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\de\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\de\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\de\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\de\panel_media_remote.xml
Delete $INSTDIR\skins\default\xui\de\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\de\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\de\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\de\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\de\panel_login.xml
Delete $INSTDIR\skins\default\xui\de\panel_login.xml
Delete $INSTDIR\skins\default\xui\de\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\de\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\de\panel_groups.xml
Delete $INSTDIR\skins\default\xui\de\panel_groups.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\de\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\de\panel_group.xml
Delete $INSTDIR\skins\default\xui\de\panel_group.xml
Delete $INSTDIR\skins\default\xui\de\panel_friends.xml
Delete $INSTDIR\skins\default\xui\de\panel_friends.xml
Delete $INSTDIR\skins\default\xui\de\panel_event.xml
Delete $INSTDIR\skins\default\xui\de\panel_event.xml
Delete $INSTDIR\skins\default\xui\de\panel_classified.xml
Delete $INSTDIR\skins\default\xui\de\panel_classified.xml
Delete $INSTDIR\skins\default\xui\de\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\de\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\de\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\de\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\de\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\de\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\de\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\de\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\de\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\de\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\de\panel_audio.xml
Delete $INSTDIR\skins\default\xui\de\panel_audio.xml
Delete $INSTDIR\skins\default\xui\de\notify.xml
Delete $INSTDIR\skins\default\xui\de\notify.xml
Delete $INSTDIR\skins\default\xui\de\notifications.xml
Delete $INSTDIR\skins\default\xui\de\notifications.xml
Delete $INSTDIR\skins\default\xui\de\mime_types.xml
Delete $INSTDIR\skins\default\xui\de\mime_types.xml
Delete $INSTDIR\skins\default\xui\de\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\de\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\de\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\de\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\de\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\de\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\de\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\de\menu_login.xml
Delete $INSTDIR\skins\default\xui\de\menu_login.xml
Delete $INSTDIR\skins\default\xui\de\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\de\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\de\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\de\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\de\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\de\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\de\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\de\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\de\floater_water.xml
Delete $INSTDIR\skins\default\xui\de\floater_water.xml
Delete $INSTDIR\skins\default\xui\de\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\de\floater_voice_wizard.xml
Delete $INSTDIR\skins\default\xui\de\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\de\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\de\floater_tos.xml
Delete $INSTDIR\skins\default\xui\de\floater_tos.xml
Delete $INSTDIR\skins\default\xui\de\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\de\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\de\floater_tools.xml
Delete $INSTDIR\skins\default\xui\de\floater_tools.xml
Delete $INSTDIR\skins\default\xui\de\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\de\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\de\floater_test.xml
Delete $INSTDIR\skins\default\xui\de\floater_test.xml
Delete $INSTDIR\skins\default\xui\de\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\de\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\de\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\de\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\de\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\de\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\de\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\de\floater_skin_preview_template.xml
Delete $INSTDIR\skins\default\xui\de\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\de\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\de\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\de\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\de\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\de\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\de\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\de\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\de\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\de\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\de\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\de\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\de\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\de\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\de\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\de\floater_profile.xml
Delete $INSTDIR\skins\default\xui\de\floater_profile.xml
Delete $INSTDIR\skins\default\xui\de\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\de\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\de\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\de\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\de\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\de\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\de\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\de\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\de\floater_post_process.xml
Delete $INSTDIR\skins\default\xui\de\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\de\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\de\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\de\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\de\floater_pay.xml
Delete $INSTDIR\skins\default\xui\de\floater_pay.xml
Delete $INSTDIR\skins\default\xui\de\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\de\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\de\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\de\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\de\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\de\floater_notifications_console.xml
Delete $INSTDIR\skins\default\xui\de\floater_notification.xml
Delete $INSTDIR\skins\default\xui\de\floater_notification.xml
Delete $INSTDIR\skins\default\xui\de\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\de\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\de\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\de\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\de\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\de\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\de\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\de\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\de\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\de\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\de\floater_mute.xml
Delete $INSTDIR\skins\default\xui\de\floater_mute.xml
Delete $INSTDIR\skins\default\xui\de\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\de\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\de\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\de\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\de\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\de\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\de\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\de\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\de\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\de\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\de\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\de\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\de\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\de\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\de\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\de\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\de\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\de\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\de\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\de\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\de\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\de\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\de\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\de\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\de\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\de\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\de\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\de\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\de\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\de\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\de\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\de\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\de\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\de\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\de\floater_import.xml
Delete $INSTDIR\skins\default\xui\de\floater_import.xml
Delete $INSTDIR\skins\default\xui\de\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_im.xml
Delete $INSTDIR\skins\default\xui\de\floater_im.xml
Delete $INSTDIR\skins\default\xui\de\floater_hud.xml
Delete $INSTDIR\skins\default\xui\de\floater_hud.xml
Delete $INSTDIR\skins\default\xui\de\floater_html.xml
Delete $INSTDIR\skins\default\xui\de\floater_html.xml
Delete $INSTDIR\skins\default\xui\de\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\de\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\de\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\de\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\de\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\de\floater_god_tools.xml
Delete $INSTDIR\skins\default\xui\de\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\de\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\de\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\de\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\de\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\de\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\de\floater_directory.xml
Delete $INSTDIR\skins\default\xui\de\floater_directory.xml
Delete $INSTDIR\skins\default\xui\de\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\de\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\de\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\de\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\de\floater_customize.xml
Delete $INSTDIR\skins\default\xui\de\floater_customize.xml
Delete $INSTDIR\skins\default\xui\de\floater_critical.xml
Delete $INSTDIR\skins\default\xui\de\floater_critical.xml
Delete $INSTDIR\skins\default\xui\de\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\de\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\de\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\de\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\de\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\de\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\de\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\de\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\de\floater_camera.xml
Delete $INSTDIR\skins\default\xui\de\floater_camera.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\de\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\de\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\de\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\de\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\de\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\de\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\de\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\de\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\de\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\de\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\de\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\de\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\de\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\de\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\de\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\de\floater_auction.xml
Delete $INSTDIR\skins\default\xui\de\floater_auction.xml
Delete $INSTDIR\skins\default\xui\de\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\de\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\de\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\de\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\de\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\de\floater_about.xml
Delete $INSTDIR\skins\default\xui\de\floater_about.xml
Delete $INSTDIR\skins\default\xui\de\alerts.xml
Delete $INSTDIR\skins\default\xui\de\alerts.xml
Delete $INSTDIR\skins\default\xui\de\LCD_text.xml
Delete $INSTDIR\skins\default\xui\de\LCD_text.xml
Delete $INSTDIR\skins\default\xui\da\xui_version.xml
Delete $INSTDIR\skins\default\xui\da\xui_version.xml
Delete $INSTDIR\skins\default\xui\da\ui_strings.xml
Delete $INSTDIR\skins\default\xui\da\ui_strings.xml
Delete $INSTDIR\skins\default\xui\da\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\da\teleport_strings.xml
Delete $INSTDIR\skins\default\xui\da\strings.xml
Delete $INSTDIR\skins\default\xui\da\strings.xml
Delete $INSTDIR\skins\default\xui\da\role_actions.xml
Delete $INSTDIR\skins\default\xui\da\role_actions.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_options.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_enable.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\da\panel_voice_controls.xml
Delete $INSTDIR\skins\default\xui\da\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\da\panel_top_pick.xml
Delete $INSTDIR\skins\default\xui\da\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\da\panel_toolbar.xml
Delete $INSTDIR\skins\default\xui\da\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\da\panel_status_bar.xml
Delete $INSTDIR\skins\default\xui\da\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\da\panel_speaker_controls.xml
Delete $INSTDIR\skins\default\xui\da\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\da\panel_scrolling_param.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_texture.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_terrain.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_general.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_estate.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_debug.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\da\panel_region_covenant.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_web.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_voice.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_skins.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_popups.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_network.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_input.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_im.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_graphics1.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_general.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_chat.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\da\panel_preferences_audio.xml
Delete $INSTDIR\skins\default\xui\da\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\da\panel_place_small.xml
Delete $INSTDIR\skins\default\xui\da\panel_place.xml
Delete $INSTDIR\skins\default\xui\da\panel_place.xml
Delete $INSTDIR\skins\default\xui\da\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\da\panel_overlaybar.xml
Delete $INSTDIR\skins\default\xui\da\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\da\panel_mini_map.xml
Delete $INSTDIR\skins\default\xui\da\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\da\panel_media_remote_expanded.xml
Delete $INSTDIR\skins\default\xui\da\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\da\panel_media_controls.xml
Delete $INSTDIR\skins\default\xui\da\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\da\panel_master_volume.xml
Delete $INSTDIR\skins\default\xui\da\panel_login.xml
Delete $INSTDIR\skins\default\xui\da\panel_login.xml
Delete $INSTDIR\skins\default\xui\da\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\da\panel_land_covenant.xml
Delete $INSTDIR\skins\default\xui\da\panel_groups.xml
Delete $INSTDIR\skins\default\xui\da\panel_groups.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_voting.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_roles.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_notices.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_land_money.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_invite.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_general.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\da\panel_group_finder.xml
Delete $INSTDIR\skins\default\xui\da\panel_group.xml
Delete $INSTDIR\skins\default\xui\da\panel_group.xml
Delete $INSTDIR\skins\default\xui\da\panel_friends.xml
Delete $INSTDIR\skins\default\xui\da\panel_friends.xml
Delete $INSTDIR\skins\default\xui\da\panel_event.xml
Delete $INSTDIR\skins\default\xui\da\panel_event.xml
Delete $INSTDIR\skins\default\xui\da\panel_classified.xml
Delete $INSTDIR\skins\default\xui\da\panel_classified.xml
Delete $INSTDIR\skins\default\xui\da\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\da\panel_chat_bar.xml
Delete $INSTDIR\skins\default\xui\da\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\da\panel_avatar_pick.xml
Delete $INSTDIR\skins\default\xui\da\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\da\panel_avatar_classified.xml
Delete $INSTDIR\skins\default\xui\da\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\da\panel_avatar.xml
Delete $INSTDIR\skins\default\xui\da\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\da\panel_audio_device.xml
Delete $INSTDIR\skins\default\xui\da\panel_audio.xml
Delete $INSTDIR\skins\default\xui\da\panel_audio.xml
Delete $INSTDIR\skins\default\xui\da\notify.xml
Delete $INSTDIR\skins\default\xui\da\notify.xml
Delete $INSTDIR\skins\default\xui\da\notifications.xml
Delete $INSTDIR\skins\default\xui\da\notifications.xml
Delete $INSTDIR\skins\default\xui\da\mime_types.xml
Delete $INSTDIR\skins\default\xui\da\mime_types.xml
Delete $INSTDIR\skins\default\xui\da\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\da\menu_viewer.xml
Delete $INSTDIR\skins\default\xui\da\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\da\menu_slurl.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_self.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_object.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_land.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_avatar.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\da\menu_pie_attachment.xml
Delete $INSTDIR\skins\default\xui\da\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\da\menu_mini_map.xml
Delete $INSTDIR\skins\default\xui\da\menu_login.xml
Delete $INSTDIR\skins\default\xui\da\menu_login.xml
Delete $INSTDIR\skins\default\xui\da\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\da\menu_inventory.xml
Delete $INSTDIR\skins\default\xui\da\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\da\floater_world_map.xml
Delete $INSTDIR\skins\default\xui\da\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\da\floater_windlight_options.xml
Delete $INSTDIR\skins\default\xui\da\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\da\floater_wearable_save_as.xml
Delete $INSTDIR\skins\default\xui\da\floater_water.xml
Delete $INSTDIR\skins\default\xui\da\floater_water.xml
Delete $INSTDIR\skins\default\xui\da\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\da\floater_url_entry.xml
Delete $INSTDIR\skins\default\xui\da\floater_tos.xml
Delete $INSTDIR\skins\default\xui\da\floater_tos.xml
Delete $INSTDIR\skins\default\xui\da\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\da\floater_top_objects.xml
Delete $INSTDIR\skins\default\xui\da\floater_tools.xml
Delete $INSTDIR\skins\default\xui\da\floater_tools.xml
Delete $INSTDIR\skins\default\xui\da\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\da\floater_texture_ctrl.xml
Delete $INSTDIR\skins\default\xui\da\floater_test.xml
Delete $INSTDIR\skins\default\xui\da\floater_test.xml
Delete $INSTDIR\skins\default\xui\da\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\da\floater_telehub.xml
Delete $INSTDIR\skins\default\xui\da\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\da\floater_statistics.xml
Delete $INSTDIR\skins\default\xui\da\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_sound_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\da\floater_snapshot.xml
Delete $INSTDIR\skins\default\xui\da\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\da\floater_sim_release_message.xml
Delete $INSTDIR\skins\default\xui\da\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\da\floater_settings_debug.xml
Delete $INSTDIR\skins\default\xui\da\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\da\floater_sell_land.xml
Delete $INSTDIR\skins\default\xui\da\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\da\floater_select_key.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_search.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_queue.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_ed_panel.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\da\floater_script_debug.xml
Delete $INSTDIR\skins\default\xui\da\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\da\floater_report_bug.xml
Delete $INSTDIR\skins\default\xui\da\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\da\floater_report_abuse.xml
Delete $INSTDIR\skins\default\xui\da\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\da\floater_region_info.xml
Delete $INSTDIR\skins\default\xui\da\floater_profile.xml
Delete $INSTDIR\skins\default\xui\da\floater_profile.xml
Delete $INSTDIR\skins\default\xui\da\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\da\floater_price_for_listing.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_url.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_texture_keep_discard.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_texture.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_sound.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_notecard_keep_discard.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_notecard.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_gesture.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_existing_landmark.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_event.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_embedded_texture.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_classified.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\da\floater_preview_animation.xml
Delete $INSTDIR\skins\default\xui\da\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\da\floater_preferences.xml
Delete $INSTDIR\skins\default\xui\da\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\da\floater_postcard.xml
Delete $INSTDIR\skins\default\xui\da\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\da\floater_perm_prefs.xml
Delete $INSTDIR\skins\default\xui\da\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\da\floater_pay_object.xml
Delete $INSTDIR\skins\default\xui\da\floater_pay.xml
Delete $INSTDIR\skins\default\xui\da\floater_pay.xml
Delete $INSTDIR\skins\default\xui\da\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\da\floater_openobject.xml
Delete $INSTDIR\skins\default\xui\da\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\da\floater_object_im_info.xml
Delete $INSTDIR\skins\default\xui\da\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\da\floater_new_outfit_dialog.xml
Delete $INSTDIR\skins\default\xui\da\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\da\floater_new_im.xml
Delete $INSTDIR\skins\default\xui\da\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\da\floater_name_description.xml
Delete $INSTDIR\skins\default\xui\da\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\da\floater_my_friends.xml
Delete $INSTDIR\skins\default\xui\da\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\da\floater_mute_object.xml
Delete $INSTDIR\skins\default\xui\da\floater_mute.xml
Delete $INSTDIR\skins\default\xui\da\floater_mute.xml
Delete $INSTDIR\skins\default\xui\da\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\da\floater_moveview.xml
Delete $INSTDIR\skins\default\xui\da\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\da\floater_mini_map.xml
Delete $INSTDIR\skins\default\xui\da\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\da\floater_mem_leaking.xml
Delete $INSTDIR\skins\default\xui\da\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\da\floater_media_browser.xml
Delete $INSTDIR\skins\default\xui\da\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\da\floater_lsl_guide.xml
Delete $INSTDIR\skins\default\xui\da\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\da\floater_live_lsleditor.xml
Delete $INSTDIR\skins\default\xui\da\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\da\floater_landmark_ctrl.xml
Delete $INSTDIR\skins\default\xui\da\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\da\floater_land_holdings.xml
Delete $INSTDIR\skins\default\xui\da\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\da\floater_lagmeter.xml
Delete $INSTDIR\skins\default\xui\da\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\da\floater_joystick.xml
Delete $INSTDIR\skins\default\xui\da\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\da\floater_inventory_view_finder.xml
Delete $INSTDIR\skins\default\xui\da\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\da\floater_inventory_item_properties.xml
Delete $INSTDIR\skins\default\xui\da\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\da\floater_inventory.xml
Delete $INSTDIR\skins\default\xui\da\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\da\floater_instant_message_group.xml
Delete $INSTDIR\skins\default\xui\da\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\da\floater_instant_message_ad_hoc.xml
Delete $INSTDIR\skins\default\xui\da\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\da\floater_instant_message.xml
Delete $INSTDIR\skins\default\xui\da\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\da\floater_inspect.xml
Delete $INSTDIR\skins\default\xui\da\floater_import.xml
Delete $INSTDIR\skins\default\xui\da\floater_import.xml
Delete $INSTDIR\skins\default\xui\da\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_image_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_im.xml
Delete $INSTDIR\skins\default\xui\da\floater_im.xml
Delete $INSTDIR\skins\default\xui\da\floater_hud.xml
Delete $INSTDIR\skins\default\xui\da\floater_hud.xml
Delete $INSTDIR\skins\default\xui\da\floater_html.xml
Delete $INSTDIR\skins\default\xui\da\floater_html.xml
Delete $INSTDIR\skins\default\xui\da\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\da\floater_hardware_settings.xml
Delete $INSTDIR\skins\default\xui\da\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\da\floater_group_info.xml
Delete $INSTDIR\skins\default\xui\da\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\da\floater_gesture.xml
Delete $INSTDIR\skins\default\xui\da\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\da\floater_font_test.xml
Delete $INSTDIR\skins\default\xui\da\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\da\floater_env_settings.xml
Delete $INSTDIR\skins\default\xui\da\floater_directory.xml
Delete $INSTDIR\skins\default\xui\da\floater_directory.xml
Delete $INSTDIR\skins\default\xui\da\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\da\floater_device_settings.xml
Delete $INSTDIR\skins\default\xui\da\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\da\floater_day_cycle_options.xml
Delete $INSTDIR\skins\default\xui\da\floater_customize.xml
Delete $INSTDIR\skins\default\xui\da\floater_customize.xml
Delete $INSTDIR\skins\default\xui\da\floater_critical.xml
Delete $INSTDIR\skins\default\xui\da\floater_critical.xml
Delete $INSTDIR\skins\default\xui\da\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\da\floater_color_picker.xml
Delete $INSTDIR\skins\default\xui\da\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\da\floater_choose_group.xml
Delete $INSTDIR\skins\default\xui\da\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\da\floater_chatterbox.xml
Delete $INSTDIR\skins\default\xui\da\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\da\floater_chat_history.xml
Delete $INSTDIR\skins\default\xui\da\floater_camera.xml
Delete $INSTDIR\skins\default\xui\da\floater_camera.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_object.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_land.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_currency.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\da\floater_buy_contents.xml
Delete $INSTDIR\skins\default\xui\da\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\da\floater_bumps.xml
Delete $INSTDIR\skins\default\xui\da\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\da\floater_bulk_perms.xml
Delete $INSTDIR\skins\default\xui\da\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\da\floater_build_options.xml
Delete $INSTDIR\skins\default\xui\da\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\da\floater_beacons.xml
Delete $INSTDIR\skins\default\xui\da\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\da\floater_avatar_textures.xml
Delete $INSTDIR\skins\default\xui\da\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\da\floater_avatar_picker.xml
Delete $INSTDIR\skins\default\xui\da\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\da\floater_audio_volume.xml
Delete $INSTDIR\skins\default\xui\da\floater_auction.xml
Delete $INSTDIR\skins\default\xui\da\floater_auction.xml
Delete $INSTDIR\skins\default\xui\da\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_animation_preview.xml
Delete $INSTDIR\skins\default\xui\da\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\da\floater_active_speakers.xml
Delete $INSTDIR\skins\default\xui\da\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\da\floater_about_land.xml
Delete $INSTDIR\skins\default\xui\da\floater_about.xml
Delete $INSTDIR\skins\default\xui\da\floater_about.xml
Delete $INSTDIR\skins\default\xui\da\alerts.xml
Delete $INSTDIR\skins\default\xui\da\alerts.xml
Delete $INSTDIR\app_settings\shaders\class3\lighting\sumLightsV.glsl
Delete $INSTDIR\app_settings\shaders\class3\lighting\sumLightsV.glsl
Delete $INSTDIR\app_settings\shaders\class3\lighting\sumLightsSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class3\lighting\sumLightsSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class3\avatar\avatarV.glsl
Delete $INSTDIR\app_settings\shaders\class3\avatar\avatarV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\transportF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\transportF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\skyV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\skyV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\skyF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\skyF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\gammaF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\gammaF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\cloudsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\cloudsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\cloudsF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\cloudsF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsVarsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsVarsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsVarsF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsVarsF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsHelpersV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsHelpersV.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsF.glsl
Delete $INSTDIR\app_settings\shaders\class2\windlight\atmosphericsF.glsl
Delete $INSTDIR\app_settings\shaders\class2\objects\shinyV.glsl
Delete $INSTDIR\app_settings\shaders\class2\objects\shinyV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\sumLightsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\sumLightsV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\sumLightsSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\sumLightsSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightShinyWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightShinyWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightFullbrightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightFullbrightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightFullbrightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightFullbrightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightFullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightFullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightF.glsl
Delete $INSTDIR\app_settings\shaders\class2\lighting\lightF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\waterFogF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\waterFogF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\waterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\waterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\underWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\underWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\terrainWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\terrainWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\terrainV.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\terrainV.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\terrainF.glsl
Delete $INSTDIR\app_settings\shaders\class2\environment\terrainF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\simpleF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\simpleF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\nightVisionF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\nightVisionF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\extractF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\extractF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\drawQuadV.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\drawQuadV.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\colorFilterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\colorFilterF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\blurV.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\blurV.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\blurF.glsl
Delete $INSTDIR\app_settings\shaders\class2\effects\blurF.glsl
Delete $INSTDIR\app_settings\shaders\class2\avatar\eyeballV.glsl
Delete $INSTDIR\app_settings\shaders\class2\avatar\eyeballV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\transportF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\transportF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\gammaF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\gammaF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsVarsV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsVarsV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsVarsF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsVarsF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsHelpersV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsHelpersV.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsF.glsl
Delete $INSTDIR\app_settings\shaders\class1\windlight\atmosphericsF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\simpleWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\simpleWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\simpleV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\simpleV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\simpleF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\simpleF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\shinyWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\shinyWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\shinyV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\shinyV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\shinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\shinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightShinyV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightShinyV.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\objects\fullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\sumLightsV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\sumLightsV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\sumLightsSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\sumLightsSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightShinyWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightShinyWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFuncV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFuncV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFuncSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFuncSpecularV.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFullbrightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFullbrightWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFullbrightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFullbrightShinyF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightFullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\lighting\lightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\interface\highlightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\interface\highlightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\interface\highlightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\interface\highlightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\waterV.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\waterV.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\waterFogF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\waterFogF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\waterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\waterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\underWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\underWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\terrainWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\terrainWaterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\terrainV.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\terrainV.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\terrainF.glsl
Delete $INSTDIR\app_settings\shaders\class1\environment\terrainF.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowV.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowV.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowF.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowF.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowExtractV.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowExtractV.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowExtractF.glsl
Delete $INSTDIR\app_settings\shaders\class1\effects\glowExtractF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\waterV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\waterV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\waterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\waterF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\treeV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\treeV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\treeF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\treeF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\terrainV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\terrainV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\terrainF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\terrainF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\sunLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\sunLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\sunLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\sunLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\softenLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\softenLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\softenLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\softenLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\shadowV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\shadowV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\shadowF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\shadowF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\pointLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\pointLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\pointLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\pointLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\multiPointLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\multiPointLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\impostorV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\impostorV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\impostorF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\impostorF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\fullbrightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\fullbrightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\fullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\fullbrightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\diffuseV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\diffuseV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\diffuseF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\diffuseF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\bumpV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\bumpV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\bumpF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\bumpF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\blurLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\blurLightV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\blurLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\blurLightF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarShadowV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarShadowV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarShadowF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarShadowF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarAlphaV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarAlphaV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarAlphaF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\avatarAlphaF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\alphaV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\alphaV.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\alphaF.glsl
Delete $INSTDIR\app_settings\shaders\class1\deferred\alphaF.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\pickAvatarV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\pickAvatarV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\pickAvatarF.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\pickAvatarF.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\eyeballV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\eyeballV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\eyeballF.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\eyeballF.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\avatarV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\avatarV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\avatarSkinV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\avatarSkinV.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\avatarF.glsl
Delete $INSTDIR\app_settings\shaders\class1\avatar\avatarF.glsl
Delete $INSTDIR\skins\silver\textures\white.tga
Delete $INSTDIR\skins\silver\textures\white.tga
Delete $INSTDIR\skins\silver\textures\up_arrow.png
Delete $INSTDIR\skins\silver\textures\up_arrow.png
Delete $INSTDIR\skins\silver\textures\toolbar_tab.tga
Delete $INSTDIR\skins\silver\textures\toolbar_tab.tga
Delete $INSTDIR\skins\silver\textures\toolbar_btn_selected.tga
Delete $INSTDIR\skins\silver\textures\toolbar_btn_selected.tga
Delete $INSTDIR\skins\silver\textures\toolbar_btn_enabled.tga
Delete $INSTDIR\skins\silver\textures\toolbar_btn_enabled.tga
Delete $INSTDIR\skins\silver\textures\toolbar_btn_disabled.tga
Delete $INSTDIR\skins\silver\textures\toolbar_btn_disabled.tga
Delete $INSTDIR\skins\silver\textures\toolbar_bg.tga
Delete $INSTDIR\skins\silver\textures\toolbar_bg.tga
Delete $INSTDIR\skins\silver\textures\tool_zoom_active.tga
Delete $INSTDIR\skins\silver\textures\tool_zoom_active.tga
Delete $INSTDIR\skins\silver\textures\tool_zoom.tga
Delete $INSTDIR\skins\silver\textures\tool_zoom.tga
Delete $INSTDIR\skins\silver\textures\tool_dozer_active.tga
Delete $INSTDIR\skins\silver\textures\tool_dozer_active.tga
Delete $INSTDIR\skins\silver\textures\tool_dozer.tga
Delete $INSTDIR\skins\silver\textures\tool_dozer.tga
Delete $INSTDIR\skins\silver\textures\textures.xml
Delete $INSTDIR\skins\silver\textures\textures.xml
Delete $INSTDIR\skins\silver\textures\tearoffbox.tga
Delete $INSTDIR\skins\silver\textures\tearoffbox.tga
Delete $INSTDIR\skins\silver\textures\tearoff_pressed.tga
Delete $INSTDIR\skins\silver\textures\tearoff_pressed.tga
Delete $INSTDIR\skins\silver\textures\tab_top_selected_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_top_selected_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_top_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_top_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_left_selected.tga
Delete $INSTDIR\skins\silver\textures\tab_left_selected.tga
Delete $INSTDIR\skins\silver\textures\tab_left.tga
Delete $INSTDIR\skins\silver\textures\tab_left.tga
Delete $INSTDIR\skins\silver\textures\tab_bottom_selected_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_bottom_selected_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_bottom_blue.tga
Delete $INSTDIR\skins\silver\textures\tab_bottom_blue.tga
Delete $INSTDIR\skins\silver\textures\status_voice.tga
Delete $INSTDIR\skins\silver\textures\status_voice.tga
Delete $INSTDIR\skins\silver\textures\status_search.tga
Delete $INSTDIR\skins\silver\textures\status_search.tga
Delete $INSTDIR\skins\silver\textures\status_no_voice.tga
Delete $INSTDIR\skins\silver\textures\status_no_voice.tga
Delete $INSTDIR\skins\silver\textures\status_no_scripts.tga
Delete $INSTDIR\skins\silver\textures\status_no_scripts.tga
Delete $INSTDIR\skins\silver\textures\status_no_push.tga
Delete $INSTDIR\skins\silver\textures\status_no_push.tga
Delete $INSTDIR\skins\silver\textures\status_no_fly.tga
Delete $INSTDIR\skins\silver\textures\status_no_fly.tga
Delete $INSTDIR\skins\silver\textures\status_no_build.tga
Delete $INSTDIR\skins\silver\textures\status_no_build.tga
Delete $INSTDIR\skins\silver\textures\status_health.tga
Delete $INSTDIR\skins\silver\textures\status_health.tga
Delete $INSTDIR\skins\silver\textures\status_buy_land_pressed.tga
Delete $INSTDIR\skins\silver\textures\status_buy_land_pressed.tga
Delete $INSTDIR\skins\silver\textures\status_buy_land.tga
Delete $INSTDIR\skins\silver\textures\status_buy_land.tga
Delete $INSTDIR\skins\silver\textures\status_buy_currency_pressed.tga
Delete $INSTDIR\skins\silver\textures\status_buy_currency_pressed.tga
Delete $INSTDIR\skins\silver\textures\status_buy_currency.tga
Delete $INSTDIR\skins\silver\textures\status_buy_currency.tga
Delete $INSTDIR\skins\silver\textures\startup_logo.j2c
Delete $INSTDIR\skins\silver\textures\startup_logo.j2c
Delete $INSTDIR\skins\silver\textures\square_btn_selected_32x128.tga
Delete $INSTDIR\skins\silver\textures\square_btn_selected_32x128.tga
Delete $INSTDIR\skins\silver\textures\square_btn_32x128.tga
Delete $INSTDIR\skins\silver\textures\square_btn_32x128.tga
Delete $INSTDIR\skins\silver\textures\spin_up_out_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_up_out_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_up_in_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_up_in_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_down_out_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_down_out_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_down_in_blue.tga
Delete $INSTDIR\skins\silver\textures\spin_down_in_blue.tga
Delete $INSTDIR\skins\silver\textures\sm_rounded_corners_simple.tga
Delete $INSTDIR\skins\silver\textures\sm_rounded_corners_simple.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_up_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_up_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_up_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_up_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_right_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_right_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_right_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_right_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_left_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_left_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_left_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_left_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_down_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_down_out_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_down_in_blue.tga
Delete $INSTDIR\skins\silver\textures\scrollbutton_down_in_blue.tga
Delete $INSTDIR\skins\silver\textures\rounded_square_soft.tga
Delete $INSTDIR\skins\silver\textures\rounded_square_soft.tga
Delete $INSTDIR\skins\silver\textures\rounded_square.tga
Delete $INSTDIR\skins\silver\textures\rounded_square.tga
Delete $INSTDIR\skins\silver\textures\restore_pressed.tga
Delete $INSTDIR\skins\silver\textures\restore_pressed.tga
Delete $INSTDIR\skins\silver\textures\restore_inactive.tga
Delete $INSTDIR\skins\silver\textures\restore_inactive.tga
Delete $INSTDIR\skins\silver\textures\restore.tga
Delete $INSTDIR\skins\silver\textures\restore.tga
Delete $INSTDIR\skins\silver\textures\resize_handle_bottom_right_blue.tga
Delete $INSTDIR\skins\silver\textures\resize_handle_bottom_right_blue.tga
Delete $INSTDIR\skins\silver\textures\radio_inactive_true.tga
Delete $INSTDIR\skins\silver\textures\radio_inactive_true.tga
Delete $INSTDIR\skins\silver\textures\radio_inactive_false.tga
Delete $INSTDIR\skins\silver\textures\radio_inactive_false.tga
Delete $INSTDIR\skins\silver\textures\radio_active_true.tga
Delete $INSTDIR\skins\silver\textures\radio_active_true.tga
Delete $INSTDIR\skins\silver\textures\radio_active_false.tga
Delete $INSTDIR\skins\silver\textures\radio_active_false.tga
Delete $INSTDIR\skins\silver\textures\ptt_lock_on.tga
Delete $INSTDIR\skins\silver\textures\ptt_lock_on.tga
Delete $INSTDIR\skins\silver\textures\ptt_lock_off.tga
Delete $INSTDIR\skins\silver\textures\ptt_lock_off.tga
Delete $INSTDIR\skins\silver\textures\progressbar_track.tga
Delete $INSTDIR\skins\silver\textures\progressbar_track.tga
Delete $INSTDIR\skins\silver\textures\progressbar_fill.tga
Delete $INSTDIR\skins\silver\textures\progressbar_fill.tga
Delete $INSTDIR\skins\silver\textures\object_tube_active.tga
Delete $INSTDIR\skins\silver\textures\object_tube_active.tga
Delete $INSTDIR\skins\silver\textures\object_tube.tga
Delete $INSTDIR\skins\silver\textures\object_tube.tga
Delete $INSTDIR\skins\silver\textures\object_tree_active.tga
Delete $INSTDIR\skins\silver\textures\object_tree_active.tga
Delete $INSTDIR\skins\silver\textures\object_tree.tga
Delete $INSTDIR\skins\silver\textures\object_tree.tga
Delete $INSTDIR\skins\silver\textures\object_torus_active.tga
Delete $INSTDIR\skins\silver\textures\object_torus_active.tga
Delete $INSTDIR\skins\silver\textures\object_torus.tga
Delete $INSTDIR\skins\silver\textures\object_torus.tga
Delete $INSTDIR\skins\silver\textures\object_tetrahedron_active.tga
Delete $INSTDIR\skins\silver\textures\object_tetrahedron_active.tga
Delete $INSTDIR\skins\silver\textures\object_tetrahedron.tga
Delete $INSTDIR\skins\silver\textures\object_tetrahedron.tga
Delete $INSTDIR\skins\silver\textures\object_sphere_active.tga
Delete $INSTDIR\skins\silver\textures\object_sphere_active.tga
Delete $INSTDIR\skins\silver\textures\object_sphere.tga
Delete $INSTDIR\skins\silver\textures\object_sphere.tga
Delete $INSTDIR\skins\silver\textures\object_ring_active.tga
Delete $INSTDIR\skins\silver\textures\object_ring_active.tga
Delete $INSTDIR\skins\silver\textures\object_ring.tga
Delete $INSTDIR\skins\silver\textures\object_ring.tga
Delete $INSTDIR\skins\silver\textures\object_pyramid_active.tga
Delete $INSTDIR\skins\silver\textures\object_pyramid_active.tga
Delete $INSTDIR\skins\silver\textures\object_pyramid.tga
Delete $INSTDIR\skins\silver\textures\object_pyramid.tga
Delete $INSTDIR\skins\silver\textures\object_prism_active.tga
Delete $INSTDIR\skins\silver\textures\object_prism_active.tga
Delete $INSTDIR\skins\silver\textures\object_prism.tga
Delete $INSTDIR\skins\silver\textures\object_prism.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_sphere_active.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_sphere_active.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_sphere.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_sphere.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cylinder_active.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cylinder_active.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cylinder.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cylinder.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cone_active.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cone_active.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cone.tga
Delete $INSTDIR\skins\silver\textures\object_hemi_cone.tga
Delete $INSTDIR\skins\silver\textures\object_grass_active.tga
Delete $INSTDIR\skins\silver\textures\object_grass_active.tga
Delete $INSTDIR\skins\silver\textures\object_grass.tga
Delete $INSTDIR\skins\silver\textures\object_grass.tga
Delete $INSTDIR\skins\silver\textures\object_cylinder_active.tga
Delete $INSTDIR\skins\silver\textures\object_cylinder_active.tga
Delete $INSTDIR\skins\silver\textures\object_cylinder.tga
Delete $INSTDIR\skins\silver\textures\object_cylinder.tga
Delete $INSTDIR\skins\silver\textures\object_cube_active.tga
Delete $INSTDIR\skins\silver\textures\object_cube_active.tga
Delete $INSTDIR\skins\silver\textures\object_cube.tga
Delete $INSTDIR\skins\silver\textures\object_cube.tga
Delete $INSTDIR\skins\silver\textures\object_cone_active.tga
Delete $INSTDIR\skins\silver\textures\object_cone_active.tga
Delete $INSTDIR\skins\silver\textures\object_cone.tga
Delete $INSTDIR\skins\silver\textures\object_cone.tga
Delete $INSTDIR\skins\silver\textures\notify_next.png
Delete $INSTDIR\skins\silver\textures\notify_next.png
Delete $INSTDIR\skins\silver\textures\minimize_pressed.tga
Delete $INSTDIR\skins\silver\textures\minimize_pressed.tga
Delete $INSTDIR\skins\silver\textures\minimize_inactive.tga
Delete $INSTDIR\skins\silver\textures\minimize_inactive.tga
Delete $INSTDIR\skins\silver\textures\minimize.tga
Delete $INSTDIR\skins\silver\textures\minimize.tga
Delete $INSTDIR\skins\silver\textures\map_telehub.tga
Delete $INSTDIR\skins\silver\textures\map_telehub.tga
Delete $INSTDIR\skins\silver\textures\map_infohub.tga
Delete $INSTDIR\skins\silver\textures\map_infohub.tga
Delete $INSTDIR\skins\silver\textures\map_home.tga
Delete $INSTDIR\skins\silver\textures\map_home.tga
Delete $INSTDIR\skins\silver\textures\map_event_mature.tga
Delete $INSTDIR\skins\silver\textures\map_event_mature.tga
Delete $INSTDIR\skins\silver\textures\map_event_adult.tga
Delete $INSTDIR\skins\silver\textures\map_event_adult.tga
Delete $INSTDIR\skins\silver\textures\map_event.tga
Delete $INSTDIR\skins\silver\textures\map_event.tga
Delete $INSTDIR\skins\silver\textures\map_avatar_you_8.tga
Delete $INSTDIR\skins\silver\textures\map_avatar_you_8.tga
Delete $INSTDIR\skins\silver\textures\map_avatar_16.tga
Delete $INSTDIR\skins\silver\textures\map_avatar_16.tga
Delete $INSTDIR\skins\silver\textures\lightgray.tga
Delete $INSTDIR\skins\silver\textures\lightgray.tga
Delete $INSTDIR\skins\silver\textures\inv_item_undershirt.tga
Delete $INSTDIR\skins\silver\textures\inv_item_undershirt.tga
Delete $INSTDIR\skins\silver\textures\inv_item_underpants.tga
Delete $INSTDIR\skins\silver\textures\inv_item_underpants.tga
Delete $INSTDIR\skins\silver\textures\inv_item_texture.tga
Delete $INSTDIR\skins\silver\textures\inv_item_texture.tga
Delete $INSTDIR\skins\silver\textures\inv_item_sound.tga
Delete $INSTDIR\skins\silver\textures\inv_item_sound.tga
Delete $INSTDIR\skins\silver\textures\inv_item_socks.tga
Delete $INSTDIR\skins\silver\textures\inv_item_socks.tga
Delete $INSTDIR\skins\silver\textures\inv_item_snapshot.tga
Delete $INSTDIR\skins\silver\textures\inv_item_snapshot.tga
Delete $INSTDIR\skins\silver\textures\inv_item_skirt.tga
Delete $INSTDIR\skins\silver\textures\inv_item_skirt.tga
Delete $INSTDIR\skins\silver\textures\inv_item_skin.tga
Delete $INSTDIR\skins\silver\textures\inv_item_skin.tga
Delete $INSTDIR\skins\silver\textures\inv_item_shoes.tga
Delete $INSTDIR\skins\silver\textures\inv_item_shoes.tga
Delete $INSTDIR\skins\silver\textures\inv_item_shirt.tga
Delete $INSTDIR\skins\silver\textures\inv_item_shirt.tga
Delete $INSTDIR\skins\silver\textures\inv_item_shape.tga
Delete $INSTDIR\skins\silver\textures\inv_item_shape.tga
Delete $INSTDIR\skins\silver\textures\inv_item_script.tga
Delete $INSTDIR\skins\silver\textures\inv_item_script.tga
Delete $INSTDIR\skins\silver\textures\inv_item_pants.tga
Delete $INSTDIR\skins\silver\textures\inv_item_pants.tga
Delete $INSTDIR\skins\silver\textures\inv_item_object_multi.tga
Delete $INSTDIR\skins\silver\textures\inv_item_object_multi.tga
Delete $INSTDIR\skins\silver\textures\inv_item_object.tga
Delete $INSTDIR\skins\silver\textures\inv_item_object.tga
Delete $INSTDIR\skins\silver\textures\inv_item_notecard.tga
Delete $INSTDIR\skins\silver\textures\inv_item_notecard.tga
Delete $INSTDIR\skins\silver\textures\inv_item_landmark_visited.tga
Delete $INSTDIR\skins\silver\textures\inv_item_landmark_visited.tga
Delete $INSTDIR\skins\silver\textures\inv_item_landmark.tga
Delete $INSTDIR\skins\silver\textures\inv_item_landmark.tga
Delete $INSTDIR\skins\silver\textures\inv_item_jacket.tga
Delete $INSTDIR\skins\silver\textures\inv_item_jacket.tga
Delete $INSTDIR\skins\silver\textures\inv_item_hair.tga
Delete $INSTDIR\skins\silver\textures\inv_item_hair.tga
Delete $INSTDIR\skins\silver\textures\inv_item_gloves.tga
Delete $INSTDIR\skins\silver\textures\inv_item_gloves.tga
Delete $INSTDIR\skins\silver\textures\inv_item_gesture.tga
Delete $INSTDIR\skins\silver\textures\inv_item_gesture.tga
Delete $INSTDIR\skins\silver\textures\inv_item_eyes.tga
Delete $INSTDIR\skins\silver\textures\inv_item_eyes.tga
Delete $INSTDIR\skins\silver\textures\inv_item_clothing.tga
Delete $INSTDIR\skins\silver\textures\inv_item_clothing.tga
Delete $INSTDIR\skins\silver\textures\inv_item_callingcard_online.tga
Delete $INSTDIR\skins\silver\textures\inv_item_callingcard_online.tga
Delete $INSTDIR\skins\silver\textures\inv_item_callingcard_offline.tga
Delete $INSTDIR\skins\silver\textures\inv_item_callingcard_offline.tga
Delete $INSTDIR\skins\silver\textures\inv_item_attach.tga
Delete $INSTDIR\skins\silver\textures\inv_item_attach.tga
Delete $INSTDIR\skins\silver\textures\inv_item_animation.tga
Delete $INSTDIR\skins\silver\textures\inv_item_animation.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_trash.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_trash.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_texture.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_texture.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_sound.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_sound.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_snapshot.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_snapshot.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_script.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_script.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_plain_open.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_plain_open.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_plain_closed.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_plain_closed.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_object.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_object.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_notecard.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_notecard.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_lostandfound.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_lostandfound.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_landmark.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_landmark.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_gesture.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_gesture.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_clothing.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_clothing.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_callingcard.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_callingcard.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_bodypart.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_bodypart.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_animation.tga
Delete $INSTDIR\skins\silver\textures\inv_folder_animation.tga
Delete $INSTDIR\skins\silver\textures\icon_top_pick.tga
Delete $INSTDIR\skins\silver\textures\icon_top_pick.tga
Delete $INSTDIR\skins\silver\textures\icon_lock.tga
Delete $INSTDIR\skins\silver\textures\icon_lock.tga
Delete $INSTDIR\skins\silver\textures\icon_groupnoticeinventory.tga
Delete $INSTDIR\skins\silver\textures\icon_groupnoticeinventory.tga
Delete $INSTDIR\skins\silver\textures\icon_groupnotice.tga
Delete $INSTDIR\skins\silver\textures\icon_groupnotice.tga
Delete $INSTDIR\skins\silver\textures\icon_group.tga
Delete $INSTDIR\skins\silver\textures\icon_group.tga
Delete $INSTDIR\skins\silver\textures\icon_for_sale.tga
Delete $INSTDIR\skins\silver\textures\icon_for_sale.tga
Delete $INSTDIR\skins\silver\textures\icon_event_mature.tga
Delete $INSTDIR\skins\silver\textures\icon_event_mature.tga
Delete $INSTDIR\skins\silver\textures\icon_event.tga
Delete $INSTDIR\skins\silver\textures\icon_event.tga
Delete $INSTDIR\skins\silver\textures\icon_day_cycle.tga
Delete $INSTDIR\skins\silver\textures\icon_day_cycle.tga
Delete $INSTDIR\skins\silver\textures\icon_avatar_online.tga
Delete $INSTDIR\skins\silver\textures\icon_avatar_online.tga
Delete $INSTDIR\skins\silver\textures\icon_avatar_offline.tga
Delete $INSTDIR\skins\silver\textures\icon_avatar_offline.tga
Delete $INSTDIR\skins\silver\textures\icon_auction.tga
Delete $INSTDIR\skins\silver\textures\icon_auction.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on-lvl3.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on-lvl3.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on-lvl2.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on-lvl2.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on-lvl1.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-on-lvl1.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-off.tga
Delete $INSTDIR\skins\silver\textures\icn_voice_ptt-off.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-pvtfocus.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-pvtfocus.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-localchat.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-localchat.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-groupfocus.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-groupfocus.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-call-start.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-call-start.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-call-end.tga
Delete $INSTDIR\skins\silver\textures\icn_voice-call-end.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_snapshot.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_snapshot.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_search.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_search.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_minimap.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_minimap.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_map.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_map.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_inventory.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_inventory.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_fly.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_fly.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_build.tga
Delete $INSTDIR\skins\silver\textures\icn_toolbar_build.tga
Delete $INSTDIR\skins\silver\textures\icn_stop.tga
Delete $INSTDIR\skins\silver\textures\icn_stop.tga
Delete $INSTDIR\skins\silver\textures\icn_speaker_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_speaker_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_speaker-muted_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_speaker-muted_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_slide-thumb_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_slide-thumb_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_slide-highlight.tga
Delete $INSTDIR\skins\silver\textures\icn_slide-highlight.tga
Delete $INSTDIR\skins\silver\textures\icn_slide-groove_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_slide-groove_dark.tga
Delete $INSTDIR\skins\silver\textures\icn_rounded-text-field.tga
Delete $INSTDIR\skins\silver\textures\icn_rounded-text-field.tga
Delete $INSTDIR\skins\silver\textures\icn_play.tga
Delete $INSTDIR\skins\silver\textures\icn_play.tga
Delete $INSTDIR\skins\silver\textures\icn_pause.tga
Delete $INSTDIR\skins\silver\textures\icn_pause.tga
Delete $INSTDIR\skins\silver\textures\icn_music-play.tga
Delete $INSTDIR\skins\silver\textures\icn_music-play.tga
Delete $INSTDIR\skins\silver\textures\icn_music-pause.tga
Delete $INSTDIR\skins\silver\textures\icn_music-pause.tga
Delete $INSTDIR\skins\silver\textures\icn_media_web.tga
Delete $INSTDIR\skins\silver\textures\icn_media_web.tga
Delete $INSTDIR\skins\silver\textures\icn_media_movie.tga
Delete $INSTDIR\skins\silver\textures\icn_media_movie.tga
Delete $INSTDIR\skins\silver\textures\icn_media-play.tga
Delete $INSTDIR\skins\silver\textures\icn_media-play.tga
Delete $INSTDIR\skins\silver\textures\icn_media-pause.tga
Delete $INSTDIR\skins\silver\textures\icn_media-pause.tga
Delete $INSTDIR\skins\silver\textures\icn_label_web.tga
Delete $INSTDIR\skins\silver\textures\icn_label_web.tga
Delete $INSTDIR\skins\silver\textures\icn_label_music.tga
Delete $INSTDIR\skins\silver\textures\icn_label_music.tga
Delete $INSTDIR\skins\silver\textures\icn_label_media.tga
Delete $INSTDIR\skins\silver\textures\icn_label_media.tga
Delete $INSTDIR\skins\silver\textures\icn_chatbar.tga
Delete $INSTDIR\skins\silver\textures\icn_chatbar.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_right_selected.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_right_selected.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_right_disabled.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_right_disabled.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_right.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_right.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_left_selected.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_left_selected.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_left_disabled.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_left_disabled.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_left.tga
Delete $INSTDIR\skins\silver\textures\flyout_btn_left.tga
Delete $INSTDIR\skins\silver\textures\ff_visible_online_button.tga
Delete $INSTDIR\skins\silver\textures\ff_visible_online_button.tga
Delete $INSTDIR\skins\silver\textures\ff_visible_map_button.tga
Delete $INSTDIR\skins\silver\textures\ff_visible_map_button.tga
Delete $INSTDIR\skins\silver\textures\ff_online_status_button.tga
Delete $INSTDIR\skins\silver\textures\ff_online_status_button.tga
Delete $INSTDIR\skins\silver\textures\ff_edit_theirs_button.tga
Delete $INSTDIR\skins\silver\textures\ff_edit_theirs_button.tga
Delete $INSTDIR\skins\silver\textures\ff_edit_mine_button.tga
Delete $INSTDIR\skins\silver\textures\ff_edit_mine_button.tga
Delete $INSTDIR\skins\silver\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
Delete $INSTDIR\skins\silver\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
Delete $INSTDIR\skins\silver\textures\eye_button_inactive.tga
Delete $INSTDIR\skins\silver\textures\eye_button_inactive.tga
Delete $INSTDIR\skins\silver\textures\eye_button_active.tga
Delete $INSTDIR\skins\silver\textures\eye_button_active.tga
Delete $INSTDIR\skins\silver\textures\down_arrow.png
Delete $INSTDIR\skins\silver\textures\down_arrow.png
Delete $INSTDIR\skins\silver\textures\darkgray.tga
Delete $INSTDIR\skins\silver\textures\darkgray.tga
Delete $INSTDIR\skins\silver\textures\combobox_arrow.tga
Delete $INSTDIR\skins\silver\textures\combobox_arrow.tga
Delete $INSTDIR\skins\silver\textures\closebox.tga
Delete $INSTDIR\skins\silver\textures\closebox.tga
Delete $INSTDIR\skins\silver\textures\close_inactive_blue.tga
Delete $INSTDIR\skins\silver\textures\close_inactive_blue.tga
Delete $INSTDIR\skins\silver\textures\close_in_blue.tga
Delete $INSTDIR\skins\silver\textures\close_in_blue.tga
Delete $INSTDIR\skins\silver\textures\checkbox_enabled_true.tga
Delete $INSTDIR\skins\silver\textures\checkbox_enabled_true.tga
Delete $INSTDIR\skins\silver\textures\checkbox_enabled_false.tga
Delete $INSTDIR\skins\silver\textures\checkbox_enabled_false.tga
Delete $INSTDIR\skins\silver\textures\checkbox_disabled_true.tga
Delete $INSTDIR\skins\silver\textures\checkbox_disabled_true.tga
Delete $INSTDIR\skins\silver\textures\checkbox_disabled_false.tga
Delete $INSTDIR\skins\silver\textures\checkbox_disabled_false.tga
Delete $INSTDIR\skins\silver\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
Delete $INSTDIR\skins\silver\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
Delete $INSTDIR\skins\silver\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
Delete $INSTDIR\skins\silver\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
Delete $INSTDIR\skins\silver\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
Delete $INSTDIR\skins\silver\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
Delete $INSTDIR\skins\silver\textures\button_enabled_selected_32x128.tga
Delete $INSTDIR\skins\silver\textures\button_enabled_selected_32x128.tga
Delete $INSTDIR\skins\silver\textures\button_enabled_32x128.tga
Delete $INSTDIR\skins\silver\textures\button_enabled_32x128.tga
Delete $INSTDIR\skins\silver\textures\button_disabled_32x128.tga
Delete $INSTDIR\skins\silver\textures\button_disabled_32x128.tga
Delete $INSTDIR\skins\silver\textures\button_anim_stop_selected.tga
Delete $INSTDIR\skins\silver\textures\button_anim_stop_selected.tga
Delete $INSTDIR\skins\silver\textures\button_anim_stop.tga
Delete $INSTDIR\skins\silver\textures\button_anim_stop.tga
Delete $INSTDIR\skins\silver\textures\button_anim_play_selected.tga
Delete $INSTDIR\skins\silver\textures\button_anim_play_selected.tga
Delete $INSTDIR\skins\silver\textures\button_anim_play.tga
Delete $INSTDIR\skins\silver\textures\button_anim_play.tga
Delete $INSTDIR\skins\silver\textures\button_anim_pause_selected.tga
Delete $INSTDIR\skins\silver\textures\button_anim_pause_selected.tga
Delete $INSTDIR\skins\silver\textures\button_anim_pause.tga
Delete $INSTDIR\skins\silver\textures\button_anim_pause.tga
Delete $INSTDIR\skins\silver\textures\btn_chatbar_selected.tga
Delete $INSTDIR\skins\silver\textures\btn_chatbar_selected.tga
Delete $INSTDIR\skins\silver\textures\btn_chatbar.tga
Delete $INSTDIR\skins\silver\textures\btn_chatbar.tga
Delete $INSTDIR\skins\silver\textures\black.tga
Delete $INSTDIR\skins\silver\textures\black.tga
Delete $INSTDIR\skins\silver\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
Delete $INSTDIR\skins\silver\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
Delete $INSTDIR\skins\silver\textures\arrow_up.tga
Delete $INSTDIR\skins\silver\textures\arrow_up.tga
Delete $INSTDIR\skins\silver\textures\arrow_down.tga
Delete $INSTDIR\skins\silver\textures\arrow_down.tga
Delete $INSTDIR\skins\silver\textures\active_voice_tab.tga
Delete $INSTDIR\skins\silver\textures\active_voice_tab.tga
Delete $INSTDIR\skins\silver\textures\active_speakers.tga
Delete $INSTDIR\skins\silver\textures\active_speakers.tga
Delete $INSTDIR\skins\silver\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
Delete $INSTDIR\skins\silver\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
Delete $INSTDIR\skins\silver\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
Delete $INSTDIR\skins\silver\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
Delete $INSTDIR\skins\silver\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
Delete $INSTDIR\skins\silver\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
Delete $INSTDIR\skins\silver\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
Delete $INSTDIR\skins\silver\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
Delete $INSTDIR\skins\silver\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
Delete $INSTDIR\skins\silver\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
Delete $INSTDIR\skins\default\textures\white.tga
Delete $INSTDIR\skins\default\textures\white.tga
Delete $INSTDIR\skins\default\textures\uv_test2.tga
Delete $INSTDIR\skins\default\textures\uv_test2.tga
Delete $INSTDIR\skins\default\textures\uv_test1.j2c
Delete $INSTDIR\skins\default\textures\uv_test1.j2c
Delete $INSTDIR\skins\default\textures\up_arrow.tga
Delete $INSTDIR\skins\default\textures\up_arrow.tga
Delete $INSTDIR\skins\default\textures\up_arrow.png
Delete $INSTDIR\skins\default\textures\up_arrow.png
Delete $INSTDIR\skins\default\textures\toolbar_tab.tga
Delete $INSTDIR\skins\default\textures\toolbar_tab.tga
Delete $INSTDIR\skins\default\textures\toolbar_btn_selected.tga
Delete $INSTDIR\skins\default\textures\toolbar_btn_selected.tga
Delete $INSTDIR\skins\default\textures\toolbar_btn_enabled.tga
Delete $INSTDIR\skins\default\textures\toolbar_btn_enabled.tga
Delete $INSTDIR\skins\default\textures\toolbar_btn_disabled.tga
Delete $INSTDIR\skins\default\textures\toolbar_btn_disabled.tga
Delete $INSTDIR\skins\default\textures\toolbar_bg.tga
Delete $INSTDIR\skins\default\textures\toolbar_bg.tga
Delete $INSTDIR\skins\default\textures\tool_zoom_active.tga
Delete $INSTDIR\skins\default\textures\tool_zoom_active.tga
Delete $INSTDIR\skins\default\textures\tool_zoom.tga
Delete $INSTDIR\skins\default\textures\tool_zoom.tga
Delete $INSTDIR\skins\default\textures\tool_dozer_active.tga
Delete $INSTDIR\skins\default\textures\tool_dozer_active.tga
Delete $INSTDIR\skins\default\textures\tool_dozer.tga
Delete $INSTDIR\skins\default\textures\tool_dozer.tga
Delete $INSTDIR\skins\default\textures\textures.xml
Delete $INSTDIR\skins\default\textures\textures.xml
Delete $INSTDIR\skins\default\textures\tearoffbox.tga
Delete $INSTDIR\skins\default\textures\tearoffbox.tga
Delete $INSTDIR\skins\default\textures\tearoff_pressed.tga
Delete $INSTDIR\skins\default\textures\tearoff_pressed.tga
Delete $INSTDIR\skins\default\textures\tabarea.tga
Delete $INSTDIR\skins\default\textures\tabarea.tga
Delete $INSTDIR\skins\default\textures\tab_top_selected_blue.tga
Delete $INSTDIR\skins\default\textures\tab_top_selected_blue.tga
Delete $INSTDIR\skins\default\textures\tab_top_blue.tga
Delete $INSTDIR\skins\default\textures\tab_top_blue.tga
Delete $INSTDIR\skins\default\textures\tab_left_selected.tga
Delete $INSTDIR\skins\default\textures\tab_left_selected.tga
Delete $INSTDIR\skins\default\textures\tab_left.tga
Delete $INSTDIR\skins\default\textures\tab_left.tga
Delete $INSTDIR\skins\default\textures\tab_bottom_selected_blue.tga
Delete $INSTDIR\skins\default\textures\tab_bottom_selected_blue.tga
Delete $INSTDIR\skins\default\textures\tab_bottom_blue.tga
Delete $INSTDIR\skins\default\textures\tab_bottom_blue.tga
Delete $INSTDIR\skins\default\textures\tab_background_purple.tga
Delete $INSTDIR\skins\default\textures\tab_background_purple.tga
Delete $INSTDIR\skins\default\textures\tab_background_lightgrey.tga
Delete $INSTDIR\skins\default\textures\tab_background_lightgrey.tga
Delete $INSTDIR\skins\default\textures\tab_background_darkpurple.tga
Delete $INSTDIR\skins\default\textures\tab_background_darkpurple.tga
Delete $INSTDIR\skins\default\textures\status_voice.tga
Delete $INSTDIR\skins\default\textures\status_voice.tga
Delete $INSTDIR\skins\default\textures\status_search_btn_pressed.png
Delete $INSTDIR\skins\default\textures\status_search_btn_pressed.png
Delete $INSTDIR\skins\default\textures\status_search_btn.png
Delete $INSTDIR\skins\default\textures\status_search_btn.png
Delete $INSTDIR\skins\default\textures\status_search.tga
Delete $INSTDIR\skins\default\textures\status_search.tga
Delete $INSTDIR\skins\default\textures\status_script_debug.tga
Delete $INSTDIR\skins\default\textures\status_script_debug.tga
Delete $INSTDIR\skins\default\textures\status_no_voice.tga
Delete $INSTDIR\skins\default\textures\status_no_voice.tga
Delete $INSTDIR\skins\default\textures\status_no_scripts.tga
Delete $INSTDIR\skins\default\textures\status_no_scripts.tga
Delete $INSTDIR\skins\default\textures\status_no_push.tga
Delete $INSTDIR\skins\default\textures\status_no_push.tga
Delete $INSTDIR\skins\default\textures\status_no_fly.tga
Delete $INSTDIR\skins\default\textures\status_no_fly.tga
Delete $INSTDIR\skins\default\textures\status_no_build.tga
Delete $INSTDIR\skins\default\textures\status_no_build.tga
Delete $INSTDIR\skins\default\textures\status_money.tga
Delete $INSTDIR\skins\default\textures\status_money.tga
Delete $INSTDIR\skins\default\textures\status_health.tga
Delete $INSTDIR\skins\default\textures\status_health.tga
Delete $INSTDIR\skins\default\textures\status_buy_land_pressed.tga
Delete $INSTDIR\skins\default\textures\status_buy_land_pressed.tga
Delete $INSTDIR\skins\default\textures\status_buy_land.tga
Delete $INSTDIR\skins\default\textures\status_buy_land.tga
Delete $INSTDIR\skins\default\textures\status_buy_currency_pressed.tga
Delete $INSTDIR\skins\default\textures\status_buy_currency_pressed.tga
Delete $INSTDIR\skins\default\textures\status_buy_currency.tga
Delete $INSTDIR\skins\default\textures\status_buy_currency.tga
Delete $INSTDIR\skins\default\textures\status_busy.tga
Delete $INSTDIR\skins\default\textures\status_busy.tga
Delete $INSTDIR\skins\default\textures\startup_logo.j2c
Delete $INSTDIR\skins\default\textures\startup_logo.j2c
Delete $INSTDIR\skins\default\textures\square_btn_selected_32x128.tga
Delete $INSTDIR\skins\default\textures\square_btn_selected_32x128.tga
Delete $INSTDIR\skins\default\textures\square_btn_32x128.tga
Delete $INSTDIR\skins\default\textures\square_btn_32x128.tga
Delete $INSTDIR\skins\default\textures\spin_up_out_blue.tga
Delete $INSTDIR\skins\default\textures\spin_up_out_blue.tga
Delete $INSTDIR\skins\default\textures\spin_up_in_blue.tga
Delete $INSTDIR\skins\default\textures\spin_up_in_blue.tga
Delete $INSTDIR\skins\default\textures\spin_down_out_blue.tga
Delete $INSTDIR\skins\default\textures\spin_down_out_blue.tga
Delete $INSTDIR\skins\default\textures\spin_down_in_blue.tga
Delete $INSTDIR\skins\default\textures\spin_down_in_blue.tga
Delete $INSTDIR\skins\default\textures\spacer35.tga
Delete $INSTDIR\skins\default\textures\spacer35.tga
Delete $INSTDIR\skins\default\textures\spacer24.tga
Delete $INSTDIR\skins\default\textures\spacer24.tga
Delete $INSTDIR\skins\default\textures\smicon_warn.tga
Delete $INSTDIR\skins\default\textures\smicon_warn.tga
Delete $INSTDIR\skins\default\textures\sm_rounded_corners_simple.tga
Delete $INSTDIR\skins\default\textures\sm_rounded_corners_simple.tga
Delete $INSTDIR\skins\default\textures\slim_icon_16_viewer.tga
Delete $INSTDIR\skins\default\textures\slim_icon_16_viewer.tga
Delete $INSTDIR\skins\default\textures\skin_thumbnail_silver.png
Delete $INSTDIR\skins\default\textures\skin_thumbnail_silver.png
Delete $INSTDIR\skins\default\textures\skin_thumbnail_default.png
Delete $INSTDIR\skins\default\textures\skin_thumbnail_default.png
Delete $INSTDIR\skins\default\textures\silhouette.j2c
Delete $INSTDIR\skins\default\textures\silhouette.j2c
Delete $INSTDIR\skins\default\textures\scrollbutton_up_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_up_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_up_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_up_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_right_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_right_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_right_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_right_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_left_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_left_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_left_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_left_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_down_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_down_out_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_down_in_blue.tga
Delete $INSTDIR\skins\default\textures\scrollbutton_down_in_blue.tga
Delete $INSTDIR\skins\default\textures\script_error.j2c
Delete $INSTDIR\skins\default\textures\script_error.j2c
Delete $INSTDIR\skins\default\textures\rounded_square_soft.j2c
Delete $INSTDIR\skins\default\textures\rounded_square_soft.j2c
Delete $INSTDIR\skins\default\textures\rounded_square.j2c
Delete $INSTDIR\skins\default\textures\rounded_square.j2c
Delete $INSTDIR\skins\default\textures\restore_pressed.tga
Delete $INSTDIR\skins\default\textures\restore_pressed.tga
Delete $INSTDIR\skins\default\textures\restore_inactive.tga
Delete $INSTDIR\skins\default\textures\restore_inactive.tga
Delete $INSTDIR\skins\default\textures\restore.tga
Delete $INSTDIR\skins\default\textures\restore.tga
Delete $INSTDIR\skins\default\textures\resize_handle_bottom_right_blue.tga
Delete $INSTDIR\skins\default\textures\resize_handle_bottom_right_blue.tga
Delete $INSTDIR\skins\default\textures\radio_inactive_true.tga
Delete $INSTDIR\skins\default\textures\radio_inactive_true.tga
Delete $INSTDIR\skins\default\textures\radio_inactive_false.tga
Delete $INSTDIR\skins\default\textures\radio_inactive_false.tga
Delete $INSTDIR\skins\default\textures\radio_active_true.tga
Delete $INSTDIR\skins\default\textures\radio_active_true.tga
Delete $INSTDIR\skins\default\textures\radio_active_false.tga
Delete $INSTDIR\skins\default\textures\radio_active_false.tga
Delete $INSTDIR\skins\default\textures\ptt_lock_on.tga
Delete $INSTDIR\skins\default\textures\ptt_lock_on.tga
Delete $INSTDIR\skins\default\textures\ptt_lock_off.tga
Delete $INSTDIR\skins\default\textures\ptt_lock_off.tga
Delete $INSTDIR\skins\default\textures\propertyline.tga
Delete $INSTDIR\skins\default\textures\propertyline.tga
Delete $INSTDIR\skins\default\textures\progressbar_track.tga
Delete $INSTDIR\skins\default\textures\progressbar_track.tga
Delete $INSTDIR\skins\default\textures\progressbar_fill.tga
Delete $INSTDIR\skins\default\textures\progressbar_fill.tga
Delete $INSTDIR\skins\default\textures\progress_fill.tga
Delete $INSTDIR\skins\default\textures\progress_fill.tga
Delete $INSTDIR\skins\default\textures\pos_lock_on.tga
Delete $INSTDIR\skins\default\textures\pos_lock_on.tga
Delete $INSTDIR\skins\default\textures\pos_lock_off.tga
Delete $INSTDIR\skins\default\textures\pos_lock_off.tga
Delete $INSTDIR\skins\default\textures\pixiesmall.j2c
Delete $INSTDIR\skins\default\textures\pixiesmall.j2c
Delete $INSTDIR\skins\default\textures\object_tube_active.tga
Delete $INSTDIR\skins\default\textures\object_tube_active.tga
Delete $INSTDIR\skins\default\textures\object_tube.tga
Delete $INSTDIR\skins\default\textures\object_tube.tga
Delete $INSTDIR\skins\default\textures\object_tree_active.tga
Delete $INSTDIR\skins\default\textures\object_tree_active.tga
Delete $INSTDIR\skins\default\textures\object_tree.tga
Delete $INSTDIR\skins\default\textures\object_tree.tga
Delete $INSTDIR\skins\default\textures\object_torus_active.tga
Delete $INSTDIR\skins\default\textures\object_torus_active.tga
Delete $INSTDIR\skins\default\textures\object_torus.tga
Delete $INSTDIR\skins\default\textures\object_torus.tga
Delete $INSTDIR\skins\default\textures\object_tetrahedron_active.tga
Delete $INSTDIR\skins\default\textures\object_tetrahedron_active.tga
Delete $INSTDIR\skins\default\textures\object_tetrahedron.tga
Delete $INSTDIR\skins\default\textures\object_tetrahedron.tga
Delete $INSTDIR\skins\default\textures\object_sphere_active.tga
Delete $INSTDIR\skins\default\textures\object_sphere_active.tga
Delete $INSTDIR\skins\default\textures\object_sphere.tga
Delete $INSTDIR\skins\default\textures\object_sphere.tga
Delete $INSTDIR\skins\default\textures\object_ring_active.tga
Delete $INSTDIR\skins\default\textures\object_ring_active.tga
Delete $INSTDIR\skins\default\textures\object_ring.tga
Delete $INSTDIR\skins\default\textures\object_ring.tga
Delete $INSTDIR\skins\default\textures\object_pyramid_active.tga
Delete $INSTDIR\skins\default\textures\object_pyramid_active.tga
Delete $INSTDIR\skins\default\textures\object_pyramid.tga
Delete $INSTDIR\skins\default\textures\object_pyramid.tga
Delete $INSTDIR\skins\default\textures\object_prism_active.tga
Delete $INSTDIR\skins\default\textures\object_prism_active.tga
Delete $INSTDIR\skins\default\textures\object_prism.tga
Delete $INSTDIR\skins\default\textures\object_prism.tga
Delete $INSTDIR\skins\default\textures\object_hemi_sphere_active.tga
Delete $INSTDIR\skins\default\textures\object_hemi_sphere_active.tga
Delete $INSTDIR\skins\default\textures\object_hemi_sphere.tga
Delete $INSTDIR\skins\default\textures\object_hemi_sphere.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cylinder_active.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cylinder_active.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cylinder.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cylinder.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cone_active.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cone_active.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cone.tga
Delete $INSTDIR\skins\default\textures\object_hemi_cone.tga
Delete $INSTDIR\skins\default\textures\object_grass_active.tga
Delete $INSTDIR\skins\default\textures\object_grass_active.tga
Delete $INSTDIR\skins\default\textures\object_grass.tga
Delete $INSTDIR\skins\default\textures\object_grass.tga
Delete $INSTDIR\skins\default\textures\object_cylinder_active.tga
Delete $INSTDIR\skins\default\textures\object_cylinder_active.tga
Delete $INSTDIR\skins\default\textures\object_cylinder.tga
Delete $INSTDIR\skins\default\textures\object_cylinder.tga
Delete $INSTDIR\skins\default\textures\object_cube_active.tga
Delete $INSTDIR\skins\default\textures\object_cube_active.tga
Delete $INSTDIR\skins\default\textures\object_cube.tga
Delete $INSTDIR\skins\default\textures\object_cube.tga
Delete $INSTDIR\skins\default\textures\object_cone_active.tga
Delete $INSTDIR\skins\default\textures\object_cone_active.tga
Delete $INSTDIR\skins\default\textures\object_cone.tga
Delete $INSTDIR\skins\default\textures\object_cone.tga
Delete $INSTDIR\skins\default\textures\notify_tip_icon.tga
Delete $INSTDIR\skins\default\textures\notify_tip_icon.tga
Delete $INSTDIR\skins\default\textures\notify_next.png
Delete $INSTDIR\skins\default\textures\notify_next.png
Delete $INSTDIR\skins\default\textures\notify_caution_icon.tga
Delete $INSTDIR\skins\default\textures\notify_caution_icon.tga
Delete $INSTDIR\skins\default\textures\notify_box_icon.tga
Delete $INSTDIR\skins\default\textures\notify_box_icon.tga
Delete $INSTDIR\skins\default\textures\noentrypasslines.j2c
Delete $INSTDIR\skins\default\textures\noentrypasslines.j2c
Delete $INSTDIR\skins\default\textures\noentrylines.j2c
Delete $INSTDIR\skins\default\textures\noentrylines.j2c
Delete $INSTDIR\skins\default\textures\mute_icon.tga
Delete $INSTDIR\skins\default\textures\mute_icon.tga
Delete $INSTDIR\skins\default\textures\music_icon.tga
Delete $INSTDIR\skins\default\textures\music_icon.tga
Delete $INSTDIR\skins\default\textures\move_up_out.tga
Delete $INSTDIR\skins\default\textures\move_up_out.tga
Delete $INSTDIR\skins\default\textures\move_up_in.tga
Delete $INSTDIR\skins\default\textures\move_up_in.tga
Delete $INSTDIR\skins\default\textures\move_turn_right_out.tga
Delete $INSTDIR\skins\default\textures\move_turn_right_out.tga
Delete $INSTDIR\skins\default\textures\move_turn_right_in.tga
Delete $INSTDIR\skins\default\textures\move_turn_right_in.tga
Delete $INSTDIR\skins\default\textures\move_turn_left_out.tga
Delete $INSTDIR\skins\default\textures\move_turn_left_out.tga
Delete $INSTDIR\skins\default\textures\move_turn_left_in.tga
Delete $INSTDIR\skins\default\textures\move_turn_left_in.tga
Delete $INSTDIR\skins\default\textures\move_right_out.tga
Delete $INSTDIR\skins\default\textures\move_right_out.tga
Delete $INSTDIR\skins\default\textures\move_right_in.tga
Delete $INSTDIR\skins\default\textures\move_right_in.tga
Delete $INSTDIR\skins\default\textures\move_left_out.tga
Delete $INSTDIR\skins\default\textures\move_left_out.tga
Delete $INSTDIR\skins\default\textures\move_left_in.tga
Delete $INSTDIR\skins\default\textures\move_left_in.tga
Delete $INSTDIR\skins\default\textures\move_forward_out.tga
Delete $INSTDIR\skins\default\textures\move_forward_out.tga
Delete $INSTDIR\skins\default\textures\move_forward_in.tga
Delete $INSTDIR\skins\default\textures\move_forward_in.tga
Delete $INSTDIR\skins\default\textures\move_down_out.tga
Delete $INSTDIR\skins\default\textures\move_down_out.tga
Delete $INSTDIR\skins\default\textures\move_down_in.tga
Delete $INSTDIR\skins\default\textures\move_down_in.tga
Delete $INSTDIR\skins\default\textures\move_backward_out.tga
Delete $INSTDIR\skins\default\textures\move_backward_out.tga
Delete $INSTDIR\skins\default\textures\move_backward_in.tga
Delete $INSTDIR\skins\default\textures\move_backward_in.tga
Delete $INSTDIR\skins\default\textures\missing_asset.tga
Delete $INSTDIR\skins\default\textures\missing_asset.tga
Delete $INSTDIR\skins\default\textures\minimize_pressed.tga
Delete $INSTDIR\skins\default\textures\minimize_pressed.tga
Delete $INSTDIR\skins\default\textures\minimize_inactive.tga
Delete $INSTDIR\skins\default\textures\minimize_inactive.tga
Delete $INSTDIR\skins\default\textures\minimize.tga
Delete $INSTDIR\skins\default\textures\minimize.tga
Delete $INSTDIR\skins\default\textures\media_panel_scrollbg.png
Delete $INSTDIR\skins\default\textures\media_panel_scrollbg.png
Delete $INSTDIR\skins\default\textures\media_panel_hoverrectangle.png
Delete $INSTDIR\skins\default\textures\media_panel_hoverrectangle.png
Delete $INSTDIR\skins\default\textures\media_panel_divider.png
Delete $INSTDIR\skins\default\textures\media_panel_divider.png
Delete $INSTDIR\skins\default\textures\media_panel_bg.png
Delete $INSTDIR\skins\default\textures\media_panel_bg.png
Delete $INSTDIR\skins\default\textures\media_icon.tga
Delete $INSTDIR\skins\default\textures\media_icon.tga
Delete $INSTDIR\skins\default\textures\media_btn_stoploading.png
Delete $INSTDIR\skins\default\textures\media_btn_stoploading.png
Delete $INSTDIR\skins\default\textures\media_btn_scrollup.png
Delete $INSTDIR\skins\default\textures\media_btn_scrollup.png
Delete $INSTDIR\skins\default\textures\media_btn_scrollright.png
Delete $INSTDIR\skins\default\textures\media_btn_scrollright.png
Delete $INSTDIR\skins\default\textures\media_btn_scrollleft.png
Delete $INSTDIR\skins\default\textures\media_btn_scrollleft.png
Delete $INSTDIR\skins\default\textures\media_btn_scrolldown.png
Delete $INSTDIR\skins\default\textures\media_btn_scrolldown.png
Delete $INSTDIR\skins\default\textures\media_btn_reload.png
Delete $INSTDIR\skins\default\textures\media_btn_reload.png
Delete $INSTDIR\skins\default\textures\media_btn_optimalzoom.png
Delete $INSTDIR\skins\default\textures\media_btn_optimalzoom.png
Delete $INSTDIR\skins\default\textures\media_btn_newwindow.png
Delete $INSTDIR\skins\default\textures\media_btn_newwindow.png
Delete $INSTDIR\skins\default\textures\media_btn_home.png
Delete $INSTDIR\skins\default\textures\media_btn_home.png
Delete $INSTDIR\skins\default\textures\media_btn_forward.png
Delete $INSTDIR\skins\default\textures\media_btn_forward.png
Delete $INSTDIR\skins\default\textures\media_btn_done.png
Delete $INSTDIR\skins\default\textures\media_btn_done.png
Delete $INSTDIR\skins\default\textures\media_btn_back.png
Delete $INSTDIR\skins\default\textures\media_btn_back.png
Delete $INSTDIR\skins\default\textures\map_track_8.tga
Delete $INSTDIR\skins\default\textures\map_track_8.tga
Delete $INSTDIR\skins\default\textures\map_track_16.tga
Delete $INSTDIR\skins\default\textures\map_track_16.tga
Delete $INSTDIR\skins\default\textures\map_telehub.tga
Delete $INSTDIR\skins\default\textures\map_telehub.tga
Delete $INSTDIR\skins\default\textures\map_infohub.tga
Delete $INSTDIR\skins\default\textures\map_infohub.tga
Delete $INSTDIR\skins\default\textures\map_home.tga
Delete $INSTDIR\skins\default\textures\map_home.tga
Delete $INSTDIR\skins\default\textures\map_event_mature.tga
Delete $INSTDIR\skins\default\textures\map_event_mature.tga
Delete $INSTDIR\skins\default\textures\map_event_adult.tga
Delete $INSTDIR\skins\default\textures\map_event_adult.tga
Delete $INSTDIR\skins\default\textures\map_event.tga
Delete $INSTDIR\skins\default\textures\map_event.tga
Delete $INSTDIR\skins\default\textures\map_avatar_you_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_you_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_you_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_you_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_below_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_below_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_below_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_below_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_above_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_above_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_above_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_above_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_8.tga
Delete $INSTDIR\skins\default\textures\map_avatar_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_32.tga
Delete $INSTDIR\skins\default\textures\map_avatar_16.tga
Delete $INSTDIR\skins\default\textures\map_avatar_16.tga
Delete $INSTDIR\skins\default\textures\locked_image.j2c
Delete $INSTDIR\skins\default\textures\locked_image.j2c
Delete $INSTDIR\skins\default\textures\lightgray.tga
Delete $INSTDIR\skins\default\textures\lightgray.tga
Delete $INSTDIR\skins\default\textures\legend.tga
Delete $INSTDIR\skins\default\textures\legend.tga
Delete $INSTDIR\skins\default\textures\lag_status_warning.tga
Delete $INSTDIR\skins\default\textures\lag_status_warning.tga
Delete $INSTDIR\skins\default\textures\lag_status_good.tga
Delete $INSTDIR\skins\default\textures\lag_status_good.tga
Delete $INSTDIR\skins\default\textures\lag_status_critical.tga
Delete $INSTDIR\skins\default\textures\lag_status_critical.tga
Delete $INSTDIR\skins\default\textures\inv_item_undershirt.tga
Delete $INSTDIR\skins\default\textures\inv_item_undershirt.tga
Delete $INSTDIR\skins\default\textures\inv_item_underpants.tga
Delete $INSTDIR\skins\default\textures\inv_item_underpants.tga
Delete $INSTDIR\skins\default\textures\inv_item_texture.tga
Delete $INSTDIR\skins\default\textures\inv_item_texture.tga
Delete $INSTDIR\skins\default\textures\inv_item_tattoo.tga
Delete $INSTDIR\skins\default\textures\inv_item_tattoo.tga
Delete $INSTDIR\skins\default\textures\inv_item_sound.tga
Delete $INSTDIR\skins\default\textures\inv_item_sound.tga
Delete $INSTDIR\skins\default\textures\inv_item_socks.tga
Delete $INSTDIR\skins\default\textures\inv_item_socks.tga
Delete $INSTDIR\skins\default\textures\inv_item_snapshot.tga
Delete $INSTDIR\skins\default\textures\inv_item_snapshot.tga
Delete $INSTDIR\skins\default\textures\inv_item_skirt.tga
Delete $INSTDIR\skins\default\textures\inv_item_skirt.tga
Delete $INSTDIR\skins\default\textures\inv_item_skin.tga
Delete $INSTDIR\skins\default\textures\inv_item_skin.tga
Delete $INSTDIR\skins\default\textures\inv_item_shoes.tga
Delete $INSTDIR\skins\default\textures\inv_item_shoes.tga
Delete $INSTDIR\skins\default\textures\inv_item_shirt.tga
Delete $INSTDIR\skins\default\textures\inv_item_shirt.tga
Delete $INSTDIR\skins\default\textures\inv_item_shape.tga
Delete $INSTDIR\skins\default\textures\inv_item_shape.tga
Delete $INSTDIR\skins\default\textures\inv_item_script_dangerous.tga
Delete $INSTDIR\skins\default\textures\inv_item_script_dangerous.tga
Delete $INSTDIR\skins\default\textures\inv_item_script.tga
Delete $INSTDIR\skins\default\textures\inv_item_script.tga
Delete $INSTDIR\skins\default\textures\inv_item_pants.tga
Delete $INSTDIR\skins\default\textures\inv_item_pants.tga
Delete $INSTDIR\skins\default\textures\inv_item_object_multi.tga
Delete $INSTDIR\skins\default\textures\inv_item_object_multi.tga
Delete $INSTDIR\skins\default\textures\inv_item_object.tga
Delete $INSTDIR\skins\default\textures\inv_item_object.tga
Delete $INSTDIR\skins\default\textures\inv_item_notecard.tga
Delete $INSTDIR\skins\default\textures\inv_item_notecard.tga
Delete $INSTDIR\skins\default\textures\inv_item_landmark_visited.tga
Delete $INSTDIR\skins\default\textures\inv_item_landmark_visited.tga
Delete $INSTDIR\skins\default\textures\inv_item_landmark.tga
Delete $INSTDIR\skins\default\textures\inv_item_landmark.tga
Delete $INSTDIR\skins\default\textures\inv_item_jacket.tga
Delete $INSTDIR\skins\default\textures\inv_item_jacket.tga
Delete $INSTDIR\skins\default\textures\inv_item_hair.tga
Delete $INSTDIR\skins\default\textures\inv_item_hair.tga
Delete $INSTDIR\skins\default\textures\inv_item_gloves.tga
Delete $INSTDIR\skins\default\textures\inv_item_gloves.tga
Delete $INSTDIR\skins\default\textures\inv_item_gesture.tga
Delete $INSTDIR\skins\default\textures\inv_item_gesture.tga
Delete $INSTDIR\skins\default\textures\inv_item_eyes.tga
Delete $INSTDIR\skins\default\textures\inv_item_eyes.tga
Delete $INSTDIR\skins\default\textures\inv_item_clothing.tga
Delete $INSTDIR\skins\default\textures\inv_item_clothing.tga
Delete $INSTDIR\skins\default\textures\inv_item_callingcard_online.tga
Delete $INSTDIR\skins\default\textures\inv_item_callingcard_online.tga
Delete $INSTDIR\skins\default\textures\inv_item_callingcard_offline.tga
Delete $INSTDIR\skins\default\textures\inv_item_callingcard_offline.tga
Delete $INSTDIR\skins\default\textures\inv_item_attach.tga
Delete $INSTDIR\skins\default\textures\inv_item_attach.tga
Delete $INSTDIR\skins\default\textures\inv_item_animation.tga
Delete $INSTDIR\skins\default\textures\inv_item_animation.tga
Delete $INSTDIR\skins\default\textures\inv_item_alpha.tga
Delete $INSTDIR\skins\default\textures\inv_item_alpha.tga
Delete $INSTDIR\skins\default\textures\inv_folder_trash.tga
Delete $INSTDIR\skins\default\textures\inv_folder_trash.tga
Delete $INSTDIR\skins\default\textures\inv_folder_texture.tga
Delete $INSTDIR\skins\default\textures\inv_folder_texture.tga
Delete $INSTDIR\skins\default\textures\inv_folder_sound.tga
Delete $INSTDIR\skins\default\textures\inv_folder_sound.tga
Delete $INSTDIR\skins\default\textures\inv_folder_snapshot.tga
Delete $INSTDIR\skins\default\textures\inv_folder_snapshot.tga
Delete $INSTDIR\skins\default\textures\inv_folder_script.tga
Delete $INSTDIR\skins\default\textures\inv_folder_script.tga
Delete $INSTDIR\skins\default\textures\inv_folder_plain_open.tga
Delete $INSTDIR\skins\default\textures\inv_folder_plain_open.tga
Delete $INSTDIR\skins\default\textures\inv_folder_plain_closed.tga
Delete $INSTDIR\skins\default\textures\inv_folder_plain_closed.tga
Delete $INSTDIR\skins\default\textures\inv_folder_object.tga
Delete $INSTDIR\skins\default\textures\inv_folder_object.tga
Delete $INSTDIR\skins\default\textures\inv_folder_notecard.tga
Delete $INSTDIR\skins\default\textures\inv_folder_notecard.tga
Delete $INSTDIR\skins\default\textures\inv_folder_lostandfound.tga
Delete $INSTDIR\skins\default\textures\inv_folder_lostandfound.tga
Delete $INSTDIR\skins\default\textures\inv_folder_landmark.tga
Delete $INSTDIR\skins\default\textures\inv_folder_landmark.tga
Delete $INSTDIR\skins\default\textures\inv_folder_gesture.tga
Delete $INSTDIR\skins\default\textures\inv_folder_gesture.tga
Delete $INSTDIR\skins\default\textures\inv_folder_clothing.tga
Delete $INSTDIR\skins\default\textures\inv_folder_clothing.tga
Delete $INSTDIR\skins\default\textures\inv_folder_callingcard.tga
Delete $INSTDIR\skins\default\textures\inv_folder_callingcard.tga
Delete $INSTDIR\skins\default\textures\inv_folder_bodypart.tga
Delete $INSTDIR\skins\default\textures\inv_folder_bodypart.tga
Delete $INSTDIR\skins\default\textures\inv_folder_animation.tga
Delete $INSTDIR\skins\default\textures\inv_folder_animation.tga
Delete $INSTDIR\skins\default\textures\icon_top_pick.tga
Delete $INSTDIR\skins\default\textures\icon_top_pick.tga
Delete $INSTDIR\skins\default\textures\icon_popular.tga
Delete $INSTDIR\skins\default\textures\icon_popular.tga
Delete $INSTDIR\skins\default\textures\icon_place.tga
Delete $INSTDIR\skins\default\textures\icon_place.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_eventpoll_gray.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_eventpoll_gray.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_eventpoll.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_eventpoll.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_circuit_gray.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_circuit_gray.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_circuit.tga
Delete $INSTDIR\skins\default\textures\icon_net_close_circuit.tga
Delete $INSTDIR\skins\default\textures\icon_lock.tga
Delete $INSTDIR\skins\default\textures\icon_lock.tga
Delete $INSTDIR\skins\default\textures\icon_groupnoticeinventory.tga
Delete $INSTDIR\skins\default\textures\icon_groupnoticeinventory.tga
Delete $INSTDIR\skins\default\textures\icon_groupnotice.tga
Delete $INSTDIR\skins\default\textures\icon_groupnotice.tga
Delete $INSTDIR\skins\default\textures\icon_group.tga
Delete $INSTDIR\skins\default\textures\icon_group.tga
Delete $INSTDIR\skins\default\textures\icon_for_sale_adult.tga
Delete $INSTDIR\skins\default\textures\icon_for_sale_adult.tga
Delete $INSTDIR\skins\default\textures\icon_for_sale.tga
Delete $INSTDIR\skins\default\textures\icon_for_sale.tga
Delete $INSTDIR\skins\default\textures\icon_event_mature.tga
Delete $INSTDIR\skins\default\textures\icon_event_mature.tga
Delete $INSTDIR\skins\default\textures\icon_event_adult.tga
Delete $INSTDIR\skins\default\textures\icon_event_adult.tga
Delete $INSTDIR\skins\default\textures\icon_event.tga
Delete $INSTDIR\skins\default\textures\icon_event.tga
Delete $INSTDIR\skins\default\textures\icon_diurnal.tga
Delete $INSTDIR\skins\default\textures\icon_diurnal.tga
Delete $INSTDIR\skins\default\textures\icon_day_cycle.tga
Delete $INSTDIR\skins\default\textures\icon_day_cycle.tga
Delete $INSTDIR\skins\default\textures\icon_avatar_online.tga
Delete $INSTDIR\skins\default\textures\icon_avatar_online.tga
Delete $INSTDIR\skins\default\textures\icon_avatar_offline.tga
Delete $INSTDIR\skins\default\textures\icon_avatar_offline.tga
Delete $INSTDIR\skins\default\textures\icon_auction.tga
Delete $INSTDIR\skins\default\textures\icon_auction.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on-lvl3.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on-lvl3.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on-lvl2.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on-lvl2.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on-lvl1.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-on-lvl1.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-off.tga
Delete $INSTDIR\skins\default\textures\icn_voice_ptt-off.tga
Delete $INSTDIR\skins\default\textures\icn_voice-pvtfocus.tga
Delete $INSTDIR\skins\default\textures\icn_voice-pvtfocus.tga
Delete $INSTDIR\skins\default\textures\icn_voice-localchat.tga
Delete $INSTDIR\skins\default\textures\icn_voice-localchat.tga
Delete $INSTDIR\skins\default\textures\icn_voice-groupfocus.tga
Delete $INSTDIR\skins\default\textures\icn_voice-groupfocus.tga
Delete $INSTDIR\skins\default\textures\icn_voice-call-start.tga
Delete $INSTDIR\skins\default\textures\icn_voice-call-start.tga
Delete $INSTDIR\skins\default\textures\icn_voice-call-end.tga
Delete $INSTDIR\skins\default\textures\icn_voice-call-end.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_snapshot.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_snapshot.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_search.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_search.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_radar.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_radar.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_minimap.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_minimap.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_map.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_map.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_inventory.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_inventory.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_fly.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_fly.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_build.tga
Delete $INSTDIR\skins\default\textures\icn_toolbar_build.tga
Delete $INSTDIR\skins\default\textures\icn_textfield_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_textfield_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_stop.tga
Delete $INSTDIR\skins\default\textures\icn_stop.tga
Delete $INSTDIR\skins\default\textures\icn_speaker_dark.tga
Delete $INSTDIR\skins\default\textures\icn_speaker_dark.tga
Delete $INSTDIR\skins\default\textures\icn_speaker-muted_dark.tga
Delete $INSTDIR\skins\default\textures\icn_speaker-muted_dark.tga
Delete $INSTDIR\skins\default\textures\icn_slide-thumb_dark.tga
Delete $INSTDIR\skins\default\textures\icn_slide-thumb_dark.tga
Delete $INSTDIR\skins\default\textures\icn_slide-highlight.tga
Delete $INSTDIR\skins\default\textures\icn_slide-highlight.tga
Delete $INSTDIR\skins\default\textures\icn_slide-groove_dark.tga
Delete $INSTDIR\skins\default\textures\icn_slide-groove_dark.tga
Delete $INSTDIR\skins\default\textures\icn_scrollbar_thumb.tga
Delete $INSTDIR\skins\default\textures\icn_scrollbar_thumb.tga
Delete $INSTDIR\skins\default\textures\icn_scrollbar_bg.tga
Delete $INSTDIR\skins\default\textures\icn_scrollbar_bg.tga
Delete $INSTDIR\skins\default\textures\icn_scrollbar.tga
Delete $INSTDIR\skins\default\textures\icn_scrollbar.tga
Delete $INSTDIR\skins\default\textures\icn_rounded-text-field.tga
Delete $INSTDIR\skins\default\textures\icn_rounded-text-field.tga
Delete $INSTDIR\skins\default\textures\icn_play.tga
Delete $INSTDIR\skins\default\textures\icn_play.tga
Delete $INSTDIR\skins\default\textures\icn_pause.tga
Delete $INSTDIR\skins\default\textures\icn_pause.tga
Delete $INSTDIR\skins\default\textures\icn_music.tga
Delete $INSTDIR\skins\default\textures\icn_music.tga
Delete $INSTDIR\skins\default\textures\icn_music-play.tga
Delete $INSTDIR\skins\default\textures\icn_music-play.tga
Delete $INSTDIR\skins\default\textures\icn_music-pause.tga
Delete $INSTDIR\skins\default\textures\icn_music-pause.tga
Delete $INSTDIR\skins\default\textures\icn_media_web.tga
Delete $INSTDIR\skins\default\textures\icn_media_web.tga
Delete $INSTDIR\skins\default\textures\icn_media_movie.tga
Delete $INSTDIR\skins\default\textures\icn_media_movie.tga
Delete $INSTDIR\skins\default\textures\icn_media.tga
Delete $INSTDIR\skins\default\textures\icn_media.tga
Delete $INSTDIR\skins\default\textures\icn_media-stop_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-stop_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-stop_disabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-stop_disabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-stop_active.tga
Delete $INSTDIR\skins\default\textures\icn_media-stop_active.tga
Delete $INSTDIR\skins\default\textures\icn_media-play_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-play_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-play_disabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-play_disabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-play_active.tga
Delete $INSTDIR\skins\default\textures\icn_media-play_active.tga
Delete $INSTDIR\skins\default\textures\icn_media-play.tga
Delete $INSTDIR\skins\default\textures\icn_media-play.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause_enabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause_disabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause_disabled.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause_active.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause_active.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause.tga
Delete $INSTDIR\skins\default\textures\icn_media-pause.tga
Delete $INSTDIR\skins\default\textures\icn_label_web.tga
Delete $INSTDIR\skins\default\textures\icn_label_web.tga
Delete $INSTDIR\skins\default\textures\icn_label_music.tga
Delete $INSTDIR\skins\default\textures\icn_label_music.tga
Delete $INSTDIR\skins\default\textures\icn_label_media.tga
Delete $INSTDIR\skins\default\textures\icn_label_media.tga
Delete $INSTDIR\skins\default\textures\icn_clear_lineeditor.tga
Delete $INSTDIR\skins\default\textures\icn_clear_lineeditor.tga
Delete $INSTDIR\skins\default\textures\icn_chatbar.tga
Delete $INSTDIR\skins\default\textures\icn_chatbar.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-typing3.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-typing3.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-typing2.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-typing2.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-typing1.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-typing1.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-dot-lvl2.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-dot-lvl2.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-dot-lvl1.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-dot-lvl1.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-dot-lvl0.tga
Delete $INSTDIR\skins\default\textures\icn_active-speakers-dot-lvl0.tga
Delete $INSTDIR\skins\default\textures\foot_shadow.j2c
Delete $INSTDIR\skins\default\textures\foot_shadow.j2c
Delete $INSTDIR\skins\default\textures\folder_arrow.tga
Delete $INSTDIR\skins\default\textures\folder_arrow.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_right_selected.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_right_selected.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_right_disabled.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_right_disabled.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_right.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_right.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_left_selected.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_left_selected.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_left_disabled.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_left_disabled.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_left.tga
Delete $INSTDIR\skins\default\textures\flyout_btn_left.tga
Delete $INSTDIR\skins\default\textures\ff_visible_online_button.tga
Delete $INSTDIR\skins\default\textures\ff_visible_online_button.tga
Delete $INSTDIR\skins\default\textures\ff_visible_online.tga
Delete $INSTDIR\skins\default\textures\ff_visible_online.tga
Delete $INSTDIR\skins\default\textures\ff_visible_map_button.tga
Delete $INSTDIR\skins\default\textures\ff_visible_map_button.tga
Delete $INSTDIR\skins\default\textures\ff_visible_map.tga
Delete $INSTDIR\skins\default\textures\ff_visible_map.tga
Delete $INSTDIR\skins\default\textures\ff_online_status_button.tga
Delete $INSTDIR\skins\default\textures\ff_online_status_button.tga
Delete $INSTDIR\skins\default\textures\ff_edit_theirs_button.tga
Delete $INSTDIR\skins\default\textures\ff_edit_theirs_button.tga
Delete $INSTDIR\skins\default\textures\ff_edit_theirs.tga
Delete $INSTDIR\skins\default\textures\ff_edit_theirs.tga
Delete $INSTDIR\skins\default\textures\ff_edit_mine_button.tga
Delete $INSTDIR\skins\default\textures\ff_edit_mine_button.tga
Delete $INSTDIR\skins\default\textures\ff_edit_mine.tga
Delete $INSTDIR\skins\default\textures\ff_edit_mine.tga
Delete $INSTDIR\skins\default\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
Delete $INSTDIR\skins\default\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
Delete $INSTDIR\skins\default\textures\fc987bf9-b8cb-f8e5-45f2-d664ca6bd3eb.j2c
Delete $INSTDIR\skins\default\textures\fc987bf9-b8cb-f8e5-45f2-d664ca6bd3eb.j2c
Delete $INSTDIR\skins\default\textures\fb2ae204-3fd1-df33-594f-c9f882830e66.j2c
Delete $INSTDIR\skins\default\textures\fb2ae204-3fd1-df33-594f-c9f882830e66.j2c
Delete $INSTDIR\skins\default\textures\fb1fecba-9585-415b-ad15-6e6e3d6c5479.j2c
Delete $INSTDIR\skins\default\textures\fb1fecba-9585-415b-ad15-6e6e3d6c5479.j2c
Delete $INSTDIR\skins\default\textures\f9bbb2fe-584b-4c01-86fc-599c69534c1b.tga
Delete $INSTDIR\skins\default\textures\f9bbb2fe-584b-4c01-86fc-599c69534c1b.tga
Delete $INSTDIR\skins\default\textures\f54a0c32-3cd1-d49a-5b4f-7b792bebc204.j2c
Delete $INSTDIR\skins\default\textures\f54a0c32-3cd1-d49a-5b4f-7b792bebc204.j2c
Delete $INSTDIR\skins\default\textures\f4b6b161-6530-6679-1a84-adfcb71a8b12.j2c
Delete $INSTDIR\skins\default\textures\f4b6b161-6530-6679-1a84-adfcb71a8b12.j2c
Delete $INSTDIR\skins\default\textures\f2d7b6f6-4200-1e9a-fd5b-96459e950f94.j2c
Delete $INSTDIR\skins\default\textures\f2d7b6f6-4200-1e9a-fd5b-96459e950f94.j2c
Delete $INSTDIR\skins\default\textures\eye_button_inactive.tga
Delete $INSTDIR\skins\default\textures\eye_button_inactive.tga
Delete $INSTDIR\skins\default\textures\eye_button_active.tga
Delete $INSTDIR\skins\default\textures\eye_button_active.tga
Delete $INSTDIR\skins\default\textures\ebf2aa19-6c34-c5d8-4f14-853da1241f91.j2c
Delete $INSTDIR\skins\default\textures\ebf2aa19-6c34-c5d8-4f14-853da1241f91.j2c
Delete $INSTDIR\skins\default\textures\e97cf410-8e61-7005-ec06-629eba4cd1fb.tga
Delete $INSTDIR\skins\default\textures\e97cf410-8e61-7005-ec06-629eba4cd1fb.tga
Delete $INSTDIR\skins\default\textures\e674ca0c-a387-4dae-a0b4-db6bd073faa5.j2c
Delete $INSTDIR\skins\default\textures\e674ca0c-a387-4dae-a0b4-db6bd073faa5.j2c
Delete $INSTDIR\skins\default\textures\e569711a-27c2-aad4-9246-0c910239a179.j2c
Delete $INSTDIR\skins\default\textures\e569711a-27c2-aad4-9246-0c910239a179.j2c
Delete $INSTDIR\skins\default\textures\e38248f9-f2ee-2c9f-aa49-4860857e3b08.j2c
Delete $INSTDIR\skins\default\textures\e38248f9-f2ee-2c9f-aa49-4860857e3b08.j2c
Delete $INSTDIR\skins\default\textures\e3369e02-93e1-43dc-b9c0-4533db0963d0.tga
Delete $INSTDIR\skins\default\textures\e3369e02-93e1-43dc-b9c0-4533db0963d0.tga
Delete $INSTDIR\skins\default\textures\e121e2fc-7573-740f-edfd-0d45a9ba486e.j2c
Delete $INSTDIR\skins\default\textures\e121e2fc-7573-740f-edfd-0d45a9ba486e.j2c
Delete $INSTDIR\skins\default\textures\down_arrow.png
Delete $INSTDIR\skins\default\textures\down_arrow.png
Delete $INSTDIR\skins\default\textures\direction_arrow.tga
Delete $INSTDIR\skins\default\textures\direction_arrow.tga
Delete $INSTDIR\skins\default\textures\default_profile_picture.j2c
Delete $INSTDIR\skins\default\textures\default_profile_picture.j2c
Delete $INSTDIR\skins\default\textures\default_land_picture.j2c
Delete $INSTDIR\skins\default\textures\default_land_picture.j2c
Delete $INSTDIR\skins\default\textures\de651394-f926-48db-b666-e49d83af1bbc.j2c
Delete $INSTDIR\skins\default\textures\de651394-f926-48db-b666-e49d83af1bbc.j2c
Delete $INSTDIR\skins\default\textures\db9d39ec-a896-c287-1ced-64566217021e.j2c
Delete $INSTDIR\skins\default\textures\db9d39ec-a896-c287-1ced-64566217021e.j2c
Delete $INSTDIR\skins\default\textures\darkgray.tga
Delete $INSTDIR\skins\default\textures\darkgray.tga
Delete $INSTDIR\skins\default\textures\d9258671-868f-7511-c321-7baef9e948a4.j2c
Delete $INSTDIR\skins\default\textures\d9258671-868f-7511-c321-7baef9e948a4.j2c
Delete $INSTDIR\skins\default\textures\d7d99e40-10e2-5739-d063-91dcbdefc492.j2c
Delete $INSTDIR\skins\default\textures\d7d99e40-10e2-5739-d063-91dcbdefc492.j2c
Delete $INSTDIR\skins\default\textures\d691a01c-13b7-578d-57c0-5caef0b4e7e1.j2c
Delete $INSTDIR\skins\default\textures\d691a01c-13b7-578d-57c0-5caef0b4e7e1.j2c
Delete $INSTDIR\skins\default\textures\d319ce44-0821-932a-cd18-cd1afb9d3ead.j2c
Delete $INSTDIR\skins\default\textures\d319ce44-0821-932a-cd18-cd1afb9d3ead.j2c
Delete $INSTDIR\skins\default\textures\d21e44ca-ff1c-a96e-b2ef-c0753426b7d9.j2c
Delete $INSTDIR\skins\default\textures\d21e44ca-ff1c-a96e-b2ef-c0753426b7d9.j2c
Delete $INSTDIR\skins\default\textures\d2114404-dd59-4a4d-8e6c-49359e91bbf0.j2c
Delete $INSTDIR\skins\default\textures\d2114404-dd59-4a4d-8e6c-49359e91bbf0.j2c
Delete $INSTDIR\skins\default\textures\d07f6eed-b96a-47cd-b51d-400ad4a1c428.j2c
Delete $INSTDIR\skins\default\textures\d07f6eed-b96a-47cd-b51d-400ad4a1c428.j2c
Delete $INSTDIR\skins\default\textures\crosshairs.tga
Delete $INSTDIR\skins\default\textures\crosshairs.tga
Delete $INSTDIR\skins\default\textures\combobox_arrow.tga
Delete $INSTDIR\skins\default\textures\combobox_arrow.tga
Delete $INSTDIR\skins\default\textures\color_swatch_alpha.tga
Delete $INSTDIR\skins\default\textures\color_swatch_alpha.tga
Delete $INSTDIR\skins\default\textures\cloud-particle.j2c
Delete $INSTDIR\skins\default\textures\cloud-particle.j2c
Delete $INSTDIR\skins\default\textures\closebox.tga
Delete $INSTDIR\skins\default\textures\closebox.tga
Delete $INSTDIR\skins\default\textures\close_inactive_blue.tga
Delete $INSTDIR\skins\default\textures\close_inactive_blue.tga
Delete $INSTDIR\skins\default\textures\close_in_blue.tga
Delete $INSTDIR\skins\default\textures\close_in_blue.tga
Delete $INSTDIR\skins\default\textures\circle.tga
Delete $INSTDIR\skins\default\textures\circle.tga
Delete $INSTDIR\skins\default\textures\checkerboard_transparency_bg.png
Delete $INSTDIR\skins\default\textures\checkerboard_transparency_bg.png
Delete $INSTDIR\skins\default\textures\checkbox_enabled_true.tga
Delete $INSTDIR\skins\default\textures\checkbox_enabled_true.tga
Delete $INSTDIR\skins\default\textures\checkbox_enabled_false.tga
Delete $INSTDIR\skins\default\textures\checkbox_enabled_false.tga
Delete $INSTDIR\skins\default\textures\checkbox_disabled_true.tga
Delete $INSTDIR\skins\default\textures\checkbox_disabled_true.tga
Delete $INSTDIR\skins\default\textures\checkbox_disabled_false.tga
Delete $INSTDIR\skins\default\textures\checkbox_disabled_false.tga
Delete $INSTDIR\skins\default\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
Delete $INSTDIR\skins\default\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
Delete $INSTDIR\skins\default\textures\cdd9a9fc-6d0b-f90d-8416-c72b6019bca8.j2c
Delete $INSTDIR\skins\default\textures\cdd9a9fc-6d0b-f90d-8416-c72b6019bca8.j2c
Delete $INSTDIR\skins\default\textures\cce0f112-878f-4586-a2e2-a8f104bba271.j2c
Delete $INSTDIR\skins\default\textures\cce0f112-878f-4586-a2e2-a8f104bba271.j2c
Delete $INSTDIR\skins\default\textures\cam_zoom_plus_in.tga
Delete $INSTDIR\skins\default\textures\cam_zoom_plus_in.tga
Delete $INSTDIR\skins\default\textures\cam_zoom_out.tga
Delete $INSTDIR\skins\default\textures\cam_zoom_out.tga
Delete $INSTDIR\skins\default\textures\cam_zoom_minus_in.tga
Delete $INSTDIR\skins\default\textures\cam_zoom_minus_in.tga
Delete $INSTDIR\skins\default\textures\cam_tracking_out.tga
Delete $INSTDIR\skins\default\textures\cam_tracking_out.tga
Delete $INSTDIR\skins\default\textures\cam_tracking_in.tga
Delete $INSTDIR\skins\default\textures\cam_tracking_in.tga
Delete $INSTDIR\skins\default\textures\cam_rotate_out.tga
Delete $INSTDIR\skins\default\textures\cam_rotate_out.tga
Delete $INSTDIR\skins\default\textures\cam_rotate_in.tga
Delete $INSTDIR\skins\default\textures\cam_rotate_in.tga
Delete $INSTDIR\skins\default\textures\ca7609c6-6ec6-32d9-332e-0d8f437ef644.tga
Delete $INSTDIR\skins\default\textures\ca7609c6-6ec6-32d9-332e-0d8f437ef644.tga
Delete $INSTDIR\skins\default\textures\ca4e8c27-473c-eb1c-2f5d-50ee3f07d85c.j2c
Delete $INSTDIR\skins\default\textures\ca4e8c27-473c-eb1c-2f5d-50ee3f07d85c.j2c
Delete $INSTDIR\skins\default\textures\c7d8bbf3-21ee-4f6e-9b20-3cf18425af1d.j2c
Delete $INSTDIR\skins\default\textures\c7d8bbf3-21ee-4f6e-9b20-3cf18425af1d.j2c
Delete $INSTDIR\skins\default\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
Delete $INSTDIR\skins\default\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
Delete $INSTDIR\skins\default\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
Delete $INSTDIR\skins\default\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
Delete $INSTDIR\skins\default\textures\button_enabled_selected_32x128.tga
Delete $INSTDIR\skins\default\textures\button_enabled_selected_32x128.tga
Delete $INSTDIR\skins\default\textures\button_enabled_32x128.tga
Delete $INSTDIR\skins\default\textures\button_enabled_32x128.tga
Delete $INSTDIR\skins\default\textures\button_disabled_32x128.tga
Delete $INSTDIR\skins\default\textures\button_disabled_32x128.tga
Delete $INSTDIR\skins\default\textures\button_anim_stop_selected.tga
Delete $INSTDIR\skins\default\textures\button_anim_stop_selected.tga
Delete $INSTDIR\skins\default\textures\button_anim_stop_disabled.tga
Delete $INSTDIR\skins\default\textures\button_anim_stop_disabled.tga
Delete $INSTDIR\skins\default\textures\button_anim_stop.tga
Delete $INSTDIR\skins\default\textures\button_anim_stop.tga
Delete $INSTDIR\skins\default\textures\button_anim_play_selected.tga
Delete $INSTDIR\skins\default\textures\button_anim_play_selected.tga
Delete $INSTDIR\skins\default\textures\button_anim_play_disabled.tga
Delete $INSTDIR\skins\default\textures\button_anim_play_disabled.tga
Delete $INSTDIR\skins\default\textures\button_anim_play.tga
Delete $INSTDIR\skins\default\textures\button_anim_play.tga
Delete $INSTDIR\skins\default\textures\button_anim_pause_selected.tga
Delete $INSTDIR\skins\default\textures\button_anim_pause_selected.tga
Delete $INSTDIR\skins\default\textures\button_anim_pause_disabled.tga
Delete $INSTDIR\skins\default\textures\button_anim_pause_disabled.tga
Delete $INSTDIR\skins\default\textures\button_anim_pause.tga
Delete $INSTDIR\skins\default\textures\button_anim_pause.tga
Delete $INSTDIR\skins\default\textures\btn_chatbar_selected.tga
Delete $INSTDIR\skins\default\textures\btn_chatbar_selected.tga
Delete $INSTDIR\skins\default\textures\btn_chatbar.tga
Delete $INSTDIR\skins\default\textures\btn_chatbar.tga
Delete $INSTDIR\skins\default\textures\black.tga
Delete $INSTDIR\skins\default\textures\black.tga
Delete $INSTDIR\skins\default\textures\beb169c7-11ea-fff2-efe5-0f24dc881df2.j2c
Delete $INSTDIR\skins\default\textures\beb169c7-11ea-fff2-efe5-0f24dc881df2.j2c
Delete $INSTDIR\skins\default\textures\badge_warn.j2c
Delete $INSTDIR\skins\default\textures\badge_warn.j2c
Delete $INSTDIR\skins\default\textures\badge_ok.j2c
Delete $INSTDIR\skins\default\textures\badge_ok.j2c
Delete $INSTDIR\skins\default\textures\badge_note.j2c
Delete $INSTDIR\skins\default\textures\badge_note.j2c
Delete $INSTDIR\skins\default\textures\badge_error.j2c
Delete $INSTDIR\skins\default\textures\badge_error.j2c
Delete $INSTDIR\skins\default\textures\b9f1a3b8-933e-b7c8-e6f5-dba1bc666bed.j2c
Delete $INSTDIR\skins\default\textures\b9f1a3b8-933e-b7c8-e6f5-dba1bc666bed.j2c
Delete $INSTDIR\skins\default\textures\b9e1cf8a-9660-c020-0c69-18f1ea27268a.j2c
Delete $INSTDIR\skins\default\textures\b9e1cf8a-9660-c020-0c69-18f1ea27268a.j2c
Delete $INSTDIR\skins\default\textures\b8eed5f0-64b7-6e12-b67f-43fa8e773440.j2c
Delete $INSTDIR\skins\default\textures\b8eed5f0-64b7-6e12-b67f-43fa8e773440.j2c
Delete $INSTDIR\skins\default\textures\b8d3965a-ad78-bf43-699b-bff8eca6c975.j2c
Delete $INSTDIR\skins\default\textures\b8d3965a-ad78-bf43-699b-bff8eca6c975.j2c
Delete $INSTDIR\skins\default\textures\b4ba225c-373f-446d-9f7e-6cb7b5cf9b3d.j2c
Delete $INSTDIR\skins\default\textures\b4ba225c-373f-446d-9f7e-6cb7b5cf9b3d.j2c
Delete $INSTDIR\skins\default\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
Delete $INSTDIR\skins\default\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
Delete $INSTDIR\skins\default\textures\b2ef2d31-9714-a07b-6ca7-31638166364b.tga
Delete $INSTDIR\skins\default\textures\b2ef2d31-9714-a07b-6ca7-31638166364b.tga
Delete $INSTDIR\skins\default\textures\avatar_thumb_bkgrnd.j2c
Delete $INSTDIR\skins\default\textures\avatar_thumb_bkgrnd.j2c
Delete $INSTDIR\skins\default\textures\arrow_up.tga
Delete $INSTDIR\skins\default\textures\arrow_up.tga
Delete $INSTDIR\skins\default\textures\arrow_down.tga
Delete $INSTDIR\skins\default\textures\arrow_down.tga
Delete $INSTDIR\skins\default\textures\alpha_gradient_2d.j2c
Delete $INSTDIR\skins\default\textures\alpha_gradient_2d.j2c
Delete $INSTDIR\skins\default\textures\alpha_gradient.tga
Delete $INSTDIR\skins\default\textures\alpha_gradient.tga
Delete $INSTDIR\skins\default\textures\ae874d1a-93ef-54fb-5fd3-eb0cb156afc0.j2c
Delete $INSTDIR\skins\default\textures\ae874d1a-93ef-54fb-5fd3-eb0cb156afc0.j2c
Delete $INSTDIR\skins\default\textures\active_voice_tab.tga
Delete $INSTDIR\skins\default\textures\active_voice_tab.tga
Delete $INSTDIR\skins\default\textures\active_speakers.tga
Delete $INSTDIR\skins\default\textures\active_speakers.tga
Delete $INSTDIR\skins\default\textures\account_id_orange.tga
Delete $INSTDIR\skins\default\textures\account_id_orange.tga
Delete $INSTDIR\skins\default\textures\account_id_green.tga
Delete $INSTDIR\skins\default\textures\account_id_green.tga
Delete $INSTDIR\skins\default\textures\abb783e6-3e93-26c0-248a-247666855da3.j2c
Delete $INSTDIR\skins\default\textures\abb783e6-3e93-26c0-248a-247666855da3.j2c
Delete $INSTDIR\skins\default\textures\a85ac674-cb75-4af6-9499-df7c5aaf7a28.j2c
Delete $INSTDIR\skins\default\textures\a85ac674-cb75-4af6-9499-df7c5aaf7a28.j2c
Delete $INSTDIR\skins\default\textures\a6162133-724b-54df-a12f-51cd070ad6f3.j2c
Delete $INSTDIR\skins\default\textures\a6162133-724b-54df-a12f-51cd070ad6f3.j2c
Delete $INSTDIR\skins\default\textures\9deab416-9c63-78d6-d558-9a156f12044c.j2c
Delete $INSTDIR\skins\default\textures\9deab416-9c63-78d6-d558-9a156f12044c.j2c
Delete $INSTDIR\skins\default\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
Delete $INSTDIR\skins\default\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
Delete $INSTDIR\skins\default\textures\9c88539c-fd04-46b8-bea2-ddf1bcffe3bd.j2c
Delete $INSTDIR\skins\default\textures\9c88539c-fd04-46b8-bea2-ddf1bcffe3bd.j2c
Delete $INSTDIR\skins\default\textures\9beb8cdd-3dce-53c2-b28e-e1f3bc2ec0a4.tga
Delete $INSTDIR\skins\default\textures\9beb8cdd-3dce-53c2-b28e-e1f3bc2ec0a4.tga
Delete $INSTDIR\skins\default\textures\99bd60a2-3250-efc9-2e39-2fbcadefbecc.j2c
Delete $INSTDIR\skins\default\textures\99bd60a2-3250-efc9-2e39-2fbcadefbecc.j2c
Delete $INSTDIR\skins\default\textures\988dd995-1769-bdc9-8842-51f8f2b03884.j2c
Delete $INSTDIR\skins\default\textures\988dd995-1769-bdc9-8842-51f8f2b03884.j2c
Delete $INSTDIR\skins\default\textures\978380f0-aaf7-c459-14e3-9808833fd372.j2c
Delete $INSTDIR\skins\default\textures\978380f0-aaf7-c459-14e3-9808833fd372.j2c
Delete $INSTDIR\skins\default\textures\96b4de31-f4fa-337d-ec78-451e3609769e.j2c
Delete $INSTDIR\skins\default\textures\96b4de31-f4fa-337d-ec78-451e3609769e.j2c
Delete $INSTDIR\skins\default\textures\95281d5c-d27a-ee13-e067-08295b67b58a.j2c
Delete $INSTDIR\skins\default\textures\95281d5c-d27a-ee13-e067-08295b67b58a.j2c
Delete $INSTDIR\skins\default\textures\92e66e00-f56f-598a-7997-048aa64cde18.j2c
Delete $INSTDIR\skins\default\textures\92e66e00-f56f-598a-7997-048aa64cde18.j2c
Delete $INSTDIR\skins\default\textures\8f761ce3-5939-4d3a-8991-00064fdfacf9.tga
Delete $INSTDIR\skins\default\textures\8f761ce3-5939-4d3a-8991-00064fdfacf9.tga
Delete $INSTDIR\skins\default\textures\8f458549-173b-23ff-d4ff-bfaa5ea2371b.j2c
Delete $INSTDIR\skins\default\textures\8f458549-173b-23ff-d4ff-bfaa5ea2371b.j2c
Delete $INSTDIR\skins\default\textures\8dcd4a48-2d37-4909-9f78-f7a9eb4ef903.j2c
Delete $INSTDIR\skins\default\textures\8dcd4a48-2d37-4909-9f78-f7a9eb4ef903.j2c
Delete $INSTDIR\skins\default\textures\8a515889-eac9-fb55-8eba-d2dc09eb32c8.j2c
Delete $INSTDIR\skins\default\textures\8a515889-eac9-fb55-8eba-d2dc09eb32c8.j2c
Delete $INSTDIR\skins\default\textures\89e9fc7c-0b16-457d-be4f-136270759c4d.tga
Delete $INSTDIR\skins\default\textures\89e9fc7c-0b16-457d-be4f-136270759c4d.tga
Delete $INSTDIR\skins\default\textures\8872f2b8-31db-42d8-580a-b3e4a91262de.j2c
Delete $INSTDIR\skins\default\textures\8872f2b8-31db-42d8-580a-b3e4a91262de.j2c
Delete $INSTDIR\skins\default\textures\83b77fc6-10b4-63ec-4de7-f40629f238c5.j2c
Delete $INSTDIR\skins\default\textures\83b77fc6-10b4-63ec-4de7-f40629f238c5.j2c
Delete $INSTDIR\skins\default\textures\827ff765-8c1d-a8b1-23f7-fdcba560effc.j2c
Delete $INSTDIR\skins\default\textures\827ff765-8c1d-a8b1-23f7-fdcba560effc.j2c
Delete $INSTDIR\skins\default\textures\822ded49-9a6c-f61c-cb89-6df54f42cdf4.j2c
Delete $INSTDIR\skins\default\textures\822ded49-9a6c-f61c-cb89-6df54f42cdf4.j2c
Delete $INSTDIR\skins\default\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
Delete $INSTDIR\skins\default\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
Delete $INSTDIR\skins\default\textures\7cb070bc-fc00-4527-9c4d-7f7e0c4191be.j2c
Delete $INSTDIR\skins\default\textures\7cb070bc-fc00-4527-9c4d-7f7e0c4191be.j2c
Delete $INSTDIR\skins\default\textures\7ca39b4c-bd19-4699-aff7-f93fd03d3e7b.j2c
Delete $INSTDIR\skins\default\textures\7ca39b4c-bd19-4699-aff7-f93fd03d3e7b.j2c
Delete $INSTDIR\skins\default\textures\7c0cf89b-44b1-1ce2-dd74-07102a98ac2a.j2c
Delete $INSTDIR\skins\default\textures\7c0cf89b-44b1-1ce2-dd74-07102a98ac2a.j2c
Delete $INSTDIR\skins\default\textures\7a2b3a4a-53c2-53ac-5716-aac7d743c020.j2c
Delete $INSTDIR\skins\default\textures\7a2b3a4a-53c2-53ac-5716-aac7d743c020.j2c
Delete $INSTDIR\skins\default\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
Delete $INSTDIR\skins\default\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
Delete $INSTDIR\skins\default\textures\79504bf5-c3ec-0763-6563-d843de66d0a1.j2c
Delete $INSTDIR\skins\default\textures\79504bf5-c3ec-0763-6563-d843de66d0a1.j2c
Delete $INSTDIR\skins\default\textures\78af921a-3c49-47a1-9c4e-2608951164ae.j2c
Delete $INSTDIR\skins\default\textures\78af921a-3c49-47a1-9c4e-2608951164ae.j2c
Delete $INSTDIR\skins\default\textures\7581f2f4-d0d2-481a-bc75-69a13d9caeaa.j2c
Delete $INSTDIR\skins\default\textures\7581f2f4-d0d2-481a-bc75-69a13d9caeaa.j2c
Delete $INSTDIR\skins\default\textures\74e1a96f-4833-a24d-a1bb-1bce1468b0e7.tga
Delete $INSTDIR\skins\default\textures\74e1a96f-4833-a24d-a1bb-1bce1468b0e7.tga
Delete $INSTDIR\skins\default\textures\74ba3584-58ea-9984-5b76-62d37942ab77.tga
Delete $INSTDIR\skins\default\textures\74ba3584-58ea-9984-5b76-62d37942ab77.tga
Delete $INSTDIR\skins\default\textures\73577b7b-19c3-4050-a19d-36bc2408aa79.tga
Delete $INSTDIR\skins\default\textures\73577b7b-19c3-4050-a19d-36bc2408aa79.tga
Delete $INSTDIR\skins\default\textures\735198cf-6ea0-2550-e222-21d3c6a341ae.j2c
Delete $INSTDIR\skins\default\textures\735198cf-6ea0-2550-e222-21d3c6a341ae.j2c
Delete $INSTDIR\skins\default\textures\6de37e4e-7029-61f5-54b8-f5e63f983f58.j2c
Delete $INSTDIR\skins\default\textures\6de37e4e-7029-61f5-54b8-f5e63f983f58.j2c
Delete $INSTDIR\skins\default\textures\6c9fa78a-1c69-2168-325b-3e03ffa348ce.j2c
Delete $INSTDIR\skins\default\textures\6c9fa78a-1c69-2168-325b-3e03ffa348ce.j2c
Delete $INSTDIR\skins\default\textures\6c4727b8-ac79-ba44-3b81-f9aa887b47eb.j2c
Delete $INSTDIR\skins\default\textures\6c4727b8-ac79-ba44-3b81-f9aa887b47eb.j2c
Delete $INSTDIR\skins\default\textures\67931331-0c02-4876-1255-28770896c6a2.j2c
Delete $INSTDIR\skins\default\textures\67931331-0c02-4876-1255-28770896c6a2.j2c
Delete $INSTDIR\skins\default\textures\6522e74d-1660-4e7f-b601-6f48c1659a77.j2c
Delete $INSTDIR\skins\default\textures\6522e74d-1660-4e7f-b601-6f48c1659a77.j2c
Delete $INSTDIR\skins\default\textures\64eed6af-f575-35c7-baa4-b140bdcdb00f.j2c
Delete $INSTDIR\skins\default\textures\64eed6af-f575-35c7-baa4-b140bdcdb00f.j2c
Delete $INSTDIR\skins\default\textures\64367bd1-697e-b3e6-0b65-3f862a577366.j2c
Delete $INSTDIR\skins\default\textures\64367bd1-697e-b3e6-0b65-3f862a577366.j2c
Delete $INSTDIR\skins\default\textures\63338ede-0037-c4fd-855b-015d77112fc8.j2c
Delete $INSTDIR\skins\default\textures\63338ede-0037-c4fd-855b-015d77112fc8.j2c
Delete $INSTDIR\skins\default\textures\6002a571-549c-472c-9443-9ab35b1a55ed.tga
Delete $INSTDIR\skins\default\textures\6002a571-549c-472c-9443-9ab35b1a55ed.tga
Delete $INSTDIR\skins\default\textures\5bc11cd6-2f40-071e-a8da-0903394204f9.j2c
Delete $INSTDIR\skins\default\textures\5bc11cd6-2f40-071e-a8da-0903394204f9.j2c
Delete $INSTDIR\skins\default\textures\5abfabc2-5d6d-4912-acd8-d7e38ae93d02.j2c
Delete $INSTDIR\skins\default\textures\5abfabc2-5d6d-4912-acd8-d7e38ae93d02.j2c
Delete $INSTDIR\skins\default\textures\5ab48dd5-05d0-4f1a-ace6-efd4e2fb3508.j2c
Delete $INSTDIR\skins\default\textures\5ab48dd5-05d0-4f1a-ace6-efd4e2fb3508.j2c
Delete $INSTDIR\skins\default\textures\5894e2e7-ab8d-edfa-e61c-18cf16854ba3.j2c
Delete $INSTDIR\skins\default\textures\5894e2e7-ab8d-edfa-e61c-18cf16854ba3.j2c
Delete $INSTDIR\skins\default\textures\5748decc-f629-461c-9a36-a35a221fe21f.tga
Delete $INSTDIR\skins\default\textures\5748decc-f629-461c-9a36-a35a221fe21f.tga
Delete $INSTDIR\skins\default\textures\53a2f406-4895-1d13-d541-d2e3b86bc19c.j2c
Delete $INSTDIR\skins\default\textures\53a2f406-4895-1d13-d541-d2e3b86bc19c.j2c
Delete $INSTDIR\skins\default\textures\48766d75-6e58-de84-68fe-1980c64feaee.j2c
Delete $INSTDIR\skins\default\textures\48766d75-6e58-de84-68fe-1980c64feaee.j2c
Delete $INSTDIR\skins\default\textures\47a8c844-cd2a-4b1a-be01-df8b1612fe5d.tga
Delete $INSTDIR\skins\default\textures\47a8c844-cd2a-4b1a-be01-df8b1612fe5d.tga
Delete $INSTDIR\skins\default\textures\4726f13e-bd07-f2fb-feb0-bfa2ac58ab61.j2c
Delete $INSTDIR\skins\default\textures\4726f13e-bd07-f2fb-feb0-bfa2ac58ab61.j2c
Delete $INSTDIR\skins\default\textures\43f0a590-f3d3-48b5-b460-f5b3e6e03626.tga
Delete $INSTDIR\skins\default\textures\43f0a590-f3d3-48b5-b460-f5b3e6e03626.tga
Delete $INSTDIR\skins\default\textures\402f8b24-5f9d-4905-b5f8-37baff603e88.j2c
Delete $INSTDIR\skins\default\textures\402f8b24-5f9d-4905-b5f8-37baff603e88.j2c
Delete $INSTDIR\skins\default\textures\3cddf591-a726-4702-87b3-70c1daf88f90.j2c
Delete $INSTDIR\skins\default\textures\3cddf591-a726-4702-87b3-70c1daf88f90.j2c
Delete $INSTDIR\skins\default\textures\3c59f7fe-9dc8-47f9-8aaf-a9dd1fbc3bef.j2c
Delete $INSTDIR\skins\default\textures\3c59f7fe-9dc8-47f9-8aaf-a9dd1fbc3bef.j2c
Delete $INSTDIR\skins\default\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
Delete $INSTDIR\skins\default\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
Delete $INSTDIR\skins\default\textures\39801651-26cb-4926-af57-7af9352c273c.tga
Delete $INSTDIR\skins\default\textures\39801651-26cb-4926-af57-7af9352c273c.tga
Delete $INSTDIR\skins\default\textures\37c8e000-6aa2-41ef-8f86-e0c2e60bfa42.tga
Delete $INSTDIR\skins\default\textures\37c8e000-6aa2-41ef-8f86-e0c2e60bfa42.tga
Delete $INSTDIR\skins\default\textures\35f217a3-f618-49cf-bbca-c86d486551a9.j2c
Delete $INSTDIR\skins\default\textures\35f217a3-f618-49cf-bbca-c86d486551a9.j2c
Delete $INSTDIR\skins\default\textures\34c9398d-bb78-4643-9633-46a2fa3e9637.tga
Delete $INSTDIR\skins\default\textures\34c9398d-bb78-4643-9633-46a2fa3e9637.tga
Delete $INSTDIR\skins\default\textures\335f8f14-f2db-db7c-1c04-734dc7657439.j2c
Delete $INSTDIR\skins\default\textures\335f8f14-f2db-db7c-1c04-734dc7657439.j2c
Delete $INSTDIR\skins\default\textures\303cd381-8560-7579-23f1-f0a880799740.j2c
Delete $INSTDIR\skins\default\textures\303cd381-8560-7579-23f1-f0a880799740.j2c
Delete $INSTDIR\skins\default\textures\30047cec-269d-408e-0c30-b2603b887268.j2c
Delete $INSTDIR\skins\default\textures\30047cec-269d-408e-0c30-b2603b887268.j2c
Delete $INSTDIR\skins\default\textures\2d784476-d0db-9979-0cff-9408745a7cf3.j2c
Delete $INSTDIR\skins\default\textures\2d784476-d0db-9979-0cff-9408745a7cf3.j2c
Delete $INSTDIR\skins\default\textures\2caf1179-7861-6ff3-4b7d-46e17780bdfa.j2c
Delete $INSTDIR\skins\default\textures\2caf1179-7861-6ff3-4b7d-46e17780bdfa.j2c
Delete $INSTDIR\skins\default\textures\2a4880b6-b7a3-690a-2049-bfbe38eafb9f.j2c
Delete $INSTDIR\skins\default\textures\2a4880b6-b7a3-690a-2049-bfbe38eafb9f.j2c
Delete $INSTDIR\skins\default\textures\29de489d-0491-fb00-7dab-f9e686d31e83.j2c
Delete $INSTDIR\skins\default\textures\29de489d-0491-fb00-7dab-f9e686d31e83.j2c
Delete $INSTDIR\skins\default\textures\28f0f9ca-0423-4d1b-9e76-616ffce99544.j2c
Delete $INSTDIR\skins\default\textures\28f0f9ca-0423-4d1b-9e76-616ffce99544.j2c
Delete $INSTDIR\skins\default\textures\2660b114-1d66-3cde-e148-ebc2d1f963d5.j2c
Delete $INSTDIR\skins\default\textures\2660b114-1d66-3cde-e148-ebc2d1f963d5.j2c
Delete $INSTDIR\skins\default\textures\1e63e323-5fe0-452e-92f8-b98bd0f764e3.j2c
Delete $INSTDIR\skins\default\textures\1e63e323-5fe0-452e-92f8-b98bd0f764e3.j2c
Delete $INSTDIR\skins\default\textures\19c76b49-c5f4-aeca-7cd8-17010f2969c3.j2c
Delete $INSTDIR\skins\default\textures\19c76b49-c5f4-aeca-7cd8-17010f2969c3.j2c
Delete $INSTDIR\skins\default\textures\18fb888b-e8f1-dce7-7da7-321d651ea6b0.j2c
Delete $INSTDIR\skins\default\textures\18fb888b-e8f1-dce7-7da7-321d651ea6b0.j2c
Delete $INSTDIR\skins\default\textures\179cdabd-398a-9b6b-1391-4dc333ba321f.j2c
Delete $INSTDIR\skins\default\textures\179cdabd-398a-9b6b-1391-4dc333ba321f.j2c
Delete $INSTDIR\skins\default\textures\13dd1d96-6836-461e-8a4c-36003065c59b.tga
Delete $INSTDIR\skins\default\textures\13dd1d96-6836-461e-8a4c-36003065c59b.tga
Delete $INSTDIR\skins\default\textures\11ee27f5-43c0-414e-afd5-d7f5688c351f.j2c
Delete $INSTDIR\skins\default\textures\11ee27f5-43c0-414e-afd5-d7f5688c351f.j2c
Delete $INSTDIR\skins\default\textures\111b39de-8928-4690-b7b2-e17d5c960277.tga
Delete $INSTDIR\skins\default\textures\111b39de-8928-4690-b7b2-e17d5c960277.tga
Delete $INSTDIR\skins\default\textures\10d2a01a-0818-84b9-4b96-c2eb63256519.j2c
Delete $INSTDIR\skins\default\textures\10d2a01a-0818-84b9-4b96-c2eb63256519.j2c
Delete $INSTDIR\skins\default\textures\0ff70ead-4562-45f9-9e8a-52b1a3286868.j2c
Delete $INSTDIR\skins\default\textures\0ff70ead-4562-45f9-9e8a-52b1a3286868.j2c
Delete $INSTDIR\skins\default\textures\0e82d24e-ed45-41bc-b090-94c97c1caab2.tga
Delete $INSTDIR\skins\default\textures\0e82d24e-ed45-41bc-b090-94c97c1caab2.tga
Delete $INSTDIR\skins\default\textures\0bc58228-74a0-7e83-89bc-5c23464bcec5.j2c
Delete $INSTDIR\skins\default\textures\0bc58228-74a0-7e83-89bc-5c23464bcec5.j2c
Delete $INSTDIR\skins\default\textures\0b444c3a-75c2-4891-9d1e-ac35c8d13d62.j2c
Delete $INSTDIR\skins\default\textures\0b444c3a-75c2-4891-9d1e-ac35c8d13d62.j2c
Delete $INSTDIR\skins\default\textures\0a94b42f-ec84-5f9c-14b7-1ef8505ceead.j2c
Delete $INSTDIR\skins\default\textures\0a94b42f-ec84-5f9c-14b7-1ef8505ceead.j2c
Delete $INSTDIR\skins\default\textures\09a324a8-acc1-d9cd-2cbd-7465d90d3a98.tga
Delete $INSTDIR\skins\default\textures\09a324a8-acc1-d9cd-2cbd-7465d90d3a98.tga
Delete $INSTDIR\skins\default\textures\07d0ea4c-af0c-aad1-dbbf-c24020ff2b80.tga
Delete $INSTDIR\skins\default\textures\07d0ea4c-af0c-aad1-dbbf-c24020ff2b80.tga
Delete $INSTDIR\skins\default\textures\073c9723-540c-5449-cdd4-0e87fdc159e3.j2c
Delete $INSTDIR\skins\default\textures\073c9723-540c-5449-cdd4-0e87fdc159e3.j2c
Delete $INSTDIR\skins\default\textures\058c75c0-a0d5-f2f8-43f3-e9699a89c2fc.j2c
Delete $INSTDIR\skins\default\textures\058c75c0-a0d5-f2f8-43f3-e9699a89c2fc.j2c
Delete $INSTDIR\skins\default\textures\0498c309-5306-43cd-82a2-ae31d096cdef.tga
Delete $INSTDIR\skins\default\textures\0498c309-5306-43cd-82a2-ae31d096cdef.tga
Delete $INSTDIR\skins\default\textures\041ee5a0-cb6a-9ac5-6e49-41e9320507d5.j2c
Delete $INSTDIR\skins\default\textures\041ee5a0-cb6a-9ac5-6e49-41e9320507d5.j2c
Delete $INSTDIR\skins\default\textures\0187babf-6c0d-5891-ebed-4ecab1426683.j2c
Delete $INSTDIR\skins\default\textures\0187babf-6c0d-5891-ebed-4ecab1426683.j2c
Delete $INSTDIR\skins\default\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
Delete $INSTDIR\skins\default\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
Delete $INSTDIR\skins\default\html\unabletoconnect.png
Delete $INSTDIR\skins\default\html\unabletoconnect.png
Delete $INSTDIR\skins\default\html\btn_purplepill_bg.png
Delete $INSTDIR\skins\default\html\btn_purplepill_bg.png
Delete $INSTDIR\app_settings\windlight\water\Valdez.xml
Delete $INSTDIR\app_settings\windlight\water\Valdez.xml
Delete $INSTDIR\app_settings\windlight\water\Second%20Plague.xml
Delete $INSTDIR\app_settings\windlight\water\Second%20Plague.xml
Delete $INSTDIR\app_settings\windlight\water\SNAKE%21%21%21.xml
Delete $INSTDIR\app_settings\windlight\water\SNAKE%21%21%21.xml
Delete $INSTDIR\app_settings\windlight\water\Pond.xml
Delete $INSTDIR\app_settings\windlight\water\Pond.xml
Delete $INSTDIR\app_settings\windlight\water\Murky.xml
Delete $INSTDIR\app_settings\windlight\water\Murky.xml
Delete $INSTDIR\app_settings\windlight\water\Glassy.xml
Delete $INSTDIR\app_settings\windlight\water\Glassy.xml
Delete $INSTDIR\app_settings\windlight\water\Default.xml
Delete $INSTDIR\app_settings\windlight\water\Default.xml
Delete $INSTDIR\app_settings\windlight\skies\Sheer%20Surreality.xml
Delete $INSTDIR\app_settings\windlight\skies\Sheer%20Surreality.xml
Delete $INSTDIR\app_settings\windlight\skies\Sailor%27s%20Delight.xml
Delete $INSTDIR\app_settings\windlight\skies\Sailor%27s%20Delight.xml
Delete $INSTDIR\app_settings\windlight\skies\Purple.xml
Delete $INSTDIR\app_settings\windlight\skies\Purple.xml
Delete $INSTDIR\app_settings\windlight\skies\Pirate.xml
Delete $INSTDIR\app_settings\windlight\skies\Pirate.xml
Delete $INSTDIR\app_settings\windlight\skies\Night.xml
Delete $INSTDIR\app_settings\windlight\skies\Night.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%204.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%204.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%203.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%203.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%202.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%202.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%201.xml
Delete $INSTDIR\app_settings\windlight\skies\Midday%201.xml
Delete $INSTDIR\app_settings\windlight\skies\Incongruent%20Truths.xml
Delete $INSTDIR\app_settings\windlight\skies\Incongruent%20Truths.xml
Delete $INSTDIR\app_settings\windlight\skies\Ghost.xml
Delete $INSTDIR\app_settings\windlight\skies\Ghost.xml
Delete $INSTDIR\app_settings\windlight\skies\Gelatto.xml
Delete $INSTDIR\app_settings\windlight\skies\Gelatto.xml
Delete $INSTDIR\app_settings\windlight\skies\Funky%20Funky.xml
Delete $INSTDIR\app_settings\windlight\skies\Funky%20Funky.xml
Delete $INSTDIR\app_settings\windlight\skies\Funky%20Funky%20Funky.xml
Delete $INSTDIR\app_settings\windlight\skies\Funky%20Funky%20Funky.xml
Delete $INSTDIR\app_settings\windlight\skies\Foggy.xml
Delete $INSTDIR\app_settings\windlight\skies\Foggy.xml
Delete $INSTDIR\app_settings\windlight\skies\Fluffy%20Big%20Clouds.xml
Delete $INSTDIR\app_settings\windlight\skies\Fluffy%20Big%20Clouds.xml
Delete $INSTDIR\app_settings\windlight\skies\Fine%20Day.xml
Delete $INSTDIR\app_settings\windlight\skies\Fine%20Day.xml
Delete $INSTDIR\app_settings\windlight\skies\Desert%20Sunset.xml
Delete $INSTDIR\app_settings\windlight\skies\Desert%20Sunset.xml
Delete $INSTDIR\app_settings\windlight\skies\Default.xml
Delete $INSTDIR\app_settings\windlight\skies\Default.xml
Delete $INSTDIR\app_settings\windlight\skies\Coastal%20Sunset.xml
Delete $INSTDIR\app_settings\windlight\skies\Coastal%20Sunset.xml
Delete $INSTDIR\app_settings\windlight\skies\Coastal%20Afternoon.xml
Delete $INSTDIR\app_settings\windlight\skies\Coastal%20Afternoon.xml
Delete $INSTDIR\app_settings\windlight\skies\Blue%20Midday.xml
Delete $INSTDIR\app_settings\windlight\skies\Blue%20Midday.xml
Delete $INSTDIR\app_settings\windlight\skies\Blizzard.xml
Delete $INSTDIR\app_settings\windlight\skies\Blizzard.xml
Delete $INSTDIR\app_settings\windlight\skies\Barcelona.xml
Delete $INSTDIR\app_settings\windlight\skies\Barcelona.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D9PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D9PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D9AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D9AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D6PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D6PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D6AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D6AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D3PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D3PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D3AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D3AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D12PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D12PM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D12AM.xml
Delete $INSTDIR\app_settings\windlight\skies\A%2D12AM.xml
Delete $INSTDIR\app_settings\windlight\days\Default.xml
Delete $INSTDIR\app_settings\windlight\days\Default.xml
Delete $INSTDIR\skins\silver\colors_base.xml
Delete $INSTDIR\skins\silver\colors_base.xml
Delete $INSTDIR\skins\silver\colors.xml
Delete $INSTDIR\skins\silver\colors.xml
Delete $INSTDIR\skins\default\colors_base.xml
Delete $INSTDIR\skins\default\colors_base.xml
Delete $INSTDIR\skins\default\colors.xml
Delete $INSTDIR\skins\default\colors.xml
Delete $INSTDIR\app_settings\windlight\postprocesseffects.xml
Delete $INSTDIR\app_settings\windlight\postprocesseffects.xml
Delete $INSTDIR\app_settings\windlight\clouds2.tga
Delete $INSTDIR\app_settings\windlight\clouds2.tga
Delete $INSTDIR\app_settings\shaders\shader_heirarchy.txt
Delete $INSTDIR\app_settings\shaders\shader_heirarchy.txt
Delete $INSTDIR\skins\paths.xml
Delete $INSTDIR\skins\paths.xml
Delete $INSTDIR\llplugin\ssleay32.dll
Delete $INSTDIR\llplugin\ssleay32.dll
Delete $INSTDIR\llplugin\qtwebkit4.dll
Delete $INSTDIR\llplugin\qtwebkit4.dll
Delete $INSTDIR\llplugin\qtopengl4.dll
Delete $INSTDIR\llplugin\qtopengl4.dll
Delete $INSTDIR\llplugin\qtnetwork4.dll
Delete $INSTDIR\llplugin\qtnetwork4.dll
Delete $INSTDIR\llplugin\qtgui4.dll
Delete $INSTDIR\llplugin\qtgui4.dll
Delete $INSTDIR\llplugin\qtcore4.dll
Delete $INSTDIR\llplugin\qtcore4.dll
Delete $INSTDIR\llplugin\media_plugin_webkit.dll
Delete $INSTDIR\llplugin\media_plugin_quicktime.dll
Delete $INSTDIR\llplugin\libeay32.dll
Delete $INSTDIR\llplugin\libeay32.dll
Delete $INSTDIR\llplugin\imageformats\qtiff4.dll
Delete $INSTDIR\llplugin\imageformats\qtiff4.dll
Delete $INSTDIR\llplugin\imageformats\qsvg4.dll
Delete $INSTDIR\llplugin\imageformats\qsvg4.dll
Delete $INSTDIR\llplugin\imageformats\qmng4.dll
Delete $INSTDIR\llplugin\imageformats\qmng4.dll
Delete $INSTDIR\llplugin\imageformats\qjpeg4.dll
Delete $INSTDIR\llplugin\imageformats\qjpeg4.dll
Delete $INSTDIR\llplugin\imageformats\qico4.dll
Delete $INSTDIR\llplugin\imageformats\qico4.dll
Delete $INSTDIR\llplugin\imageformats\qgif4.dll
Delete $INSTDIR\llplugin\imageformats\qgif4.dll
Delete $INSTDIR\fonts\Readme.txt
Delete $INSTDIR\fonts\Readme.txt
Delete $INSTDIR\fonts\MtBkLfRg.ttf
Delete $INSTDIR\fonts\MtBkLfRg.ttf
Delete $INSTDIR\fonts\MtBdLfRg.ttf
Delete $INSTDIR\fonts\MtBdLfRg.ttf
Delete $INSTDIR\fonts\DejaVuSansMono.ttf
Delete $INSTDIR\fonts\DejaVuSansMono.ttf
Delete $INSTDIR\fonts\DejaVuSansCondensed.ttf
Delete $INSTDIR\fonts\DejaVuSansCondensed.ttf
Delete $INSTDIR\fonts\DejaVu-license.txt
Delete $INSTDIR\fonts\DejaVu-license.txt
Delete $INSTDIR\character\upperbodyfreckles_alpha.tga
Delete $INSTDIR\character\upperbodyfreckles_alpha.tga
Delete $INSTDIR\character\upperbody_shading_alpha.tga
Delete $INSTDIR\character\upperbody_shading_alpha.tga
Delete $INSTDIR\character\upperbody_highlights_alpha.tga
Delete $INSTDIR\character\upperbody_highlights_alpha.tga
Delete $INSTDIR\character\upperbody_color.tga
Delete $INSTDIR\character\upperbody_color.tga
Delete $INSTDIR\character\undershirt_trial_female.tga
Delete $INSTDIR\character\undershirt_trial_female.tga
Delete $INSTDIR\character\underpants_trial_male.tga
Delete $INSTDIR\character\underpants_trial_male.tga
Delete $INSTDIR\character\underpants_trial_female.tga
Delete $INSTDIR\character\underpants_trial_female.tga
Delete $INSTDIR\character\skirt_slit_right_alpha.tga
Delete $INSTDIR\character\skirt_slit_right_alpha.tga
Delete $INSTDIR\character\skirt_slit_left_alpha.tga
Delete $INSTDIR\character\skirt_slit_left_alpha.tga
Delete $INSTDIR\character\skirt_slit_front_alpha.tga
Delete $INSTDIR\character\skirt_slit_front_alpha.tga
Delete $INSTDIR\character\skirt_slit_back_alpha.tga
Delete $INSTDIR\character\skirt_slit_back_alpha.tga
Delete $INSTDIR\character\skirt_length_alpha.tga
Delete $INSTDIR\character\skirt_length_alpha.tga
Delete $INSTDIR\character\shoe_height_alpha.tga
Delete $INSTDIR\character\shoe_height_alpha.tga
Delete $INSTDIR\character\shirt_sleeve_alpha.tga
Delete $INSTDIR\character\shirt_sleeve_alpha.tga
Delete $INSTDIR\character\shirt_collar_back_alpha.tga
Delete $INSTDIR\character\shirt_collar_back_alpha.tga
Delete $INSTDIR\character\shirt_collar_alpha.tga
Delete $INSTDIR\character\shirt_collar_alpha.tga
Delete $INSTDIR\character\shirt_bottom_alpha.tga
Delete $INSTDIR\character\shirt_bottom_alpha.tga
Delete $INSTDIR\character\rouge_alpha.tga
Delete $INSTDIR\character\rouge_alpha.tga
Delete $INSTDIR\character\rosyface_alpha.tga
Delete $INSTDIR\character\rosyface_alpha.tga
Delete $INSTDIR\character\pants_waist_alpha.tga
Delete $INSTDIR\character\pants_waist_alpha.tga
Delete $INSTDIR\character\pants_length_alpha.tga
Delete $INSTDIR\character\pants_length_alpha.tga
Delete $INSTDIR\character\nailpolish_alpha.tga
Delete $INSTDIR\character\nailpolish_alpha.tga
Delete $INSTDIR\character\lowerbody_shading_alpha.tga
Delete $INSTDIR\character\lowerbody_shading_alpha.tga
Delete $INSTDIR\character\lowerbody_highlights_alpha.tga
Delete $INSTDIR\character\lowerbody_highlights_alpha.tga
Delete $INSTDIR\character\lowerbody_color.tga
Delete $INSTDIR\character\lowerbody_color.tga
Delete $INSTDIR\character\lipstick_alpha.tga
Delete $INSTDIR\character\lipstick_alpha.tga
Delete $INSTDIR\character\lips_mask.tga
Delete $INSTDIR\character\lips_mask.tga
Delete $INSTDIR\character\lipgloss_alpha.tga
Delete $INSTDIR\character\lipgloss_alpha.tga
Delete $INSTDIR\character\jacket_open_upper_alpha.tga
Delete $INSTDIR\character\jacket_open_upper_alpha.tga
Delete $INSTDIR\character\jacket_open_lower_alpha.tga
Delete $INSTDIR\character\jacket_open_lower_alpha.tga
Delete $INSTDIR\character\jacket_length_upper_alpha.tga
Delete $INSTDIR\character\jacket_length_upper_alpha.tga
Delete $INSTDIR\character\jacket_length_lower_alpha.tga
Delete $INSTDIR\character\jacket_length_lower_alpha.tga
Delete $INSTDIR\character\head_skingrain.tga
Delete $INSTDIR\character\head_skingrain.tga
Delete $INSTDIR\character\head_shading_alpha.tga
Delete $INSTDIR\character\head_shading_alpha.tga
Delete $INSTDIR\character\head_highlights_alpha.tga
Delete $INSTDIR\character\head_highlights_alpha.tga
Delete $INSTDIR\character\head_hair.tga
Delete $INSTDIR\character\head_hair.tga
Delete $INSTDIR\character\head_color.tga
Delete $INSTDIR\character\head_color.tga
Delete $INSTDIR\character\head_alpha.tga
Delete $INSTDIR\character\head_alpha.tga
Delete $INSTDIR\character\gloves_fingers_alpha.tga
Delete $INSTDIR\character\gloves_fingers_alpha.tga
Delete $INSTDIR\character\glove_length_alpha.tga
Delete $INSTDIR\character\glove_length_alpha.tga
Delete $INSTDIR\character\genepool.xml
Delete $INSTDIR\character\genepool.xml
Delete $INSTDIR\character\freckles_alpha.tga
Delete $INSTDIR\character\freckles_alpha.tga
Delete $INSTDIR\character\facehair_soulpatch_alpha.tga
Delete $INSTDIR\character\facehair_soulpatch_alpha.tga
Delete $INSTDIR\character\facehair_sideburns_alpha.tga
Delete $INSTDIR\character\facehair_sideburns_alpha.tga
Delete $INSTDIR\character\facehair_moustache_alpha.tga
Delete $INSTDIR\character\facehair_moustache_alpha.tga
Delete $INSTDIR\character\facehair_chincurtains_alpha.tga
Delete $INSTDIR\character\facehair_chincurtains_alpha.tga
Delete $INSTDIR\character\eyewhite.tga
Delete $INSTDIR\character\eyewhite.tga
Delete $INSTDIR\character\eyeshadow_outer_alpha.tga
Delete $INSTDIR\character\eyeshadow_outer_alpha.tga
Delete $INSTDIR\character\eyeshadow_inner_alpha.tga
Delete $INSTDIR\character\eyeshadow_inner_alpha.tga
Delete $INSTDIR\character\eyeliner_alpha.tga
Delete $INSTDIR\character\eyeliner_alpha.tga
Delete $INSTDIR\character\eyebrows_alpha.tga
Delete $INSTDIR\character\eyebrows_alpha.tga
Delete $INSTDIR\character\bump_upperbody_base.tga
Delete $INSTDIR\character\bump_upperbody_base.tga
Delete $INSTDIR\character\bump_shirt_wrinkles.tga
Delete $INSTDIR\character\bump_shirt_wrinkles.tga
Delete $INSTDIR\character\bump_pants_wrinkles.tga
Delete $INSTDIR\character\bump_pants_wrinkles.tga
Delete $INSTDIR\character\bump_lowerbody_base.tga
Delete $INSTDIR\character\bump_lowerbody_base.tga
Delete $INSTDIR\character\bump_head_base.tga
Delete $INSTDIR\character\bump_head_base.tga
Delete $INSTDIR\character\bump_face_wrinkles.tga
Delete $INSTDIR\character\bump_face_wrinkles.tga
Delete $INSTDIR\character\bodyfreckles_alpha.tga
Delete $INSTDIR\character\bodyfreckles_alpha.tga
Delete $INSTDIR\character\body_skingrain.tga
Delete $INSTDIR\character\body_skingrain.tga
Delete $INSTDIR\character\blush_alpha.tga
Delete $INSTDIR\character\blush_alpha.tga
Delete $INSTDIR\character\avatar_upper_body_4.llm
Delete $INSTDIR\character\avatar_upper_body_4.llm
Delete $INSTDIR\character\avatar_upper_body_3.llm
Delete $INSTDIR\character\avatar_upper_body_3.llm
Delete $INSTDIR\character\avatar_upper_body_2.llm
Delete $INSTDIR\character\avatar_upper_body_2.llm
Delete $INSTDIR\character\avatar_upper_body_1.llm
Delete $INSTDIR\character\avatar_upper_body_1.llm
Delete $INSTDIR\character\avatar_upper_body.llm
Delete $INSTDIR\character\avatar_upper_body.llm
Delete $INSTDIR\character\avatar_skirt_4.llm
Delete $INSTDIR\character\avatar_skirt_4.llm
Delete $INSTDIR\character\avatar_skirt_3.llm
Delete $INSTDIR\character\avatar_skirt_3.llm
Delete $INSTDIR\character\avatar_skirt_2.llm
Delete $INSTDIR\character\avatar_skirt_2.llm
Delete $INSTDIR\character\avatar_skirt_1.llm
Delete $INSTDIR\character\avatar_skirt_1.llm
Delete $INSTDIR\character\avatar_skirt.llm
Delete $INSTDIR\character\avatar_skirt.llm
Delete $INSTDIR\character\avatar_skeleton.xml
Delete $INSTDIR\character\avatar_skeleton.xml
Delete $INSTDIR\character\avatar_lower_body_4.llm
Delete $INSTDIR\character\avatar_lower_body_4.llm
Delete $INSTDIR\character\avatar_lower_body_3.llm
Delete $INSTDIR\character\avatar_lower_body_3.llm
Delete $INSTDIR\character\avatar_lower_body_2.llm
Delete $INSTDIR\character\avatar_lower_body_2.llm
Delete $INSTDIR\character\avatar_lower_body_1.llm
Delete $INSTDIR\character\avatar_lower_body_1.llm
Delete $INSTDIR\character\avatar_lower_body.llm
Delete $INSTDIR\character\avatar_lower_body.llm
Delete $INSTDIR\character\avatar_lad.xml
Delete $INSTDIR\character\avatar_lad.xml
Delete $INSTDIR\character\avatar_head_4.llm
Delete $INSTDIR\character\avatar_head_4.llm
Delete $INSTDIR\character\avatar_head_3.llm
Delete $INSTDIR\character\avatar_head_3.llm
Delete $INSTDIR\character\avatar_head_2.llm
Delete $INSTDIR\character\avatar_head_2.llm
Delete $INSTDIR\character\avatar_head_1.llm
Delete $INSTDIR\character\avatar_head_1.llm
Delete $INSTDIR\character\avatar_head.llm
Delete $INSTDIR\character\avatar_head.llm
Delete $INSTDIR\character\avatar_hair_5.llm
Delete $INSTDIR\character\avatar_hair_5.llm
Delete $INSTDIR\character\avatar_hair_4.llm
Delete $INSTDIR\character\avatar_hair_4.llm
Delete $INSTDIR\character\avatar_hair_3.llm
Delete $INSTDIR\character\avatar_hair_3.llm
Delete $INSTDIR\character\avatar_hair_2.llm
Delete $INSTDIR\character\avatar_hair_2.llm
Delete $INSTDIR\character\avatar_hair_1.llm
Delete $INSTDIR\character\avatar_hair_1.llm
Delete $INSTDIR\character\avatar_hair.llm
Delete $INSTDIR\character\avatar_hair.llm
Delete $INSTDIR\character\avatar_eyelashes.llm
Delete $INSTDIR\character\avatar_eyelashes.llm
Delete $INSTDIR\character\avatar_eye_1.llm
Delete $INSTDIR\character\avatar_eye_1.llm
Delete $INSTDIR\character\avatar_eye.llm
Delete $INSTDIR\character\avatar_eye.llm
Delete $INSTDIR\character\attentionsN.xml
Delete $INSTDIR\character\attentionsN.xml
Delete $INSTDIR\character\attentions.xml
Delete $INSTDIR\character\attentions.xml
Delete $INSTDIR\app_settings\viewerart.xml
Delete $INSTDIR\app_settings\viewerart.xml
Delete $INSTDIR\app_settings\ultra_graphics.xml
Delete $INSTDIR\app_settings\ultra_graphics.xml
Delete $INSTDIR\app_settings\trees.xml
Delete $INSTDIR\app_settings\trees.xml
Delete $INSTDIR\app_settings\std_bump.ini
Delete $INSTDIR\app_settings\std_bump.ini
Delete $INSTDIR\app_settings\static_index.db2
Delete $INSTDIR\app_settings\static_index.db2
Delete $INSTDIR\app_settings\static_data.db2
Delete $INSTDIR\app_settings\static_data.db2
Delete $INSTDIR\app_settings\settings_per_account.xml
Delete $INSTDIR\app_settings\settings_per_account.xml
Delete $INSTDIR\app_settings\settings_files.xml
Delete $INSTDIR\app_settings\settings_files.xml
Delete $INSTDIR\app_settings\settings_crash_behavior.xml
Delete $INSTDIR\app_settings\settings_crash_behavior.xml
Delete $INSTDIR\app_settings\settings.xml
Delete $INSTDIR\app_settings\settings.xml
Delete $INSTDIR\app_settings\client_definitions.xml
Delete $INSTDIR\app_settings\client_definitions.xml
Delete $INSTDIR\app_settings\mime_types.xml
Delete $INSTDIR\app_settings\mime_types.xml
Delete $INSTDIR\app_settings\mid_graphics.xml
Delete $INSTDIR\app_settings\mid_graphics.xml
Delete $INSTDIR\app_settings\low_graphics.xml
Delete $INSTDIR\app_settings\low_graphics.xml
Delete $INSTDIR\app_settings\logcontrol.xml
Delete $INSTDIR\app_settings\logcontrol.xml
Delete $INSTDIR\app_settings\keywords.ini
Delete $INSTDIR\app_settings\keywords.ini
Delete $INSTDIR\app_settings\keys.ini
Delete $INSTDIR\app_settings\keys.ini
Delete $INSTDIR\app_settings\high_graphics.xml
Delete $INSTDIR\app_settings\high_graphics.xml
Delete $INSTDIR\app_settings\grass.xml
Delete $INSTDIR\app_settings\grass.xml
Delete $INSTDIR\app_settings\cmd_line.xml
Delete $INSTDIR\app_settings\cmd_line.xml
Delete $INSTDIR\app_settings\anim.ini
Delete $INSTDIR\app_settings\anim.ini
Delete $INSTDIR\app_settings\CA.pem
Delete $INSTDIR\app_settings\CA.pem
Delete $INSTDIR\wrap_oal.dll
Delete $INSTDIR\wrap_oal.dll
Delete $INSTDIR\win_crash_logger.exe
Delete $INSTDIR\vivoxsdk.dll
Delete $INSTDIR\vivoxsdk.dll
Delete $INSTDIR\updater.exe
Delete $INSTDIR\skins\default\xui\en-us\mime_types.xml
Delete $INSTDIR\skins\default\xui\en-us\mime_types.xml
Delete $INSTDIR\ortp.dll
Delete $INSTDIR\ortp.dll
Delete $INSTDIR\openjpeg.dll
Delete $INSTDIR\openjpeg.dll
Delete $INSTDIR\msvcr80.dll
Delete $INSTDIR\msvcp80.dll
Delete $INSTDIR\licenses.txt
Delete $INSTDIR\licenses.txt
Delete $INSTDIR\gpu_table.txt
Delete $INSTDIR\gpu_table.txt
Delete $INSTDIR\fmod.dll
Delete $INSTDIR\fmod.dll
Delete $INSTDIR\fmod.dll
Delete $INSTDIR\featuretable.txt
Delete $INSTDIR\featuretable.txt
Delete $INSTDIR\dbghelp.dll
Delete $INSTDIR\dbghelp.dll
Delete $INSTDIR\app_settings\message_template.msg
Delete $INSTDIR\app_settings\message_template.msg
Delete $INSTDIR\app_settings\message.xml
Delete $INSTDIR\app_settings\message.xml
Delete $INSTDIR\alut.dll
Delete $INSTDIR\alut.dll
Delete $INSTDIR\SLVoice.exe
Delete $INSTDIR\SLVoice.exe
Delete $INSTDIR\SLPlugin.exe
Delete $INSTDIR\Microsoft.VC80.CRT.manifest
Delete $INSTDIR\Ascent.exe.config
Delete $INSTDIR\Ascent.exe
RMDir $INSTDIR\skins\default\html\zh\loading-error
RMDir $INSTDIR\skins\default\html\pt\loading-error
RMDir $INSTDIR\skins\default\html\ko\loading-error
RMDir $INSTDIR\skins\default\html\ja\loading-error
RMDir $INSTDIR\skins\default\html\fr\loading-error
RMDir $INSTDIR\skins\default\html\es\loading-error
RMDir $INSTDIR\skins\default\html\en-us\loading-error
RMDir $INSTDIR\skins\default\html\en-us\loading
RMDir $INSTDIR\skins\default\html\de\loading-error
RMDir $INSTDIR\skins\silver\xui\en-us
RMDir $INSTDIR\skins\default\xui\zh
RMDir $INSTDIR\skins\default\xui\uk
RMDir $INSTDIR\skins\default\xui\tr
RMDir $INSTDIR\skins\default\xui\ru
RMDir $INSTDIR\skins\default\xui\pt
RMDir $INSTDIR\skins\default\xui\pl
RMDir $INSTDIR\skins\default\xui\ko
RMDir $INSTDIR\skins\default\xui\ja
RMDir $INSTDIR\skins\default\xui\it
RMDir $INSTDIR\skins\default\xui\hu
RMDir $INSTDIR\skins\default\xui\fr
RMDir $INSTDIR\skins\default\xui\es
RMDir $INSTDIR\skins\default\xui\en-us
RMDir $INSTDIR\skins\default\xui\de
RMDir $INSTDIR\skins\default\xui\da
RMDir $INSTDIR\skins\default\html\zh
RMDir $INSTDIR\skins\default\html\pt
RMDir $INSTDIR\skins\default\html\ko
RMDir $INSTDIR\skins\default\html\ja
RMDir $INSTDIR\skins\default\html\fr
RMDir $INSTDIR\skins\default\html\es
RMDir $INSTDIR\skins\default\html\en-us
RMDir $INSTDIR\skins\default\html\de
RMDir $INSTDIR\app_settings\shaders\class3\lighting
RMDir $INSTDIR\app_settings\shaders\class3\avatar
RMDir $INSTDIR\app_settings\shaders\class2\windlight
RMDir $INSTDIR\app_settings\shaders\class2\objects
RMDir $INSTDIR\app_settings\shaders\class2\lighting
RMDir $INSTDIR\app_settings\shaders\class2\environment
RMDir $INSTDIR\app_settings\shaders\class2\effects
RMDir $INSTDIR\app_settings\shaders\class2\avatar
RMDir $INSTDIR\app_settings\shaders\class1\windlight
RMDir $INSTDIR\app_settings\shaders\class1\objects
RMDir $INSTDIR\app_settings\shaders\class1\lighting
RMDir $INSTDIR\app_settings\shaders\class1\interface
RMDir $INSTDIR\app_settings\shaders\class1\environment
RMDir $INSTDIR\app_settings\shaders\class1\effects
RMDir $INSTDIR\app_settings\shaders\class1\deferred
RMDir $INSTDIR\app_settings\shaders\class1\avatar
RMDir $INSTDIR\skins\silver\xui
RMDir $INSTDIR\skins\silver\textures
RMDir $INSTDIR\skins\default\xui
RMDir $INSTDIR\skins\default\textures
RMDir $INSTDIR\skins\default\html
RMDir $INSTDIR\app_settings\windlight\water
RMDir $INSTDIR\app_settings\windlight\skies
RMDir $INSTDIR\app_settings\windlight\days
RMDir $INSTDIR\app_settings\shaders\class3
RMDir $INSTDIR\app_settings\shaders\class2
RMDir $INSTDIR\app_settings\shaders\class1
RMDir $INSTDIR\skins\silver
RMDir $INSTDIR\skins\default
RMDir $INSTDIR\llplugin\imageformats
RMDir $INSTDIR\app_settings\windlight
RMDir $INSTDIR\app_settings\shaders
RMDir $INSTDIR\skins
RMDir $INSTDIR\llplugin
RMDir $INSTDIR\fonts
RMDir $INSTDIR\character
RMDir $INSTDIR\app_settings
RMDir $INSTDIR\.


;; Optional/obsolete files.  Delete won't fail if they don't exist.
Delete "$INSTDIR\dronesettings.ini"
Delete "$INSTDIR\message_template.msg"
Delete "$INSTDIR\newview.pdb"
Delete "$INSTDIR\newview.map"
Delete "$INSTDIR\SecondLife.pdb"
Delete "$INSTDIR\SecondLife.map"
Delete "$INSTDIR\comm.dat"
Delete "$INSTDIR\*.glsl"
Delete "$INSTDIR\motions\*.lla"
Delete "$INSTDIR\trial\*.html"
Delete "$INSTDIR\newview.exe"
;; Remove entire help directory
Delete "$INSTDIR\help\Advanced\*"
RMDir  "$INSTDIR\help\Advanced"
Delete "$INSTDIR\help\basics\*"
RMDir  "$INSTDIR\help\basics"
Delete "$INSTDIR\help\Concepts\*"
RMDir  "$INSTDIR\help\Concepts"
Delete "$INSTDIR\help\welcome\*"
RMDir  "$INSTDIR\help\welcome"
Delete "$INSTDIR\help\*"
RMDir  "$INSTDIR\help"

Delete "$INSTDIR\uninst.exe"
RMDir "$INSTDIR"

IfFileExists "$INSTDIR" FOLDERFOUND NOFOLDER

FOLDERFOUND:
  MessageBox MB_YESNO $(DeleteProgramFilesMB) IDNO NOFOLDER
  RMDir /r "$INSTDIR"

NOFOLDER:

FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Uninstall settings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
UninstallText $(UninstallTextMsg)
ShowUninstDetails show

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Uninstall section
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Section Uninstall

; Start with some default values.
StrCpy $INSTFLAGS ""
StrCpy $INSTPROG "${INSTNAME}"
StrCpy $INSTEXE "${INSTEXE}"
StrCpy $INSTSHORTCUT "${SHORTCUT}"
Call un.CheckIfAdministrator		; Make sure the user can install/uninstall

; uninstall for all users (if you change this, change it in the install as well)
SetShellVarContext all			

; Make sure we're not running
Call un.CloseSecondLife

; Clean up registry keys and subkeys (these should all be !defines somewhere)
DeleteRegKey HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG"
DeleteRegKey HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\$INSTPROG"

; Clean up shortcuts
Delete "$SMPROGRAMS\$INSTSHORTCUT\*.*"
RMDir  "$SMPROGRAMS\$INSTSHORTCUT"

Delete "$DESKTOP\$INSTSHORTCUT.lnk"
Delete "$INSTDIR\$INSTSHORTCUT.lnk"
Delete "$INSTDIR\Uninstall $INSTSHORTCUT.lnk"

; Clean up cache and log files.
; Leave them in-place for non AGNI installs.

!ifdef UNINSTALL_SETTINGS
Call un.DocumentsAndSettingsFolder
!endif

; remove stored password on uninstall
Call un.RemovePassword

Call un.ProgramFiles

SectionEnd 				; end of uninstall section


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; (From the NSIS documentation, JC)
; GetWindowsVersion
;
; Based on Yazno's function, http://yazno.tripod.com/powerpimpit/
; Updated by Joost Verburg
;
; Returns on top of stack
;
; Windows Version (95, 98, ME, NT x.x, 2000, XP, 2003)
; or
; '' (Unknown Windows Version)
;
; Usage:
;   Call GetWindowsVersion
;   Pop $R0
;   ; at this point $R0 is "NT 4.0" or whatnot
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function GetWindowsVersion
 
   Push $R0
   Push $R1
 
   ReadRegStr $R0 HKLM \
   "SOFTWARE\Microsoft\Windows NT\CurrentVersion" CurrentVersion

   IfErrors 0 lbl_winnt
   
   ; we are not NT
   ReadRegStr $R0 HKLM \
   "SOFTWARE\Microsoft\Windows\CurrentVersion" VersionNumber
 
   StrCpy $R1 $R0 1
   StrCmp $R1 '4' 0 lbl_error
 
   StrCpy $R1 $R0 3
 
   StrCmp $R1 '4.0' lbl_win32_95
   StrCmp $R1 '4.9' lbl_win32_ME lbl_win32_98
 
   lbl_win32_95:
     StrCpy $R0 '95'
   Goto lbl_done
 
   lbl_win32_98:
     StrCpy $R0 '98'
   Goto lbl_done
 
   lbl_win32_ME:
     StrCpy $R0 'ME'
   Goto lbl_done
 
   lbl_winnt:
 
   StrCpy $R1 $R0 1
 
   StrCmp $R1 '3' lbl_winnt_x
   StrCmp $R1 '4' lbl_winnt_x
 
   StrCpy $R1 $R0 3
 
   StrCmp $R1 '5.0' lbl_winnt_2000
   StrCmp $R1 '5.1' lbl_winnt_XP
   StrCmp $R1 '5.2' lbl_winnt_2003 lbl_error
 
   lbl_winnt_x:
     StrCpy $R0 "NT $R0" 6
   Goto lbl_done
 
   lbl_winnt_2000:
     Strcpy $R0 '2000'
   Goto lbl_done
 
   lbl_winnt_XP:
     Strcpy $R0 'XP'
   Goto lbl_done
 
   lbl_winnt_2003:
     Strcpy $R0 '2003'
   Goto lbl_done
 
   lbl_error:
     Strcpy $R0 ''
   lbl_done:
 
   Pop $R1
   Exch $R0
 
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Note: to add new languages, add a language file include to the list 
;;  at the top of this file, add an entry to the menu and then add an 
;;  entry to the language ID selector below
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function .onInit
    Push $0
    ${GetParameters} $COMMANDLINE              ; get our command line
    ${GetOptions} $COMMANDLINE "/LANGID=" $0   ; /LANGID=1033 implies US English
    ; If no language (error), then proceed
    IfErrors lbl_check_silent
    ; No error means we got a language, so use it
    StrCpy $LANGUAGE $0
    Goto lbl_return

lbl_check_silent:
    ; For silent installs, no language prompt, use default
    IfSilent lbl_return
    
	; If we currently have a version of SL installed, default to the language of that install
    ; Otherwise don't change $LANGUAGE and it will default to the OS UI language.
	ReadRegStr $0 HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\${INSTNAME}" "InstallerLanguage"
    IfErrors lbl_build_menu
	StrCpy $LANGUAGE $0

lbl_build_menu:
	Push ""
    # Use separate file so labels can be UTF-16 but we can still merge changes
    # into this ASCII file. JC
    !include "E:\ascentviewer\indra\newview\installers\windows\language_menu.nsi"
    
	Push A ; A means auto count languages for the auto count to work the first empty push (Push "") must remain
	LangDLL::LangDialog $(InstallerLanguageTitle) $(SelectInstallerLanguage)
	Pop $0
	StrCmp $0 "cancel" 0 +2
		Abort
    StrCpy $LANGUAGE $0

	; save language in registry		
	WriteRegStr HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\${INSTNAME}" "InstallerLanguage" $LANGUAGE
lbl_return:
    Pop $0
    Return
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Function un.onInit
	; read language from registry and set for uninstaller
    ; Key will be removed on successful uninstall
	ReadRegStr $0 HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\${INSTNAME}" "InstallerLanguage"
    IfErrors lbl_end
	StrCpy $LANGUAGE $0
lbl_end:
    Return
FunctionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MAIN SECTION
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Section ""						; (default section)

SetShellVarContext all			; install for all users (if you change this, change it in the uninstall as well)

; Start with some default values.
StrCpy $INSTFLAGS "${INSTFLAGS}"
StrCpy $INSTPROG "${INSTNAME}"
StrCpy $INSTEXE "${INSTEXE}"
StrCpy $INSTSHORTCUT "${SHORTCUT}"

Call CheckWindowsVersion		; warn if on Windows 98/ME
Call CheckIfAdministrator		; Make sure the user can install/uninstall
Call CheckIfAlreadyCurrent		; Make sure that we haven't already installed this version
Call CloseSecondLife			; Make sure we're not running
Call CheckNetworkConnection		; ping secondlife.com

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Don't remove cache files during a regular install, removing the inventory cache on upgrades results in lots of damage to the servers.
;Call RemoveCacheFiles			; Installing over removes potentially corrupted
								; VFS and cache files.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Need to clean out shader files from previous installs to fix DEV-5663
Call RemoveOldShaders

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Need to clean out old XUI files that predate skinning
Call RemoveOldXUI

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Clear out old releasenotes.txt files. These are now on the public wiki.
Call RemoveOldReleaseNotes

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Remove an old xui file that should not be in Silver skin
Call RemoveOldAboutLandSilver

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Files
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; This placeholder is replaced by the complete list of all the files in the installer, by viewer_manifest.py
SetOutPath $INSTDIR\skins\default\html\zh\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\zh\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\zh\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\pt\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\pt\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\pt\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\ko\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\ko\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\ko\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\ja\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\ja\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\ja\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\fr\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\fr\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\fr\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\es\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\es\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\es\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\en-us\loading
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\en-us\loading\sl_logo_rotate_black.gif
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\en-us\loading\sl_logo_rotate_black.gif
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\en-us\loading\loading.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\en-us\loading\loading.html
SetOutPath $INSTDIR\skins\default\html\en-us\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\en-us\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\en-us\loading-error\index.html
SetOutPath $INSTDIR\skins\default\html\de\loading-error
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\de\loading-error\index.html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\de\loading-error\index.html
SetOutPath $INSTDIR\skins\silver\xui\en-us
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\xui\en-us\floater_chatterbox.xml
SetOutPath $INSTDIR\skins\default\xui\zh
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_new_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_new_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\zh\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\uk
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_progress.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_progress.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\uk\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\tr
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\tr\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\ru
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_progress.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_progress.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ru\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\pt
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_account_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_account_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pt\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\pl
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\pl\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\ko
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_LCD.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_preferences_LCD.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\LCD_text.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ko\LCD_text.xml
SetOutPath $INSTDIR\skins\default\xui\ja
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_LCD.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_preferences_LCD.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\LCD_text.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\ja\LCD_text.xml
SetOutPath $INSTDIR\skins\default\xui\it
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\it\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\hu
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\hu\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\fr
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\fr\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\es
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\es\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\en-us
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_progress.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_progress.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_ascent_vanity.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_ascent_vanity.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_ascent_system.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_ascent_system.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_bg_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_bg_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_bg_tab.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_bg_tab.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_bars.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_bars.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types_windows.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types_windows.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types_mac.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types_mac.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types_linux.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types_linux.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\fonts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_vfs_explorer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_vfs_explorer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_vfs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_vfs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_teleport_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_teleport_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_teleport.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_teleport.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_radar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_radar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_new_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_new_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_new_local_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_new_local_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_message_log.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_message_log.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_message_builder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_message_builder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_backup_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_backup_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_backup.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory_backup.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_import_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_import_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_html_simple.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_html_simple.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_hex.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_hex.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_export.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_export.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_explore_sounds.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_explore_sounds.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_explore_animations.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_explore_animations.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_dickdongs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_dickdongs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_clothing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_clothing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_blacklist.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_blacklist.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_asset_text_editor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_asset_text_editor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_area_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_area_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_ao.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_ao.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\alerts.xml
SetOutPath $INSTDIR\skins\default\xui\de
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_LCD.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_preferences_LCD.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_notifications_channel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_media_remote.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_voice_wizard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_skin_preview_template.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_post_process.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_notifications_console.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_notification.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_god_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\LCD_text.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\de\LCD_text.xml
SetOutPath $INSTDIR\skins\default\xui\da
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\xui_version.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\ui_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\teleport_strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\strings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\role_actions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_enable.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_voice_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_top_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_toolbar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_status_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_speaker_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_scrolling_param.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_terrain.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_estate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_region_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_web.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_voice.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_skins.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_popups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_network.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_input.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_graphics1.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_chat.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_preferences_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_place_small.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_place.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_overlaybar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_media_remote_expanded.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_media_controls.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_master_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_land_covenant.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_groups.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_voting.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_roles.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_notices.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_land_money.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_invite.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_general.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_chat_bar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_avatar_pick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_avatar_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_audio_device.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\panel_audio.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\notify.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\notifications.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_viewer.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_slurl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_self.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_avatar.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_pie_attachment.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_login.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\menu_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_world_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_windlight_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_wearable_save_as.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_water.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_url_entry.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_tos.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_top_objects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_tools.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_texture_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_telehub.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_statistics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_sound_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_snapshot.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_sim_release_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_settings_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_sell_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_select_key.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_search.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_queue.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_ed_panel.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_script_debug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_report_bug.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_report_abuse.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_region_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_profile.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_price_for_listing.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_url.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_texture_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_sound.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_notecard_keep_discard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_notecard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_existing_landmark.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_event.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_embedded_texture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_classified.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preview_animation.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_preferences.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_postcard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_perm_prefs.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_pay_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_pay.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_openobject.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_object_im_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_new_outfit_dialog.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_new_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_name_description.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_my_friends.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mute_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mute.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_moveview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mini_map.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_mem_leaking.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_media_browser.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_lsl_guide.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_live_lsleditor.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_landmark_ctrl.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_land_holdings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_lagmeter.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_joystick.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inventory_view_finder.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inventory_item_properties.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inventory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_instant_message_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_instant_message_ad_hoc.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_instant_message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_inspect.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_import.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_image_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_im.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_hud.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_html.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_hardware_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_group_info.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_gesture.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_font_test.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_env_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_directory.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_device_settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_day_cycle_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_customize.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_critical.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_color_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_choose_group.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_chatterbox.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_chat_history.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_camera.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_object.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_currency.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_buy_contents.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_bumps.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_bulk_perms.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_build_options.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_beacons.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_avatar_textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_avatar_picker.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_audio_volume.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_auction.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_animation_preview.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_active_speakers.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_about_land.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\floater_about.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\alerts.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\da\alerts.xml
SetOutPath $INSTDIR\app_settings\shaders\class3\lighting
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class3\lighting\sumLightsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class3\lighting\sumLightsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class3\lighting\sumLightsSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class3\lighting\sumLightsSpecularV.glsl
SetOutPath $INSTDIR\app_settings\shaders\class3\avatar
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class3\avatar\avatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class3\avatar\avatarV.glsl
SetOutPath $INSTDIR\app_settings\shaders\class2\windlight
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\transportF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\transportF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\skyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\skyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\skyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\skyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\gammaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\gammaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\cloudsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\cloudsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\cloudsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\cloudsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsVarsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsVarsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsVarsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsVarsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsHelpersV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsHelpersV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\windlight\atmosphericsF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class2\objects
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\objects\shinyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\objects\shinyV.glsl
SetOutPath $INSTDIR\app_settings\shaders\class2\lighting
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\sumLightsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\sumLightsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\sumLightsSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\sumLightsSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightShinyWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightShinyWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightFullbrightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightFullbrightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightFullbrightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightFullbrightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightFullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightFullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\lighting\lightF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class2\environment
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\waterFogF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\waterFogF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\waterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\waterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\underWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\underWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\terrainWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\terrainWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\terrainV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\terrainV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\terrainF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\environment\terrainF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class2\effects
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\simpleF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\simpleF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\nightVisionF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\nightVisionF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\extractF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\extractF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\drawQuadV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\drawQuadV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\colorFilterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\colorFilterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\blurV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\blurV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\blurF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\effects\blurF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class2\avatar
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\avatar\eyeballV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class2\avatar\eyeballV.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\windlight
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\transportF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\transportF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\gammaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\gammaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsVarsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsVarsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsVarsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsVarsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsHelpersV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsHelpersV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\windlight\atmosphericsF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\objects
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\simpleWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\simpleWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\simpleV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\simpleV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\simpleF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\simpleF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\shinyWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\shinyWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\shinyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\shinyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\shinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\shinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightShinyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightShinyV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\objects\fullbrightF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\lighting
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\sumLightsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\sumLightsV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\sumLightsSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\sumLightsSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightShinyWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightShinyWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFuncV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFuncV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFuncSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFuncSpecularV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFullbrightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFullbrightWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFullbrightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFullbrightShinyF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightFullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\lighting\lightF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\interface
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\interface\highlightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\interface\highlightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\interface\highlightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\interface\highlightF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\environment
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\waterV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\waterV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\waterFogF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\waterFogF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\waterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\waterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\underWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\underWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\terrainWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\terrainWaterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\terrainV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\terrainV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\terrainF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\environment\terrainF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\effects
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowExtractV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowExtractV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowExtractF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\effects\glowExtractF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\deferred
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\waterV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\waterV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\waterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\waterF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\treeV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\treeV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\treeF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\treeF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\terrainV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\terrainV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\terrainF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\terrainF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\sunLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\sunLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\sunLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\sunLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\softenLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\softenLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\softenLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\softenLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\shadowV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\shadowV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\shadowF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\shadowF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\pointLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\pointLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\pointLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\pointLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\multiPointLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\multiPointLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\impostorV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\impostorV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\impostorF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\impostorF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\fullbrightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\fullbrightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\fullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\fullbrightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\diffuseV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\diffuseV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\diffuseF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\diffuseF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\bumpV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\bumpV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\bumpF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\bumpF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\blurLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\blurLightV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\blurLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\blurLightF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarShadowV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarShadowV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarShadowF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarShadowF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarAlphaV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarAlphaV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarAlphaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\avatarAlphaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\alphaV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\alphaV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\alphaF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\deferred\alphaF.glsl
SetOutPath $INSTDIR\app_settings\shaders\class1\avatar
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\pickAvatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\pickAvatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\pickAvatarF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\pickAvatarF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\eyeballV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\eyeballV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\eyeballF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\eyeballF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\avatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\avatarV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\avatarSkinV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\avatarSkinV.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\avatarF.glsl
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\class1\avatar\avatarF.glsl
SetOutPath $INSTDIR\skins\silver\textures
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\white.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\white.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\up_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\up_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_btn_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_btn_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_btn_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_btn_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_btn_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_btn_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_bg.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\toolbar_bg.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_zoom_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_zoom_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_zoom.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_zoom.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_dozer_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_dozer_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_dozer.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tool_dozer.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tearoffbox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tearoffbox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tearoff_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tearoff_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_top_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_top_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_top_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_top_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_bottom_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_bottom_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_bottom_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\tab_bottom_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_scripts.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_scripts.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_push.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_push.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_no_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_health.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_health.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_land_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_land_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_land.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_land.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_currency_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_currency_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_currency.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\status_buy_currency.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\startup_logo.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\startup_logo.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\square_btn_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\square_btn_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\square_btn_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\square_btn_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\spin_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\sm_rounded_corners_simple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\sm_rounded_corners_simple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_right_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_right_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_right_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_right_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_left_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_left_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_left_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_left_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\scrollbutton_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\rounded_square_soft.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\rounded_square_soft.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\rounded_square.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\rounded_square.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\restore_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\restore_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\restore_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\restore_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\restore.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\restore.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\resize_handle_bottom_right_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\resize_handle_bottom_right_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_inactive_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_inactive_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_inactive_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_inactive_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_active_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_active_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_active_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\radio_active_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ptt_lock_on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ptt_lock_on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ptt_lock_off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ptt_lock_off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\progressbar_track.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\progressbar_track.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\progressbar_fill.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\progressbar_fill.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tree_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tree_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tree.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tree.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_torus_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_torus_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_torus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_torus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tetrahedron_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tetrahedron_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tetrahedron.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_tetrahedron.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_ring_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_ring_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_ring.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_ring.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_pyramid_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_pyramid_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_pyramid.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_pyramid.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_prism_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_prism_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_prism.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_prism.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_hemi_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_grass_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_grass_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_grass.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_grass.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\object_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\notify_next.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\notify_next.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\minimize_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\minimize_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\minimize_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\minimize_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\minimize.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\minimize.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_telehub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_telehub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_infohub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_infohub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_home.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_home.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_event_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_event_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_avatar_you_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_avatar_you_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_avatar_16.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\map_avatar_16.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\lightgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\lightgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_undershirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_undershirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_underpants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_underpants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_socks.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_socks.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_skirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_skirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_skin.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_skin.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_shoes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_shoes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_shirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_shirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_shape.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_shape.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_pants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_pants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_object_multi.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_object_multi.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_landmark_visited.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_landmark_visited.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_jacket.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_jacket.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_hair.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_hair.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_gloves.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_gloves.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_eyes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_eyes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_callingcard_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_callingcard_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_callingcard_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_callingcard_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_attach.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_attach.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_item_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_trash.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_trash.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_plain_open.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_plain_open.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_plain_closed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_plain_closed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_lostandfound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_lostandfound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_callingcard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_callingcard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_bodypart.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_bodypart.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\inv_folder_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_top_pick.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_top_pick.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_lock.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_lock.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_groupnoticeinventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_groupnoticeinventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_groupnotice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_groupnotice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_group.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_group.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_for_sale.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_for_sale.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_day_cycle.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_day_cycle.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_avatar_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_avatar_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_avatar_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_avatar_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_auction.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icon_auction.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on-lvl3.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on-lvl3.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on-lvl2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on-lvl2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on-lvl1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-on-lvl1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice_ptt-off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-pvtfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-pvtfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-localchat.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-localchat.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-groupfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-groupfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-call-start.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-call-start.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-call-end.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_voice-call-end.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_minimap.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_minimap.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_map.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_map.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_inventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_inventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_toolbar_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_speaker_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_speaker_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_speaker-muted_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_speaker-muted_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_slide-thumb_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_slide-thumb_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_slide-highlight.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_slide-highlight.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_slide-groove_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_slide-groove_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_rounded-text-field.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_rounded-text-field.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_music-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_music-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_music-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_music-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media_movie.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media_movie.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_media-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_label_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_label_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_label_music.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_label_music.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_label_media.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_label_media.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\icn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_right_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_right_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_right_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_right_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_right.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_right.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_left_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_left_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\flyout_btn_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_visible_online_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_visible_online_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_visible_map_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_visible_map_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_online_status_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_online_status_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_edit_theirs_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_edit_theirs_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_edit_mine_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff_edit_mine_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\eye_button_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\eye_button_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\eye_button_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\eye_button_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\down_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\down_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\darkgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\darkgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\combobox_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\combobox_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\closebox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\closebox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\close_inactive_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\close_inactive_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\close_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\close_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_enabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_enabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_enabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_enabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_disabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_disabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_disabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\checkbox_disabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_enabled_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_enabled_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_enabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_enabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_disabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_disabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_stop_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_stop_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_play_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_play_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_pause_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_pause_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\button_anim_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\btn_chatbar_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\btn_chatbar_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\btn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\btn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\black.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\black.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\arrow_up.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\arrow_up.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\arrow_down.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\arrow_down.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\active_voice_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\active_voice_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\active_speakers.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\active_speakers.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
SetOutPath $INSTDIR\skins\default\textures
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\white.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\white.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\uv_test2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\uv_test2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\uv_test1.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\uv_test1.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\up_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\up_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\up_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\up_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_btn_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_btn_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_btn_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_btn_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_btn_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_btn_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_bg.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\toolbar_bg.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_zoom_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_zoom_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_zoom.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_zoom.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_dozer_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_dozer_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_dozer.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tool_dozer.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\textures.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tearoffbox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tearoffbox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tearoff_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tearoff_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tabarea.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tabarea.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_top_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_top_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_top_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_top_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_bottom_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_bottom_selected_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_bottom_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_bottom_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_background_purple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_background_purple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_background_lightgrey.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_background_lightgrey.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_background_darkpurple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\tab_background_darkpurple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_search_btn_pressed.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_search_btn_pressed.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_search_btn.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_search_btn.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_script_debug.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_script_debug.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_voice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_scripts.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_scripts.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_push.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_push.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_no_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_money.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_money.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_health.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_health.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_land_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_land_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_land.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_land.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_currency_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_currency_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_currency.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_buy_currency.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_busy.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\status_busy.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\startup_logo.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\startup_logo.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\square_btn_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\square_btn_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\square_btn_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\square_btn_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spin_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spacer35.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spacer35.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spacer24.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\spacer24.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\smicon_warn.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\smicon_warn.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\sm_rounded_corners_simple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\sm_rounded_corners_simple.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\slim_icon_16_viewer.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\slim_icon_16_viewer.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\skin_thumbnail_silver.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\skin_thumbnail_silver.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\skin_thumbnail_default.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\skin_thumbnail_default.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\silhouette.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\silhouette.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_up_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_up_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_right_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_right_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_right_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_right_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_left_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_left_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_left_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_left_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_down_out_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\scrollbutton_down_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\script_error.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\script_error.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\rounded_square_soft.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\rounded_square_soft.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\rounded_square.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\rounded_square.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\restore_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\restore_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\restore_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\restore_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\restore.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\restore.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\resize_handle_bottom_right_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\resize_handle_bottom_right_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_inactive_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_inactive_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_inactive_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_inactive_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_active_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_active_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_active_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\radio_active_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ptt_lock_on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ptt_lock_on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ptt_lock_off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ptt_lock_off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\propertyline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\propertyline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\progressbar_track.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\progressbar_track.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\progressbar_fill.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\progressbar_fill.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\progress_fill.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\progress_fill.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\pos_lock_on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\pos_lock_on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\pos_lock_off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\pos_lock_off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\pixiesmall.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\pixiesmall.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tree_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tree_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tree.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tree.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_torus_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_torus_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_torus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_torus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tetrahedron_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tetrahedron_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tetrahedron.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_tetrahedron.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_ring_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_ring_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_ring.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_ring.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_pyramid_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_pyramid_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_pyramid.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_pyramid.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_prism_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_prism_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_prism.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_prism.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_sphere_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_sphere.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_hemi_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_grass_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_grass_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_grass.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_grass.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cylinder_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cylinder.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cube_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cube.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cone_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\object_cone.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_tip_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_tip_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_next.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_next.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_caution_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_caution_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_box_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\notify_box_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\noentrypasslines.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\noentrypasslines.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\noentrylines.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\noentrylines.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\mute_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\mute_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\music_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\music_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_up_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_up_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_up_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_up_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_right_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_right_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_right_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_right_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_left_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_left_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_left_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_turn_left_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_right_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_right_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_right_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_right_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_left_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_left_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_left_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_left_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_forward_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_forward_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_forward_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_forward_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_down_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_down_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_down_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_down_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_backward_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_backward_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_backward_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\move_backward_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\missing_asset.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\missing_asset.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\minimize_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\minimize_pressed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\minimize_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\minimize_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\minimize.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\minimize.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_scrollbg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_scrollbg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_hoverrectangle.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_hoverrectangle.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_divider.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_divider.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_bg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_panel_bg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_icon.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_stoploading.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_stoploading.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrollup.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrollup.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrollright.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrollright.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrollleft.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrollleft.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrolldown.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_scrolldown.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_reload.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_reload.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_optimalzoom.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_optimalzoom.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_newwindow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_newwindow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_home.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_home.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_forward.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_forward.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_done.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_done.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_back.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\media_btn_back.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_track_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_track_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_track_16.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_track_16.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_telehub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_telehub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_infohub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_infohub.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_home.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_home.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_event_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_event_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_you_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_you_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_you_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_you_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_below_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_below_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_below_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_below_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_above_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_above_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_above_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_above_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_8.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_32.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_16.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\map_avatar_16.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\locked_image.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\locked_image.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lightgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lightgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\legend.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\legend.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lag_status_warning.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lag_status_warning.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lag_status_good.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lag_status_good.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lag_status_critical.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\lag_status_critical.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_undershirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_undershirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_underpants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_underpants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_tattoo.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_tattoo.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_socks.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_socks.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_skirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_skirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_skin.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_skin.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_shoes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_shoes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_shirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_shirt.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_shape.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_shape.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_script_dangerous.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_script_dangerous.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_pants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_pants.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_object_multi.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_object_multi.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_landmark_visited.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_landmark_visited.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_jacket.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_jacket.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_hair.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_hair.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_gloves.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_gloves.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_eyes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_eyes.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_callingcard_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_callingcard_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_callingcard_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_callingcard_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_attach.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_attach.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_item_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_trash.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_trash.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_texture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_sound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_script.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_plain_open.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_plain_open.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_plain_closed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_plain_closed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_object.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_notecard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_lostandfound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_lostandfound.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_landmark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_gesture.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_clothing.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_callingcard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_callingcard.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_bodypart.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_bodypart.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\inv_folder_animation.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_top_pick.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_top_pick.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_popular.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_popular.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_place.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_place.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_eventpoll_gray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_eventpoll_gray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_eventpoll.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_eventpoll.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_circuit_gray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_circuit_gray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_circuit.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_net_close_circuit.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_lock.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_lock.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_groupnoticeinventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_groupnoticeinventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_groupnotice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_groupnotice.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_group.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_group.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_for_sale_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_for_sale_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_for_sale.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_for_sale.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_event_mature.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_event_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_event_adult.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_event.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_diurnal.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_diurnal.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_day_cycle.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_day_cycle.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_avatar_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_avatar_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_avatar_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_avatar_offline.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_auction.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icon_auction.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on-lvl3.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on-lvl3.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on-lvl2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on-lvl2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on-lvl1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-on-lvl1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice_ptt-off.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-pvtfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-pvtfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-localchat.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-localchat.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-groupfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-groupfocus.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-call-start.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-call-start.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-call-end.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_voice-call-end.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_snapshot.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_search.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_radar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_radar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_minimap.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_minimap.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_map.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_map.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_inventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_inventory.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_fly.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_toolbar_build.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_textfield_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_textfield_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_speaker_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_speaker_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_speaker-muted_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_speaker-muted_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_slide-thumb_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_slide-thumb_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_slide-highlight.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_slide-highlight.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_slide-groove_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_slide-groove_dark.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_scrollbar_thumb.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_scrollbar_thumb.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_scrollbar_bg.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_scrollbar_bg.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_scrollbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_scrollbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_rounded-text-field.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_rounded-text-field.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_music.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_music.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_music-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_music-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_music-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_music-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media_movie.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media_movie.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-stop_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-stop_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-stop_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-stop_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-stop_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-stop_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause_enabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_media-pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_label_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_label_web.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_label_music.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_label_music.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_label_media.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_label_media.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_clear_lineeditor.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_clear_lineeditor.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-typing3.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-typing3.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-typing2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-typing2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-typing1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-typing1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-dot-lvl2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-dot-lvl2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-dot-lvl1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-dot-lvl1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-dot-lvl0.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\icn_active-speakers-dot-lvl0.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\foot_shadow.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\foot_shadow.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\folder_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\folder_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_right_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_right_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_right_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_right_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_right.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_right.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_left_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_left_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_left_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\flyout_btn_left.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_online_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_online_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_online.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_map_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_map_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_map.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_visible_map.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_online_status_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_online_status_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_theirs_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_theirs_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_theirs.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_theirs.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_mine_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_mine_button.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_mine.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff_edit_mine.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ff9a71eb-7414-4cf8-866e-a701deb7c3cf.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\fc987bf9-b8cb-f8e5-45f2-d664ca6bd3eb.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\fc987bf9-b8cb-f8e5-45f2-d664ca6bd3eb.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\fb2ae204-3fd1-df33-594f-c9f882830e66.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\fb2ae204-3fd1-df33-594f-c9f882830e66.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\fb1fecba-9585-415b-ad15-6e6e3d6c5479.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\fb1fecba-9585-415b-ad15-6e6e3d6c5479.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f9bbb2fe-584b-4c01-86fc-599c69534c1b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f9bbb2fe-584b-4c01-86fc-599c69534c1b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f54a0c32-3cd1-d49a-5b4f-7b792bebc204.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f54a0c32-3cd1-d49a-5b4f-7b792bebc204.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f4b6b161-6530-6679-1a84-adfcb71a8b12.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f4b6b161-6530-6679-1a84-adfcb71a8b12.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f2d7b6f6-4200-1e9a-fd5b-96459e950f94.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\f2d7b6f6-4200-1e9a-fd5b-96459e950f94.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\eye_button_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\eye_button_inactive.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\eye_button_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\eye_button_active.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ebf2aa19-6c34-c5d8-4f14-853da1241f91.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ebf2aa19-6c34-c5d8-4f14-853da1241f91.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e97cf410-8e61-7005-ec06-629eba4cd1fb.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e97cf410-8e61-7005-ec06-629eba4cd1fb.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e674ca0c-a387-4dae-a0b4-db6bd073faa5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e674ca0c-a387-4dae-a0b4-db6bd073faa5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e569711a-27c2-aad4-9246-0c910239a179.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e569711a-27c2-aad4-9246-0c910239a179.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e38248f9-f2ee-2c9f-aa49-4860857e3b08.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e38248f9-f2ee-2c9f-aa49-4860857e3b08.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e3369e02-93e1-43dc-b9c0-4533db0963d0.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e3369e02-93e1-43dc-b9c0-4533db0963d0.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e121e2fc-7573-740f-edfd-0d45a9ba486e.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\e121e2fc-7573-740f-edfd-0d45a9ba486e.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\down_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\down_arrow.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\direction_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\direction_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\default_profile_picture.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\default_profile_picture.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\default_land_picture.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\default_land_picture.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\de651394-f926-48db-b666-e49d83af1bbc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\de651394-f926-48db-b666-e49d83af1bbc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\db9d39ec-a896-c287-1ced-64566217021e.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\db9d39ec-a896-c287-1ced-64566217021e.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\darkgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\darkgray.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d9258671-868f-7511-c321-7baef9e948a4.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d9258671-868f-7511-c321-7baef9e948a4.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d7d99e40-10e2-5739-d063-91dcbdefc492.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d7d99e40-10e2-5739-d063-91dcbdefc492.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d691a01c-13b7-578d-57c0-5caef0b4e7e1.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d691a01c-13b7-578d-57c0-5caef0b4e7e1.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d319ce44-0821-932a-cd18-cd1afb9d3ead.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d319ce44-0821-932a-cd18-cd1afb9d3ead.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d21e44ca-ff1c-a96e-b2ef-c0753426b7d9.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d21e44ca-ff1c-a96e-b2ef-c0753426b7d9.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d2114404-dd59-4a4d-8e6c-49359e91bbf0.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d2114404-dd59-4a4d-8e6c-49359e91bbf0.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d07f6eed-b96a-47cd-b51d-400ad4a1c428.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\d07f6eed-b96a-47cd-b51d-400ad4a1c428.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\crosshairs.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\crosshairs.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\combobox_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\combobox_arrow.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\color_swatch_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\color_swatch_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cloud-particle.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cloud-particle.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\closebox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\closebox.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\close_inactive_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\close_inactive_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\close_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\close_in_blue.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\circle.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\circle.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkerboard_transparency_bg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkerboard_transparency_bg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_enabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_enabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_enabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_enabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_disabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_disabled_true.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_disabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\checkbox_disabled_false.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ce15fd63-b0b6-463c-a37d-ea6393208b3e.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cdd9a9fc-6d0b-f90d-8416-c72b6019bca8.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cdd9a9fc-6d0b-f90d-8416-c72b6019bca8.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cce0f112-878f-4586-a2e2-a8f104bba271.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cce0f112-878f-4586-a2e2-a8f104bba271.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_zoom_plus_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_zoom_plus_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_zoom_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_zoom_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_zoom_minus_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_zoom_minus_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_tracking_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_tracking_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_tracking_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_tracking_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_rotate_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_rotate_out.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_rotate_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\cam_rotate_in.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ca7609c6-6ec6-32d9-332e-0d8f437ef644.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ca7609c6-6ec6-32d9-332e-0d8f437ef644.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ca4e8c27-473c-eb1c-2f5d-50ee3f07d85c.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ca4e8c27-473c-eb1c-2f5d-50ee3f07d85c.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\c7d8bbf3-21ee-4f6e-9b20-3cf18425af1d.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\c7d8bbf3-21ee-4f6e-9b20-3cf18425af1d.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\c63f124c-6340-4fbf-b59e-0869a44adb64.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\c1e21504-f136-451d-b8e9-929037812f1d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_enabled_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_enabled_selected_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_enabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_enabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_disabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_disabled_32x128.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_stop_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_stop_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_stop_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_stop_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_stop.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_play_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_play_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_play_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_play_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_play.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_pause_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_pause_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_pause_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_pause_disabled.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\button_anim_pause.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\btn_chatbar_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\btn_chatbar_selected.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\btn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\btn_chatbar.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\black.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\black.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\beb169c7-11ea-fff2-efe5-0f24dc881df2.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\beb169c7-11ea-fff2-efe5-0f24dc881df2.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_warn.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_warn.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_ok.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_ok.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_note.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_note.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_error.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\badge_error.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b9f1a3b8-933e-b7c8-e6f5-dba1bc666bed.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b9f1a3b8-933e-b7c8-e6f5-dba1bc666bed.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b9e1cf8a-9660-c020-0c69-18f1ea27268a.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b9e1cf8a-9660-c020-0c69-18f1ea27268a.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b8eed5f0-64b7-6e12-b67f-43fa8e773440.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b8eed5f0-64b7-6e12-b67f-43fa8e773440.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b8d3965a-ad78-bf43-699b-bff8eca6c975.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b8d3965a-ad78-bf43-699b-bff8eca6c975.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b4ba225c-373f-446d-9f7e-6cb7b5cf9b3d.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b4ba225c-373f-446d-9f7e-6cb7b5cf9b3d.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b4870163-6208-42a9-9801-93133bf9a6cd.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b2ef2d31-9714-a07b-6ca7-31638166364b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\b2ef2d31-9714-a07b-6ca7-31638166364b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\avatar_thumb_bkgrnd.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\avatar_thumb_bkgrnd.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\arrow_up.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\arrow_up.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\arrow_down.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\arrow_down.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\alpha_gradient_2d.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\alpha_gradient_2d.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\alpha_gradient.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\alpha_gradient.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ae874d1a-93ef-54fb-5fd3-eb0cb156afc0.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\ae874d1a-93ef-54fb-5fd3-eb0cb156afc0.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\active_voice_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\active_voice_tab.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\active_speakers.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\active_speakers.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\account_id_orange.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\account_id_orange.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\account_id_green.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\account_id_green.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\abb783e6-3e93-26c0-248a-247666855da3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\abb783e6-3e93-26c0-248a-247666855da3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\a85ac674-cb75-4af6-9499-df7c5aaf7a28.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\a85ac674-cb75-4af6-9499-df7c5aaf7a28.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\a6162133-724b-54df-a12f-51cd070ad6f3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\a6162133-724b-54df-a12f-51cd070ad6f3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9deab416-9c63-78d6-d558-9a156f12044c.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9deab416-9c63-78d6-d558-9a156f12044c.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9cad3e6d-2d6d-107d-f8ab-5ba272b5bfe1.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9c88539c-fd04-46b8-bea2-ddf1bcffe3bd.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9c88539c-fd04-46b8-bea2-ddf1bcffe3bd.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9beb8cdd-3dce-53c2-b28e-e1f3bc2ec0a4.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\9beb8cdd-3dce-53c2-b28e-e1f3bc2ec0a4.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\99bd60a2-3250-efc9-2e39-2fbcadefbecc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\99bd60a2-3250-efc9-2e39-2fbcadefbecc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\988dd995-1769-bdc9-8842-51f8f2b03884.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\988dd995-1769-bdc9-8842-51f8f2b03884.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\978380f0-aaf7-c459-14e3-9808833fd372.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\978380f0-aaf7-c459-14e3-9808833fd372.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\96b4de31-f4fa-337d-ec78-451e3609769e.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\96b4de31-f4fa-337d-ec78-451e3609769e.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\95281d5c-d27a-ee13-e067-08295b67b58a.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\95281d5c-d27a-ee13-e067-08295b67b58a.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\92e66e00-f56f-598a-7997-048aa64cde18.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\92e66e00-f56f-598a-7997-048aa64cde18.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8f761ce3-5939-4d3a-8991-00064fdfacf9.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8f761ce3-5939-4d3a-8991-00064fdfacf9.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8f458549-173b-23ff-d4ff-bfaa5ea2371b.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8f458549-173b-23ff-d4ff-bfaa5ea2371b.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8dcd4a48-2d37-4909-9f78-f7a9eb4ef903.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8dcd4a48-2d37-4909-9f78-f7a9eb4ef903.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8a515889-eac9-fb55-8eba-d2dc09eb32c8.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8a515889-eac9-fb55-8eba-d2dc09eb32c8.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\89e9fc7c-0b16-457d-be4f-136270759c4d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\89e9fc7c-0b16-457d-be4f-136270759c4d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8872f2b8-31db-42d8-580a-b3e4a91262de.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\8872f2b8-31db-42d8-580a-b3e4a91262de.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\83b77fc6-10b4-63ec-4de7-f40629f238c5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\83b77fc6-10b4-63ec-4de7-f40629f238c5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\827ff765-8c1d-a8b1-23f7-fdcba560effc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\827ff765-8c1d-a8b1-23f7-fdcba560effc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\822ded49-9a6c-f61c-cb89-6df54f42cdf4.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\822ded49-9a6c-f61c-cb89-6df54f42cdf4.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7dabc040-ec13-2309-ddf7-4f161f6de2f4.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7cb070bc-fc00-4527-9c4d-7f7e0c4191be.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7cb070bc-fc00-4527-9c4d-7f7e0c4191be.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7ca39b4c-bd19-4699-aff7-f93fd03d3e7b.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7ca39b4c-bd19-4699-aff7-f93fd03d3e7b.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7c0cf89b-44b1-1ce2-dd74-07102a98ac2a.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7c0cf89b-44b1-1ce2-dd74-07102a98ac2a.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7a2b3a4a-53c2-53ac-5716-aac7d743c020.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7a2b3a4a-53c2-53ac-5716-aac7d743c020.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7a0b1bdb-b5d9-4df5-bac2-ba230da93b5b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\79504bf5-c3ec-0763-6563-d843de66d0a1.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\79504bf5-c3ec-0763-6563-d843de66d0a1.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\78af921a-3c49-47a1-9c4e-2608951164ae.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\78af921a-3c49-47a1-9c4e-2608951164ae.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7581f2f4-d0d2-481a-bc75-69a13d9caeaa.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\7581f2f4-d0d2-481a-bc75-69a13d9caeaa.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\74e1a96f-4833-a24d-a1bb-1bce1468b0e7.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\74e1a96f-4833-a24d-a1bb-1bce1468b0e7.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\74ba3584-58ea-9984-5b76-62d37942ab77.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\74ba3584-58ea-9984-5b76-62d37942ab77.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\73577b7b-19c3-4050-a19d-36bc2408aa79.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\73577b7b-19c3-4050-a19d-36bc2408aa79.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\735198cf-6ea0-2550-e222-21d3c6a341ae.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\735198cf-6ea0-2550-e222-21d3c6a341ae.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6de37e4e-7029-61f5-54b8-f5e63f983f58.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6de37e4e-7029-61f5-54b8-f5e63f983f58.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6c9fa78a-1c69-2168-325b-3e03ffa348ce.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6c9fa78a-1c69-2168-325b-3e03ffa348ce.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6c4727b8-ac79-ba44-3b81-f9aa887b47eb.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6c4727b8-ac79-ba44-3b81-f9aa887b47eb.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\67931331-0c02-4876-1255-28770896c6a2.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\67931331-0c02-4876-1255-28770896c6a2.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6522e74d-1660-4e7f-b601-6f48c1659a77.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6522e74d-1660-4e7f-b601-6f48c1659a77.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\64eed6af-f575-35c7-baa4-b140bdcdb00f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\64eed6af-f575-35c7-baa4-b140bdcdb00f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\64367bd1-697e-b3e6-0b65-3f862a577366.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\64367bd1-697e-b3e6-0b65-3f862a577366.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\63338ede-0037-c4fd-855b-015d77112fc8.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\63338ede-0037-c4fd-855b-015d77112fc8.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6002a571-549c-472c-9443-9ab35b1a55ed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\6002a571-549c-472c-9443-9ab35b1a55ed.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5bc11cd6-2f40-071e-a8da-0903394204f9.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5bc11cd6-2f40-071e-a8da-0903394204f9.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5abfabc2-5d6d-4912-acd8-d7e38ae93d02.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5abfabc2-5d6d-4912-acd8-d7e38ae93d02.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5ab48dd5-05d0-4f1a-ace6-efd4e2fb3508.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5ab48dd5-05d0-4f1a-ace6-efd4e2fb3508.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5894e2e7-ab8d-edfa-e61c-18cf16854ba3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5894e2e7-ab8d-edfa-e61c-18cf16854ba3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5748decc-f629-461c-9a36-a35a221fe21f.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\5748decc-f629-461c-9a36-a35a221fe21f.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\53a2f406-4895-1d13-d541-d2e3b86bc19c.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\53a2f406-4895-1d13-d541-d2e3b86bc19c.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\48766d75-6e58-de84-68fe-1980c64feaee.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\48766d75-6e58-de84-68fe-1980c64feaee.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\47a8c844-cd2a-4b1a-be01-df8b1612fe5d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\47a8c844-cd2a-4b1a-be01-df8b1612fe5d.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\4726f13e-bd07-f2fb-feb0-bfa2ac58ab61.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\4726f13e-bd07-f2fb-feb0-bfa2ac58ab61.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\43f0a590-f3d3-48b5-b460-f5b3e6e03626.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\43f0a590-f3d3-48b5-b460-f5b3e6e03626.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\402f8b24-5f9d-4905-b5f8-37baff603e88.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\402f8b24-5f9d-4905-b5f8-37baff603e88.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\3cddf591-a726-4702-87b3-70c1daf88f90.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\3cddf591-a726-4702-87b3-70c1daf88f90.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\3c59f7fe-9dc8-47f9-8aaf-a9dd1fbc3bef.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\3c59f7fe-9dc8-47f9-8aaf-a9dd1fbc3bef.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\3c18c87e-5f50-14e2-e744-f44734aa365f.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\39801651-26cb-4926-af57-7af9352c273c.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\39801651-26cb-4926-af57-7af9352c273c.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\37c8e000-6aa2-41ef-8f86-e0c2e60bfa42.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\37c8e000-6aa2-41ef-8f86-e0c2e60bfa42.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\35f217a3-f618-49cf-bbca-c86d486551a9.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\35f217a3-f618-49cf-bbca-c86d486551a9.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\34c9398d-bb78-4643-9633-46a2fa3e9637.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\34c9398d-bb78-4643-9633-46a2fa3e9637.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\335f8f14-f2db-db7c-1c04-734dc7657439.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\335f8f14-f2db-db7c-1c04-734dc7657439.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\303cd381-8560-7579-23f1-f0a880799740.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\303cd381-8560-7579-23f1-f0a880799740.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\30047cec-269d-408e-0c30-b2603b887268.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\30047cec-269d-408e-0c30-b2603b887268.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2d784476-d0db-9979-0cff-9408745a7cf3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2d784476-d0db-9979-0cff-9408745a7cf3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2caf1179-7861-6ff3-4b7d-46e17780bdfa.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2caf1179-7861-6ff3-4b7d-46e17780bdfa.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2a4880b6-b7a3-690a-2049-bfbe38eafb9f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2a4880b6-b7a3-690a-2049-bfbe38eafb9f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\29de489d-0491-fb00-7dab-f9e686d31e83.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\29de489d-0491-fb00-7dab-f9e686d31e83.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\28f0f9ca-0423-4d1b-9e76-616ffce99544.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\28f0f9ca-0423-4d1b-9e76-616ffce99544.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2660b114-1d66-3cde-e148-ebc2d1f963d5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\2660b114-1d66-3cde-e148-ebc2d1f963d5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\1e63e323-5fe0-452e-92f8-b98bd0f764e3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\1e63e323-5fe0-452e-92f8-b98bd0f764e3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\19c76b49-c5f4-aeca-7cd8-17010f2969c3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\19c76b49-c5f4-aeca-7cd8-17010f2969c3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\18fb888b-e8f1-dce7-7da7-321d651ea6b0.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\18fb888b-e8f1-dce7-7da7-321d651ea6b0.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\179cdabd-398a-9b6b-1391-4dc333ba321f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\179cdabd-398a-9b6b-1391-4dc333ba321f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\13dd1d96-6836-461e-8a4c-36003065c59b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\13dd1d96-6836-461e-8a4c-36003065c59b.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\11ee27f5-43c0-414e-afd5-d7f5688c351f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\11ee27f5-43c0-414e-afd5-d7f5688c351f.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\111b39de-8928-4690-b7b2-e17d5c960277.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\111b39de-8928-4690-b7b2-e17d5c960277.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\10d2a01a-0818-84b9-4b96-c2eb63256519.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\10d2a01a-0818-84b9-4b96-c2eb63256519.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0ff70ead-4562-45f9-9e8a-52b1a3286868.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0ff70ead-4562-45f9-9e8a-52b1a3286868.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0e82d24e-ed45-41bc-b090-94c97c1caab2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0e82d24e-ed45-41bc-b090-94c97c1caab2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0bc58228-74a0-7e83-89bc-5c23464bcec5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0bc58228-74a0-7e83-89bc-5c23464bcec5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0b444c3a-75c2-4891-9d1e-ac35c8d13d62.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0b444c3a-75c2-4891-9d1e-ac35c8d13d62.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0a94b42f-ec84-5f9c-14b7-1ef8505ceead.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0a94b42f-ec84-5f9c-14b7-1ef8505ceead.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\09a324a8-acc1-d9cd-2cbd-7465d90d3a98.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\09a324a8-acc1-d9cd-2cbd-7465d90d3a98.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\07d0ea4c-af0c-aad1-dbbf-c24020ff2b80.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\07d0ea4c-af0c-aad1-dbbf-c24020ff2b80.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\073c9723-540c-5449-cdd4-0e87fdc159e3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\073c9723-540c-5449-cdd4-0e87fdc159e3.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\058c75c0-a0d5-f2f8-43f3-e9699a89c2fc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\058c75c0-a0d5-f2f8-43f3-e9699a89c2fc.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0498c309-5306-43cd-82a2-ae31d096cdef.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0498c309-5306-43cd-82a2-ae31d096cdef.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\041ee5a0-cb6a-9ac5-6e49-41e9320507d5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\041ee5a0-cb6a-9ac5-6e49-41e9320507d5.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0187babf-6c0d-5891-ebed-4ecab1426683.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0187babf-6c0d-5891-ebed-4ecab1426683.j2c
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\textures\0098b015-3daf-4cfe-a72f-915369ea97c2.tga
SetOutPath $INSTDIR\skins\default\html
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\unabletoconnect.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\unabletoconnect.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\btn_purplepill_bg.png
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\html\btn_purplepill_bg.png
SetOutPath $INSTDIR\app_settings\windlight\water
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Valdez.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Valdez.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Second%20Plague.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Second%20Plague.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\SNAKE%21%21%21.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\SNAKE%21%21%21.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Pond.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Pond.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Murky.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Murky.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Glassy.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Glassy.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Default.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\water\Default.xml
SetOutPath $INSTDIR\app_settings\windlight\skies
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Sheer%20Surreality.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Sheer%20Surreality.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Sailor%27s%20Delight.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Sailor%27s%20Delight.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Purple.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Purple.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Pirate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Pirate.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Night.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Night.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%204.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%204.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%203.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%203.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%202.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%202.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%201.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Midday%201.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Incongruent%20Truths.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Incongruent%20Truths.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Ghost.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Ghost.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Gelatto.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Gelatto.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Funky%20Funky.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Funky%20Funky.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Funky%20Funky%20Funky.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Funky%20Funky%20Funky.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Foggy.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Foggy.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Fluffy%20Big%20Clouds.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Fluffy%20Big%20Clouds.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Fine%20Day.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Fine%20Day.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Desert%20Sunset.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Desert%20Sunset.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Default.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Default.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Coastal%20Sunset.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Coastal%20Sunset.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Coastal%20Afternoon.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Coastal%20Afternoon.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Blue%20Midday.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Blue%20Midday.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Blizzard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Blizzard.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Barcelona.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\Barcelona.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D9PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D9PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D9AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D9AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D6PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D6PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D6AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D6AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D3PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D3PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D3AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D3AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D12PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D12PM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D12AM.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\skies\A%2D12AM.xml
SetOutPath $INSTDIR\app_settings\windlight\days
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\days\Default.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\days\Default.xml
SetOutPath $INSTDIR\skins\silver
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\colors_base.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\colors_base.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\colors.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\silver\colors.xml
SetOutPath $INSTDIR\skins\default
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\colors_base.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\colors_base.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\colors.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\colors.xml
SetOutPath $INSTDIR\app_settings\windlight
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\postprocesseffects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\postprocesseffects.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\clouds2.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\windlight\clouds2.tga
SetOutPath $INSTDIR\app_settings\shaders
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\shader_heirarchy.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\shaders\shader_heirarchy.txt
SetOutPath $INSTDIR\skins
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\paths.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\paths.xml
SetOutPath $INSTDIR\llplugin
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\ssleay32.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\ssleay32.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtwebkit4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtwebkit4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtopengl4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtopengl4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtnetwork4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtnetwork4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtgui4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtgui4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtcore4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\qtcore4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\media_plugin_webkit.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\media_plugin_quicktime.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\libeay32.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\libeay32.dll
SetOutPath $INSTDIR\llplugin\imageformats
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qtiff4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qtiff4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qsvg4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qsvg4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qmng4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qmng4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qjpeg4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qjpeg4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qico4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qico4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qgif4.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\llplugin\imageformats\qgif4.dll
SetOutPath $INSTDIR\fonts
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\Readme.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\Readme.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\MtBkLfRg.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\MtBkLfRg.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\MtBdLfRg.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\MtBdLfRg.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\DejaVuSansMono.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\DejaVuSansMono.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\DejaVuSansCondensed.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\DejaVuSansCondensed.ttf
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\DejaVu-license.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\fonts\DejaVu-license.txt
SetOutPath $INSTDIR\character
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbodyfreckles_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbodyfreckles_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbody_shading_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbody_shading_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbody_highlights_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbody_highlights_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbody_color.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\upperbody_color.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\undershirt_trial_female.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\undershirt_trial_female.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\underpants_trial_male.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\underpants_trial_male.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\underpants_trial_female.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\underpants_trial_female.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_right_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_right_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_left_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_left_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_front_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_front_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_back_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_slit_back_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_length_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\skirt_length_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shoe_height_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shoe_height_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_sleeve_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_sleeve_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_collar_back_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_collar_back_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_collar_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_collar_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_bottom_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\shirt_bottom_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\rouge_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\rouge_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\rosyface_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\rosyface_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\pants_waist_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\pants_waist_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\pants_length_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\pants_length_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\nailpolish_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\nailpolish_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lowerbody_shading_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lowerbody_shading_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lowerbody_highlights_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lowerbody_highlights_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lowerbody_color.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lowerbody_color.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lipstick_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lipstick_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lips_mask.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lips_mask.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lipgloss_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\lipgloss_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_open_upper_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_open_upper_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_open_lower_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_open_lower_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_length_upper_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_length_upper_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_length_lower_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\jacket_length_lower_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_skingrain.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_skingrain.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_shading_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_shading_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_highlights_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_highlights_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_hair.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_hair.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_color.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_color.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\head_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\gloves_fingers_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\gloves_fingers_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\glove_length_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\glove_length_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\genepool.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\genepool.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\freckles_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\freckles_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_soulpatch_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_soulpatch_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_sideburns_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_sideburns_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_moustache_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_moustache_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_chincurtains_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\facehair_chincurtains_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyewhite.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyewhite.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyeshadow_outer_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyeshadow_outer_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyeshadow_inner_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyeshadow_inner_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyeliner_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyeliner_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyebrows_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\eyebrows_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_upperbody_base.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_upperbody_base.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_shirt_wrinkles.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_shirt_wrinkles.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_pants_wrinkles.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_pants_wrinkles.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_lowerbody_base.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_lowerbody_base.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_head_base.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_head_base.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_face_wrinkles.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bump_face_wrinkles.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bodyfreckles_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\bodyfreckles_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\body_skingrain.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\body_skingrain.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\blush_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\blush_alpha.tga
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_upper_body.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skirt.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skeleton.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_skeleton.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lower_body.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lad.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_lad.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_head.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_5.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_5.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_4.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_3.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_2.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_hair.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_eyelashes.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_eyelashes.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_eye_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_eye_1.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_eye.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\avatar_eye.llm
File E:\ascentviewer\indra\build-VC80\newview\Release\character\attentionsN.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\attentionsN.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\attentions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\character\attentions.xml
SetOutPath $INSTDIR\app_settings
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\viewerart.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\viewerart.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\ultra_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\ultra_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\trees.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\trees.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\std_bump.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\std_bump.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\static_index.db2
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\static_index.db2
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\static_data.db2
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\static_data.db2
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings_per_account.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings_per_account.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings_files.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings_files.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings_crash_behavior.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings_crash_behavior.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\settings.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\client_definitions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\client_definitions.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\mid_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\mid_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\low_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\low_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\logcontrol.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\logcontrol.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\keywords.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\keywords.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\keys.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\keys.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\high_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\high_graphics.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\grass.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\grass.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\cmd_line.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\cmd_line.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\anim.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\anim.ini
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\CA.pem
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\CA.pem
SetOutPath $INSTDIR
File E:\ascentviewer\indra\build-VC80\newview\Release\wrap_oal.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\wrap_oal.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\win_crash_logger.exe
File E:\ascentviewer\indra\build-VC80\newview\Release\vivoxsdk.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\vivoxsdk.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\updater.exe
SetOutPath $INSTDIR\skins\default\xui\en-us
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\skins\default\xui\en-us\mime_types.xml
SetOutPath $INSTDIR
File E:\ascentviewer\indra\build-VC80\newview\Release\ortp.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\ortp.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\openjpeg.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\openjpeg.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\msvcr80.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\msvcp80.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\licenses.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\licenses.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\gpu_table.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\gpu_table.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\fmod.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\fmod.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\fmod.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\featuretable.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\featuretable.txt
File E:\ascentviewer\indra\build-VC80\newview\Release\dbghelp.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\dbghelp.dll
SetOutPath $INSTDIR\app_settings
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\message_template.msg
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\message_template.msg
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\message.xml
File E:\ascentviewer\indra\build-VC80\newview\Release\app_settings\message.xml
SetOutPath $INSTDIR
File E:\ascentviewer\indra\build-VC80\newview\Release\alut.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\alut.dll
File E:\ascentviewer\indra\build-VC80\newview\Release\SLVoice.exe
File E:\ascentviewer\indra\build-VC80\newview\Release\SLVoice.exe
File E:\ascentviewer\indra\build-VC80\newview\Release\SLPlugin.exe
File E:\ascentviewer\indra\build-VC80\newview\Release\Microsoft.VC80.CRT.manifest
File E:\ascentviewer\indra\build-VC80\newview\Release\Ascent.exe.config
File E:\ascentviewer\indra\build-VC80\newview\Release\Ascent.exe


# Pass the installer's language to the client to use as a default
StrCpy $SHORTCUT_LANG_PARAM "--set InstallLanguage $(LanguageCode)"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Shortcuts in start menu
CreateDirectory	"$SMPROGRAMS\$INSTSHORTCUT"
SetOutPath "$INSTDIR"
CreateShortCut	"$SMPROGRAMS\$INSTSHORTCUT\$INSTSHORTCUT.lnk" \
				"$INSTDIR\$INSTEXE" "$INSTFLAGS $SHORTCUT_LANG_PARAM"


WriteINIStr		"$SMPROGRAMS\$INSTSHORTCUT\SL Create Account.url" \
				"InternetShortcut" "URL" \
				"http://www.secondlife.com/registration/"
WriteINIStr		"$SMPROGRAMS\$INSTSHORTCUT\SL Your Account.url" \
				"InternetShortcut" "URL" \
				"http://www.secondlife.com/account/"
WriteINIStr		"$SMPROGRAMS\$INSTSHORTCUT\SL Scripting Language Help.url" \
				"InternetShortcut" "URL" \
                "http://wiki.secondlife.com/wiki/LSL_Portal"
CreateShortCut	"$SMPROGRAMS\$INSTSHORTCUT\Uninstall $INSTSHORTCUT.lnk" \
				'"$INSTDIR\uninst.exe"' ''

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Other shortcuts
SetOutPath "$INSTDIR"
CreateShortCut "$DESKTOP\$INSTSHORTCUT.lnk" \
        "$INSTDIR\$INSTEXE" "$INSTFLAGS $SHORTCUT_LANG_PARAM"
CreateShortCut "$INSTDIR\$INSTSHORTCUT.lnk" \
        "$INSTDIR\$INSTEXE" "$INSTFLAGS $SHORTCUT_LANG_PARAM"
CreateShortCut "$INSTDIR\Uninstall $INSTSHORTCUT.lnk" \
				'"$INSTDIR\uninst.exe"' ''


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Write registry
WriteRegStr HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG" "" "$INSTDIR"
WriteRegStr HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG" "Version" "${VERSION_LONG}"
WriteRegStr HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG" "Flags" "$INSTFLAGS"
WriteRegStr HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG" "Shortcut" "$INSTSHORTCUT"
WriteRegStr HKEY_LOCAL_MACHINE "SOFTWARE\Linden Research, Inc.\$INSTPROG" "Exe" "$INSTEXE"
WriteRegStr HKEY_LOCAL_MACHINE "Software\Microsoft\Windows\CurrentVersion\Uninstall\$INSTPROG" "DisplayName" "$INSTPROG (remove only)"
WriteRegStr HKEY_LOCAL_MACHINE "Software\Microsoft\Windows\CurrentVersion\Uninstall\$INSTPROG" "UninstallString" '"$INSTDIR\uninst.exe"'

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Write URL registry info
WriteRegStr HKEY_CLASSES_ROOT "${URLNAME}" "(default)" "URL:Second Life"
WriteRegStr HKEY_CLASSES_ROOT "${URLNAME}" "URL Protocol" ""
WriteRegStr HKEY_CLASSES_ROOT "${URLNAME}\DefaultIcon" "" '"$INSTDIR\$INSTEXE"'
;; URL param must be last item passed to viewer, it ignores subsequent params
;; to avoid parameter injection attacks.
WriteRegExpandStr HKEY_CLASSES_ROOT "${URLNAME}\shell\open\command" "" '"$INSTDIR\$INSTEXE" $INSTFLAGS -url "%1"'

; write out uninstaller
WriteUninstaller "$INSTDIR\uninst.exe"

; end of default section
SectionEnd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EOF  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
